@@ -52,6 +52,7 @@ gboolean crm_read_options(gpointer user_data);
 gboolean fsa_has_quorum = FALSE;
 crm_trigger_t *fsa_source = NULL;
 crm_trigger_t *config_read = NULL;
+bool no_quorum_suicide_escalation = FALSE;
 
 static gboolean
 election_timeout_popped(gpointer data)
@@ -85,6 +86,7 @@ do_ha_control(long long action,
 
     if (action & A_HA_CONNECT) {
         crm_set_status_callback(&peer_update_callback);
+        crm_set_autoreap(FALSE);
 
         if (is_openais_cluster()) {
 #if SUPPORT_COROSYNC
@@ -151,6 +153,24 @@ do_ha_control(long long action,
     }
 }
 
+static bool
+need_spawn_pengine_from_crmd(void)
+{
+	static int result = -1;
+
+	if (result != -1)
+		return result;
+	if (!is_heartbeat_cluster()) {
+		result = 0;
+		return result;
+	}
+
+	/* NULL, or "strange" value: rather spawn from here. */
+	result = TRUE;
+	crm_str_to_boolean(daemon_option("crmd_spawns_pengine"), &result);
+	return result;
+}
+
 /*	 A_SHUTDOWN	*/
 void
 do_shutdown(long long action,
@@ -160,7 +180,7 @@ do_shutdown(long long action,
     /* just in case */
     set_bit(fsa_input_register, R_SHUTDOWN);
 
-    if (is_heartbeat_cluster()) {
+    if (need_spawn_pengine_from_crmd()) {
         if (is_set(fsa_input_register, pe_subsystem->flag_connected)) {
             crm_info("Terminating the %s", pe_subsystem->name);
             if (stop_subsystem(pe_subsystem, TRUE) == FALSE) {
@@ -176,7 +196,7 @@ do_shutdown(long long action,
     }
 
     if (stonith_api) {
-        /* Prevent it from comming up again */
+        /* Prevent it from coming up again */
         clear_bit(fsa_input_register, R_ST_REQUIRED);
 
         crm_info("Disconnecting STONITH...");
@@ -194,7 +214,7 @@ do_shutdown_req(long long action,
     xmlNode *msg = NULL;
 
     crm_info("Sending shutdown request to %s", crm_str(fsa_our_dc));
-    msg = create_request(CRM_OP_SHUTDOWN_REQ, NULL, NULL, CRM_SYSTEM_DC, CRM_SYSTEM_CRMD, NULL);
+    msg = create_request(CRM_OP_SHUTDOWN_REQ, NULL, NULL, CRM_SYSTEM_CRMD, CRM_SYSTEM_CRMD, NULL);
 
 /* 	set_bit(fsa_input_register, R_STAYDOWN); */
     if (send_cluster_message(NULL, crm_msg_crmd, msg, TRUE) == FALSE) {
@@ -208,7 +228,7 @@ extern char *max_generation_from;
 extern xmlNode *max_generation_xml;
 extern GHashTable *resource_history;
 extern GHashTable *voted;
-extern GHashTable *reload_hash;
+extern GHashTable *metadata_hash;
 extern char *te_client_id;
 
 void log_connected_client(gpointer key, gpointer value, gpointer user_data);
@@ -322,9 +342,9 @@ crmd_exit(int rc)
     free(te_subsystem); te_subsystem = NULL;
     free(cib_subsystem); cib_subsystem = NULL;
 
-    if (reload_hash) {
-        crm_trace("Destroying reload cache with %d members", g_hash_table_size(reload_hash));
-        g_hash_table_destroy(reload_hash); reload_hash = NULL;
+    if (metadata_hash) {
+        crm_trace("Destroying reload cache with %d members", g_hash_table_size(metadata_hash));
+        g_hash_table_destroy(metadata_hash); metadata_hash = NULL;
     }
 
     election_fini(fsa_election);
@@ -633,7 +653,7 @@ do_startup(long long action,
         was_error = TRUE;
     }
 
-    if (was_error == FALSE && is_heartbeat_cluster()) {
+    if (was_error == FALSE && need_spawn_pengine_from_crmd()) {
         if (start_subsystem(pe_subsystem) == FALSE) {
             was_error = TRUE;
         }
@@ -853,27 +873,68 @@ do_recover(long long action,
 
 /* *INDENT-OFF* */
 pe_cluster_option crmd_opts[] = {
-	/* name, old-name, validate, default, description */
-	{ "dc-version", NULL, "string", NULL, "none", NULL, "Version of Pacemaker on the cluster's DC.", "Includes the hash which identifies the exact Mercurial changeset it was built from.  Used for diagnostic purposes." },
-	{ "cluster-infrastructure", NULL, "string", NULL, "heartbeat", NULL, "The messaging stack on which Pacemaker is currently running.", "Used for informational and diagnostic purposes." },
-	{ XML_CONFIG_ATTR_DC_DEADTIME, "dc_deadtime", "time", NULL, "20s", &check_time, "How long to wait for a response from other nodes during startup.", "The \"correct\" value will depend on the speed/load of your network and the type of switches used." },
+	/* name, old-name, validate, values, default, short description, long description */
+	{ "dc-version", NULL, "string", NULL, "none", NULL,
+          "Version of Pacemaker on the cluster's DC.",
+          "Includes the hash which identifies the exact changeset it was built from.  Used for diagnostic purposes."
+        },
+	{ "cluster-infrastructure", NULL, "string", NULL, "heartbeat", NULL,
+          "The messaging stack on which Pacemaker is currently running.",
+          "Used for informational and diagnostic purposes." },
+	{ XML_CONFIG_ATTR_DC_DEADTIME, "dc_deadtime", "time", NULL, "20s", &check_time,
+          "How long to wait for a response from other nodes during startup.",
+          "The \"correct\" value will depend on the speed/load of your network and the type of switches used."
+        },
 	{ XML_CONFIG_ATTR_RECHECK, "cluster_recheck_interval", "time",
-	  "Zero disables polling.  Positive values are an interval in seconds (unless other SI units are specified. eg. 5min)", "15min", &check_timer,
+	  "Zero disables polling.  Positive values are an interval in seconds (unless other SI units are specified. eg. 5min)",
+          "15min", &check_timer,
 	  "Polling interval for time based changes to options, resource parameters and constraints.",
 	  "The Cluster is primarily event driven, however the configuration can have elements that change based on time."
-	  "  To ensure these changes take effect, we can optionally poll the cluster's status for changes." },
+	  "  To ensure these changes take effect, we can optionally poll the cluster's status for changes."
+        },
+
+#ifdef RHEL7_COMPAT
+    /* this interface is expected to change but was released in RHEL 7 */
+	{ "notification-agent", NULL, "string", NULL, "/dev/null", &check_script,
+          "Notification script or tool to be called after significant cluster events",
+          "Full path to a script or binary that will be invoked when resources start/stop/fail, fencing occurs or nodes join/leave the cluster.\n"
+          "Must exist on all nodes in the cluster."
+        },
+	{ "notification-recipient", NULL, "string", NULL, "", NULL,
+          "Destination for notifications (Optional)",
+          "Where should the supplied script send notifications to.  Useful to avoid hard-coding this in the script."
+        },
+#endif
+
 	{ "load-threshold", NULL, "percentage", NULL, "80%", &check_utilization,
 	  "The maximum amount of system resources that should be used by nodes in the cluster",
 	  "The cluster will slow down its recovery process when the amount of system resources used"
-          " (currently CPU) approaches this limit", },
+          " (currently CPU) approaches this limit",
+        },
 	{ "node-action-limit", NULL, "integer", NULL, "0", &check_number,
           "The maximum number of jobs that can be scheduled per node. Defaults to 2x cores"},
-	{ XML_CONFIG_ATTR_ELECTION_FAIL, "election_timeout", "time", NULL, "2min", &check_timer, "*** Advanced Use Only ***.", "If need to adjust this value, it probably indicates the presence of a bug." },
-	{ XML_CONFIG_ATTR_FORCE_QUIT, "shutdown_escalation", "time", NULL, "20min", &check_timer, "*** Advanced Use Only ***.", "If need to adjust this value, it probably indicates the presence of a bug." },
-	{ "crmd-integration-timeout", NULL, "time", NULL, "3min", &check_timer, "*** Advanced Use Only ***.", "If need to adjust this value, it probably indicates the presence of a bug." },
-	{ "crmd-finalization-timeout", NULL, "time", NULL, "30min", &check_timer, "*** Advanced Use Only ***.", "If you need to adjust this value, it probably indicates the presence of a bug." },
-	{ "crmd-transition-delay", NULL, "time", NULL, "0s", &check_timer, "*** Advanced Use Only ***\nEnabling this option will slow down cluster recovery under all conditions", "Delay cluster recovery for the configured interval to allow for additional/related events to occur.\nUseful if your configuration is sensitive to the order in which ping updates arrive." },
-
+	{ XML_CONFIG_ATTR_ELECTION_FAIL, "election_timeout", "time", NULL, "2min", &check_timer,
+          "*** Advanced Use Only ***.", "If need to adjust this value, it probably indicates the presence of a bug."
+        },
+	{ XML_CONFIG_ATTR_FORCE_QUIT, "shutdown_escalation", "time", NULL, "20min", &check_timer,
+          "*** Advanced Use Only ***.", "If need to adjust this value, it probably indicates the presence of a bug."
+        },
+	{ "crmd-integration-timeout", NULL, "time", NULL, "3min", &check_timer,
+          "*** Advanced Use Only ***.", "If need to adjust this value, it probably indicates the presence of a bug."
+        },
+	{ "crmd-finalization-timeout", NULL, "time", NULL, "30min", &check_timer,
+          "*** Advanced Use Only ***.", "If you need to adjust this value, it probably indicates the presence of a bug."
+        },
+	{ "crmd-transition-delay", NULL, "time", NULL, "0s", &check_timer,
+          "*** Advanced Use Only ***\n"
+          "Enabling this option will slow down cluster recovery under all conditions",
+          "Delay cluster recovery for the configured interval to allow for additional/related events to occur.\n"
+          "Useful if your configuration is sensitive to the order in which ping updates arrive."
+        },
+	{ "stonith-watchdog-timeout", NULL, "time", NULL, NULL, &check_timer,
+	  "How long to wait before we can assume nodes are safely down", NULL
+        },
+	{ "no-quorum-policy", "no_quorum_policy", "enum", "stop, freeze, ignore, suicide", "stop", &check_quorum, NULL, NULL },
 
 #if SUPPORT_PLUGIN
 	{ XML_ATTR_EXPECTED_VOTES, NULL, "integer", NULL, "2", &check_number, "The number of nodes expected to be in the cluster", "Used to calculate quorum in openais based clusters." },
@@ -905,9 +966,14 @@ crmd_pref(GHashTable * options, const char *name)
 static void
 config_query_callback(xmlNode * msg, int call_id, int rc, xmlNode * output, void *user_data)
 {
+#ifdef RHEL7_COMPAT
+    const char *script = NULL;
+#endif
     const char *value = NULL;
     GHashTable *config_hash = NULL;
     crm_time_t *now = crm_time_new(NULL);
+    long st_timeout = 0;
+    long sbd_timeout = 0;
 
     if (rc != pcmk_ok) {
         fsa_data_t *msg_data = NULL;
@@ -931,6 +997,12 @@ config_query_callback(xmlNode * msg, int call_id, int rc, xmlNode * output, void
 
     verify_crmd_options(config_hash);
 
+#ifdef RHEL7_COMPAT
+    script = crmd_pref(config_hash, "notification-agent");
+    value  = crmd_pref(config_hash, "notification-recipient");
+    crmd_enable_notifications(script, value);
+#endif
+
     value = crmd_pref(config_hash, XML_CONFIG_ATTR_DC_DEADTIME);
     election_trigger->period_ms = crm_get_msec(value);
 
@@ -942,6 +1014,43 @@ config_query_callback(xmlNode * msg, int call_id, int rc, xmlNode * output, void
         throttle_load_target = strtof(value, NULL) / 100;
     }
 
+    value = getenv("SBD_WATCHDOG_TIMEOUT");
+    sbd_timeout = crm_get_msec(value);
+
+    value = crmd_pref(config_hash, "stonith-watchdog-timeout");
+    st_timeout = crm_get_msec(value);
+
+    if(st_timeout > 0 && !daemon_option_enabled(crm_system_name, "watchdog")) {
+        do_crm_log_always(LOG_EMERG, "Shutting down pacemaker, no watchdog device configured");
+        crmd_exit(DAEMON_RESPAWN_STOP);
+
+    } else if(!daemon_option_enabled(crm_system_name, "watchdog")) {
+        crm_trace("Watchdog disabled");
+
+    } else if(value == NULL && sbd_timeout > 0) {
+        char *timeout = NULL;
+
+        st_timeout = 2 * sbd_timeout / 1000;
+        timeout = crm_strdup_printf("%lds", st_timeout);
+        crm_notice("Setting stonith-watchdog-timeout=%s", timeout);
+
+        update_attr_delegate(fsa_cib_conn, cib_none, XML_CIB_TAG_CRMCONFIG, NULL, NULL, NULL, NULL,
+                             "stonith-watchdog-timeout", timeout, FALSE, NULL, NULL);
+        free(timeout);
+
+    } else if(st_timeout <= 0) {
+        crm_notice("Watchdog enabled but stonith-watchdog-timeout is disabled");
+
+    } else if(st_timeout < sbd_timeout) {
+        do_crm_log_always(LOG_EMERG, "Shutting down pacemaker, stonith-watchdog-timeout (%ldms) is too short (must be greater than %ldms)",
+                          st_timeout, sbd_timeout);
+        crmd_exit(DAEMON_RESPAWN_STOP);
+    }
+
+    value = crmd_pref(config_hash, "no-quorum-policy");
+    if (safe_str_eq(value, "suicide") && daemon_option_enabled(crm_system_name, "watchdog")) {
+        no_quorum_suicide_escalation = TRUE;
+    }
 
     value = crmd_pref(config_hash, XML_CONFIG_ATTR_FORCE_QUIT);
     shutdown_escalation_timer->period_ms = crm_get_msec(value);
@@ -1031,7 +1140,7 @@ crm_shutdown(int nsig)
                 shutdown_escalation_timer->period_ms = msec;
             }
 
-            /* cant rely on this... */
+            /* can't rely on this... */
             crm_notice("Requesting shutdown, upper limit is %dms",
                        shutdown_escalation_timer->period_ms);
             crm_timer_start(shutdown_escalation_timer);
