@@ -39,6 +39,7 @@
 GListPtr fsa_message_queue = NULL;
 extern void crm_shutdown(int nsig);
 
+extern crm_ipc_t *attrd_ipc;
 void handle_response(xmlNode * stored_msg);
 enum crmd_fsa_input handle_request(xmlNode * stored_msg, enum crmd_fsa_cause cause);
 enum crmd_fsa_input handle_shutdown_request(xmlNode * stored_msg);
@@ -175,7 +176,7 @@ register_fsa_input_adv(enum crmd_fsa_cause cause, enum crmd_fsa_input input,
     /* fsa_dump_queue(LOG_DEBUG_2); */
 
     if (old_len == g_list_length(fsa_message_queue)) {
-        crm_err("Couldnt add message to the queue");
+        crm_err("Couldn't add message to the queue");
     }
 
     if (fsa_source && input != I_WAIT_FOR_EVENT) {
@@ -251,7 +252,7 @@ delete_fsa_input(fsa_data_t * fsa_data)
 
             case fsa_dt_none:
                 if (fsa_data->data != NULL) {
-                    crm_err("Dont know how to free %s data from %s",
+                    crm_err("Don't know how to free %s data from %s",
                             fsa_cause2string(fsa_data->fsa_cause), fsa_data->origin);
                     crmd_exit(pcmk_err_generic);
                 }
@@ -666,6 +667,16 @@ handle_request(xmlNode * stored_msg, enum crmd_fsa_cause cause)
         return I_NULL;
     }
 
+    if (strcmp(op, CRM_OP_SHUTDOWN_REQ) == 0) {
+        const char *from = crm_element_value(stored_msg, F_CRM_HOST_FROM);
+        crm_node_t *node = crm_find_peer(0, from);
+
+        crm_update_peer_expected(__FUNCTION__, node, CRMD_JOINSTATE_DOWN);
+        if(AM_I_DC == FALSE) {
+            return I_NULL; /* Done */
+        }
+    }
+
     /*========== DC-Only Actions ==========*/
     if (AM_I_DC) {
         if (strcmp(op, CRM_OP_JOIN_ANNOUNCE) == 0) {
@@ -686,8 +697,8 @@ handle_request(xmlNode * stored_msg, enum crmd_fsa_cause cause)
                 return I_STOP;
 
             } else if (dc_match) {
-                crm_err("We didnt ask to be shut down, yet our"
-                        " TE is telling us too." " Better get out now!");
+                crm_err("We didn't ask to be shut down, yet our"
+                        " TE is telling us to. Better get out now!");
                 return I_TERMINATE;
 
             } else if (fsa_state != S_STOPPING) {
@@ -768,7 +779,7 @@ handle_request(xmlNode * stored_msg, enum crmd_fsa_cause cause)
 
         if (dc_match || fsa_our_dc == NULL) {
             if (is_set(fsa_input_register, R_SHUTDOWN) == FALSE) {
-                crm_err("We didn't ask to be shut down, yet our" " DC is telling us too.");
+                crm_err("We didn't ask to be shut down, yet our DC is telling us to.");
                 set_bit(fsa_input_register, R_STAYDOWN);
                 return I_STOP;
             }
@@ -794,8 +805,8 @@ handle_request(xmlNode * stored_msg, enum crmd_fsa_cause cause)
         crm_notice("Current ping state: %s", fsa_state2string(fsa_state));
 
         msg = create_reply(stored_msg, ping);
-        if(msg) {
-            relay_message(msg, TRUE);
+        if (msg) {
+            (void)relay_message(msg, TRUE);
         }
 
         free_xml(ping);
@@ -838,7 +849,7 @@ handle_response(xmlNode * stored_msg)
         crm_log_xml_err(stored_msg, "Bad message");
 
     } else if (AM_I_DC && strcmp(op, CRM_OP_PECALC) == 0) {
-        /* Check if the PE answer been superceeded by a subsequent request? */
+        /* Check if the PE answer been superseded by a subsequent request? */
         const char *msg_ref = crm_element_value(stored_msg, XML_ATTR_REFERENCE);
 
         if (msg_ref == NULL) {
@@ -870,7 +881,7 @@ enum crmd_fsa_input
 handle_shutdown_request(xmlNode * stored_msg)
 {
     /* handle here to avoid potential version issues
-     *   where the shutdown message/proceedure may have
+     *   where the shutdown message/procedure may have
      *   been changed in later versions.
      *
      * This way the DC is always in control of the shutdown
