@@ -31,7 +31,96 @@
 #include <crm/fencing/internal.h>
 
 crm_trigger_t *stonith_reconnect = NULL;
-GListPtr stonith_cleanup_list = NULL;
+
+/*
+ * stonith cleanup list
+ *
+ * If the DC is shot, proper notifications might not go out.
+ * The stonith cleanup list allows the cluster to (re-)send
+ * notifications once a new DC is elected.
+ */
+
+static GListPtr stonith_cleanup_list = NULL;
+
+/*!
+ * \internal
+ * \brief Add a node to the stonith cleanup list
+ *
+ * \param[in] target  Name of node to add
+ */
+void
+add_stonith_cleanup(const char *target) {
+    stonith_cleanup_list = g_list_append(stonith_cleanup_list, strdup(target));
+}
+
+/*!
+ * \internal
+ * \brief Remove a node from the stonith cleanup list
+ *
+ * \param[in] Name of node to remove
+ */
+void
+remove_stonith_cleanup(const char *target)
+{
+    GListPtr iter = stonith_cleanup_list;
+
+    while (iter != NULL) {
+        GListPtr tmp = iter;
+        char *iter_name = tmp->data;
+
+        iter = iter->next;
+        if (safe_str_eq(target, iter_name)) {
+            crm_trace("Removing %s from the cleanup list", iter_name);
+            stonith_cleanup_list = g_list_delete_link(stonith_cleanup_list, tmp);
+            free(iter_name);
+        }
+    }
+}
+
+/*!
+ * \internal
+ * \brief Purge all entries from the stonith cleanup list
+ */
+void
+purge_stonith_cleanup()
+{
+    if (stonith_cleanup_list) {
+        GListPtr iter = NULL;
+
+        for (iter = stonith_cleanup_list; iter != NULL; iter = iter->next) {
+            char *target = iter->data;
+
+            crm_info("Purging %s from stonith cleanup list", target);
+            free(target);
+        }
+        g_list_free(stonith_cleanup_list);
+        stonith_cleanup_list = NULL;
+    }
+}
+
+/*!
+ * \internal
+ * \brief Send stonith updates for all entries in cleanup list, then purge it
+ */
+void
+execute_stonith_cleanup()
+{
+    GListPtr iter;
+
+    for (iter = stonith_cleanup_list; iter != NULL; iter = iter->next) {
+        char *target = iter->data;
+        crm_node_t *target_node = crm_get_peer(0, target);
+        const char *uuid = crm_peer_uuid(target_node);
+
+        crm_notice("Marking %s, target of a previous stonith action, as clean", target);
+        send_stonith_update(NULL, target, uuid);
+        free(target);
+    }
+    g_list_free(stonith_cleanup_list);
+    stonith_cleanup_list = NULL;
+}
+
+/* end stonith cleanup list functions */
 
 static gboolean
 fail_incompletable_stonith(crm_graph_t * graph)
@@ -116,7 +205,7 @@ static void
 tengine_stonith_notify(stonith_t * st, stonith_event_t * st_event)
 {
     if(te_client_id == NULL) {
-        te_client_id = g_strdup_printf("%s.%d", crm_system_name, getpid());
+        te_client_id = crm_strdup_printf("%s.%d", crm_system_name, getpid());
     }
 
     if (st_event == NULL) {
@@ -124,21 +213,23 @@ tengine_stonith_notify(stonith_t * st, stonith_event_t * st_event)
         return;
     }
 
+    crmd_notify_fencing_op(st_event);
+
     if (st_event->result == pcmk_ok && safe_str_eq("on", st_event->action)) {
         crm_notice("%s was successfully unfenced by %s (at the request of %s)",
-                   st_event->target, st_event->executioner, st_event->origin);
+                   st_event->target, st_event->executioner ? st_event->executioner : "<anyone>", st_event->origin);
                 /* TODO: Hook up st_event->device */
         return;
 
     } else if (safe_str_eq("on", st_event->action)) {
         crm_err("Unfencing of %s by %s failed: %s (%d)",
-                st_event->target, st_event->executioner,
+                st_event->target, st_event->executioner ? st_event->executioner : "<anyone>",
                 pcmk_strerror(st_event->result), st_event->result);
         return;
 
     } else if (st_event->result == pcmk_ok && crm_str_eq(st_event->target, fsa_our_uname, TRUE)) {
-        crm_crit("We were alegedly just fenced by %s for %s!",
-                 st_event->executioner, st_event->origin); /* Dumps blackbox if enabled */
+        crm_crit("We were allegedly just fenced by %s for %s!",
+                 st_event->executioner ? st_event->executioner : "<anyone>", st_event->origin); /* Dumps blackbox if enabled */
 
         qb_log_fini(); /* Try to get the above log message to disk - somehow */
 
@@ -199,7 +290,7 @@ tengine_stonith_notify(stonith_t * st, stonith_event_t * st_event)
 
         /* In case fenced is already trying to shoot it */
         confirm = open("/var/run/cluster/fenced_override", O_NONBLOCK|O_WRONLY);
-        if (confirm > 0) {
+        if (confirm >= 0) {
             int ignore = 0;
             int len = strlen(target_copy);
 
@@ -224,7 +315,7 @@ tengine_stonith_notify(stonith_t * st, stonith_event_t * st_event)
 #endif
 
     if (st_event->result == pcmk_ok) {
-        crm_node_t *peer = crm_find_peer_full(0, st_event->target, CRM_GET_PEER_REMOTE | CRM_GET_PEER_CLUSTER);
+        crm_node_t *peer = crm_find_peer_full(0, st_event->target, CRM_GET_PEER_ANY);
         const char *uuid = NULL;
         gboolean we_are_executioner = safe_str_eq(st_event->executioner, fsa_our_uname);
 
@@ -249,7 +340,9 @@ tengine_stonith_notify(stonith_t * st, stonith_event_t * st_event)
             }
 
             /* Assume it was our leader if we dont currently have one */
-        } else if (fsa_our_dc == NULL || safe_str_eq(fsa_our_dc, st_event->target)) {
+        } else if (((fsa_our_dc == NULL) || safe_str_eq(fsa_our_dc, st_event->target))
+            && !is_set(peer->flags, crm_remote_node)) {
+
             crm_notice("Target %s our leader %s (recorded: %s)",
                        fsa_our_dc ? "was" : "may have been", st_event->target,
                        fsa_our_dc ? fsa_our_dc : "<unset>");
@@ -261,8 +354,7 @@ tengine_stonith_notify(stonith_t * st, stonith_event_t * st_event)
             if (we_are_executioner) {
                 send_stonith_update(NULL, st_event->target, uuid);
             }
-            stonith_cleanup_list = g_list_append(stonith_cleanup_list, strdup(st_event->target));
-
+            add_stonith_cleanup(st_event->target);
         }
 
         crmd_peer_down(peer, TRUE);
@@ -295,8 +387,12 @@ te_connect_stonith(gpointer user_data)
         }
 
         if (user_data != NULL) {
-            crm_err("Sign-in failed: triggered a retry");
-            mainloop_set_trigger(stonith_reconnect);
+            if (is_set(fsa_input_register, R_ST_REQUIRED)) {
+                crm_err("Sign-in failed: triggered a retry");
+                mainloop_set_trigger(stonith_reconnect);
+            } else {
+                crm_info("Sign-in failed, but no longer required");
+            }
             return TRUE;
         }
 
@@ -374,7 +470,8 @@ te_graph_trigger(gpointer user_data)
         graph_rc = run_graph(transition_graph);
         transition_graph->batch_limit = limit; /* Restore the configured value */
 
-        print_graph(LOG_DEBUG_3, transition_graph);
+        /* significant overhead... */
+        /* print_graph(LOG_DEBUG_3, transition_graph); */
 
         if (graph_rc == transition_active) {
             crm_trace("Transition not yet complete");
