@@ -69,7 +69,6 @@ int throttle_num_cores(void)
     if(cores) {
         return cores;
     }
-
     stream = fopen(cpufile, "r");
     if(stream == NULL) {
         int rc = errno;
@@ -93,66 +92,21 @@ int throttle_num_cores(void)
     return cores;
 }
 
+/*
+ * \internal
+ * \brief Return name of /proc file containing the CIB deamon's load statistics
+ *
+ * \return Newly allocated memory with file name on success, NULL otherwise
+ *
+ * \note It is the caller's responsibility to free the return value.
+ *       This will return NULL if the daemon is being run via valgrind.
+ *       This should be called only on Linux systems.
+ */
 static char *find_cib_loadfile(void) 
 {
-    DIR *dp;
-    struct dirent *entry;
-    struct stat statbuf;
-    char *match = NULL;
+    int pid = crm_procfs_pid_of("cib");
 
-    dp = opendir("/proc");
-    if (!dp) {
-        /* no proc directory to search through */
-        crm_notice("Can not read /proc directory to track existing components");
-        return FALSE;
-    }
-
-    while ((entry = readdir(dp)) != NULL) {
-        char procpath[128];
-        char value[64];
-        char key[16];
-        FILE *file;
-        int pid;
-
-        strcpy(procpath, "/proc/");
-        /* strlen("/proc/") + strlen("/status") + 1 = 14
-         * 128 - 14 = 114 */
-        strncat(procpath, entry->d_name, 114);
-
-        if (lstat(procpath, &statbuf)) {
-            continue;
-        }
-        if (!S_ISDIR(statbuf.st_mode) || !isdigit(entry->d_name[0])) {
-            continue;
-        }
-
-        strcat(procpath, "/status");
-
-        file = fopen(procpath, "r");
-        if (!file) {
-            continue;
-        }
-        if (fscanf(file, "%15s%63s", key, value) != 2) {
-            fclose(file);
-            continue;
-        }
-        fclose(file);
-
-        if (safe_str_neq("cib", value)) {
-            continue;
-        }
-
-        pid = atoi(entry->d_name);
-        if (pid <= 0) {
-            continue;
-        }
-
-        match = g_strdup_printf("/proc/%d/stat", pid);
-        break;
-    }
-
-    closedir(dp);
-    return match;
+    return pid? crm_strdup_printf("/proc/%d/stat", pid) : NULL;
 }
 
 static bool throttle_cib_load(float *load) 
@@ -215,6 +169,10 @@ static bool throttle_cib_load(float *load)
         last_utime = 0;
         last_stime = 0;
         loadfile = find_cib_loadfile();
+        if (loadfile == NULL) {
+            crm_warn("Couldn't find CIB load file");
+            return FALSE;
+        }
         ticks_per_s = sysconf(_SC_CLK_TCK);
         crm_trace("Found %s", loadfile);
     }
@@ -415,7 +373,7 @@ throttle_handle_load(float load, const char *desc, int cores)
         return throttle_med;
 
     } else if(adjusted_load > THROTTLE_FACTOR_LOW * throttle_load_target) {
-        crm_debug("Noticable %s detected: %f", desc, load);
+        crm_debug("Noticeable %s detected: %f", desc, load);
         return throttle_low;
     }
 
@@ -426,12 +384,16 @@ throttle_handle_load(float load, const char *desc, int cores)
 static enum throttle_state_e
 throttle_mode(void)
 {
+    int cores;
     float load;
     unsigned int blocked = 0;
-    int cores = throttle_num_cores();
-
     enum throttle_state_e mode = throttle_none;
 
+#ifdef ON_SOLARIS
+    return throttle_none;
+#endif
+
+    cores = throttle_num_cores();
     if(throttle_cib_load(&load)) {
         float cib_max_cpu = 0.95;
         const char *desc = "CIB load";
@@ -467,7 +429,7 @@ throttle_mode(void)
             mode |= throttle_med;
 
         } else if(load > cib_max_cpu * 0.8) {
-            crm_debug("Noticable %s detected: %f", desc, load);
+            crm_debug("Noticeable %s detected: %f", desc, load);
             mode |= throttle_low;
 
         } else {
@@ -505,44 +467,41 @@ static void
 throttle_send_command(enum throttle_state_e mode)
 {
     xmlNode *xml = NULL;
+    static enum throttle_state_e last = -1;
 
-    xml = create_request(CRM_OP_THROTTLE, NULL, NULL, CRM_SYSTEM_CRMD, CRM_SYSTEM_CRMD, NULL);
-    crm_xml_add_int(xml, F_CRM_THROTTLE_MODE, mode);
-    crm_xml_add_int(xml, F_CRM_THROTTLE_MAX, throttle_job_max);
+    if(mode != last) {
+        crm_info("New throttle mode: %.4x (was %.4x)", mode, last);
+        last = mode;
 
-    send_cluster_message(NULL, crm_msg_crmd, xml, TRUE);
-    free_xml(xml);
+        xml = create_request(CRM_OP_THROTTLE, NULL, NULL, CRM_SYSTEM_CRMD, CRM_SYSTEM_CRMD, NULL);
+        crm_xml_add_int(xml, F_CRM_THROTTLE_MODE, mode);
+        crm_xml_add_int(xml, F_CRM_THROTTLE_MAX, throttle_job_max);
 
-    crm_info("Updated throttle state to %.4x", mode);
+        send_cluster_message(NULL, crm_msg_crmd, xml, TRUE);
+        free_xml(xml);
+    }
 }
 
 static gboolean
 throttle_timer_cb(gpointer data)
 {
     static bool send_updates = FALSE;
-    static enum throttle_state_e last = -1;
-
     enum throttle_state_e now = throttle_none;
 
-    if(send_updates == FALSE) {
-        /* Optimize for the true case */
-        if(compare_version(fsa_our_dc_version, "3.0.8") < 0) {
-            crm_trace("DC version %s doesn't support throttling", fsa_our_dc_version);
-
-        } else {
-            send_updates = TRUE;
-        }
-    }
-
     if(send_updates) {
         now = throttle_mode();
+        throttle_send_command(now);
+
+    } else if(compare_version(fsa_our_dc_version, "3.0.8") < 0) {
+        /* Optimize for the true case */
+        crm_trace("DC version %s doesn't support throttling", fsa_our_dc_version);
+
+    } else {
+        send_updates = TRUE;
+        now = throttle_mode();
+        throttle_send_command(now);
     }
 
-    if(send_updates && now != last) {
-        crm_debug("New throttle mode: %.4x (was %.4x)", now, last);
-        throttle_send_command(now);
-        last = now;
-    }
     return TRUE;
 }
 
@@ -592,9 +551,11 @@ throttle_update_job_max(const char *preference)
 void
 throttle_init(void)
 {
-    throttle_records = g_hash_table_new_full(
-        crm_str_hash, g_str_equal, NULL, throttle_record_free);
-    throttle_timer = mainloop_timer_add("throttle", 30* 1000, TRUE, throttle_timer_cb, NULL);
+    if(throttle_records == NULL) {
+        throttle_records = g_hash_table_new_full(
+            crm_str_hash, g_str_equal, NULL, throttle_record_free);
+        throttle_timer = mainloop_timer_add("throttle", 30 * 1000, TRUE, throttle_timer_cb, NULL);
+    }
 
     throttle_update_job_max(NULL);
     mainloop_timer_start(throttle_timer);
