@@ -226,6 +226,8 @@ cib_tls_signon(cib_t * cib, crm_remote_t * connection, gboolean event_channel)
         return -ENOTCONN;
     }
 
+    connection->tcp_socket = sock;
+
     if (private->encrypted) {
         /* initialize GnuTls lib */
 #ifdef HAVE_GNUTLS_GNUTLS_H
@@ -250,8 +252,6 @@ cib_tls_signon(cib_t * cib, crm_remote_t * connection, gboolean event_channel)
 #else
         return -EPROTONOSUPPORT;
 #endif
-    } else {
-        connection->tcp_socket = sock;
     }
 
     /* login to server */
@@ -394,8 +394,8 @@ cib_remote_signon(cib_t * cib, const char *name, enum cib_conn_type type)
         if(rc == 0) {
             fprintf(stderr, "Password: ");
             private->passwd = calloc(1, 1024);
-            rc = scanf("%s", private->passwd);
-            fprintf(stdout, "\n");
+            rc = scanf("%1023s", private->passwd);
+            fprintf(stderr, "\n");
         }
 
         /* fprintf(stderr, "entered: '%s'\n", buffer); */
