@@ -67,25 +67,25 @@ const int anon_tls_kx_order[] = {
  * casts are necessary for constants, because we never know how for sure
  * how U/UL/ULL map to __u16, __u32, __u64. At least not in a portable way.
  */
-#define ___swab16(x) ((__u16)(                                  \
-        (((__u16)(x) & (__u16)0x00ffU) << 8) |                  \
-        (((__u16)(x) & (__u16)0xff00U) >> 8)))
+#define __swab16(x) ((uint16_t)(                                      \
+        (((uint16_t)(x) & (uint16_t)0x00ffU) << 8) |                  \
+        (((uint16_t)(x) & (uint16_t)0xff00U) >> 8)))
 
-#define ___swab32(x) ((__u32)(                                  \
-        (((__u32)(x) & (__u32)0x000000ffUL) << 24) |            \
-        (((__u32)(x) & (__u32)0x0000ff00UL) <<  8) |            \
-        (((__u32)(x) & (__u32)0x00ff0000UL) >>  8) |            \
-        (((__u32)(x) & (__u32)0xff000000UL) >> 24)))
+#define __swab32(x) ((uint32_t)(                                      \
+        (((uint32_t)(x) & (uint32_t)0x000000ffUL) << 24) |            \
+        (((uint32_t)(x) & (uint32_t)0x0000ff00UL) <<  8) |            \
+        (((uint32_t)(x) & (uint32_t)0x00ff0000UL) >>  8) |            \
+        (((uint32_t)(x) & (uint32_t)0xff000000UL) >> 24)))
 
-#define ___swab64(x) ((__u64)(                                  \
-        (((__u64)(x) & (__u64)0x00000000000000ffULL) << 56) |   \
-        (((__u64)(x) & (__u64)0x000000000000ff00ULL) << 40) |   \
-        (((__u64)(x) & (__u64)0x0000000000ff0000ULL) << 24) |   \
-        (((__u64)(x) & (__u64)0x00000000ff000000ULL) <<  8) |   \
-        (((__u64)(x) & (__u64)0x000000ff00000000ULL) >>  8) |   \
-        (((__u64)(x) & (__u64)0x0000ff0000000000ULL) >> 24) |   \
-        (((__u64)(x) & (__u64)0x00ff000000000000ULL) >> 40) |   \
-        (((__u64)(x) & (__u64)0xff00000000000000ULL) >> 56)))
+#define __swab64(x) ((uint64_t)(                                      \
+        (((uint64_t)(x) & (uint64_t)0x00000000000000ffULL) << 56) |   \
+        (((uint64_t)(x) & (uint64_t)0x000000000000ff00ULL) << 40) |   \
+        (((uint64_t)(x) & (uint64_t)0x0000000000ff0000ULL) << 24) |   \
+        (((uint64_t)(x) & (uint64_t)0x00000000ff000000ULL) <<  8) |   \
+        (((uint64_t)(x) & (uint64_t)0x000000ff00000000ULL) >>  8) |   \
+        (((uint64_t)(x) & (uint64_t)0x0000ff0000000000ULL) >> 24) |   \
+        (((uint64_t)(x) & (uint64_t)0x00ff000000000000ULL) >> 40) |   \
+        (((uint64_t)(x) & (uint64_t)0xff00000000000000ULL) >> 56)))
 #endif
 
 #define REMOTE_MSG_VERSION 1
@@ -247,11 +247,11 @@ crm_send_tls(gnutls_session_t * session, const char *buf, size_t len)
             break;
 
         } else if (rc < len) {
-            crm_debug("Only sent %d of %d bytes", rc, len);
+            crm_debug("Sent %d of %d bytes", rc, len);
             len -= rc;
             unsent += rc;
         } else {
-            crm_debug("Sent %d bytes", rc);
+            crm_trace("Sent all %d bytes", rc);
             break;
         }
     }
@@ -308,13 +308,16 @@ crm_remote_sendv(crm_remote_t * remote, struct iovec * iov, int iovs)
     int rc = -ESOCKTNOSUPPORT;
 
     for(; lpc < iovs; lpc++) {
-        if (remote->tcp_socket) {
-            rc = crm_send_plaintext(remote->tcp_socket, iov[lpc].iov_base, iov[lpc].iov_len);
-#ifdef HAVE_GNUTLS_GNUTLS_H
 
-        } else if (remote->tls_session) {
+#ifdef HAVE_GNUTLS_GNUTLS_H
+        if (remote->tls_session) {
             rc = crm_send_tls(remote->tls_session, iov[lpc].iov_base, iov[lpc].iov_len);
+        } else if (remote->tcp_socket) {
+#else
+        if (remote->tcp_socket) {
 #endif
+            rc = crm_send_plaintext(remote->tcp_socket, iov[lpc].iov_base, iov[lpc].iov_len);
+
         } else {
             crm_err("Unsupported connection type");
         }
@@ -436,7 +439,7 @@ crm_remote_parse_buffer(crm_remote_t * remote)
  * \internal
  * \brief Determine if a remote session has data to read
  *
- * \retval 0, timeout occured.
+ * \retval 0, timeout occurred.
  * \retval positive, data is ready to be read
  * \retval negative, session has ended
  */
@@ -448,14 +451,16 @@ crm_remote_ready(crm_remote_t * remote, int timeout /* ms */ )
     int rc = 0;
     time_t start;
 
-    if (remote->tcp_socket) {
-        sock = remote->tcp_socket;
 #ifdef HAVE_GNUTLS_GNUTLS_H
-    } else if (remote->tls_session) {
+    if (remote->tls_session) {
         void *sock_ptr = gnutls_transport_get_ptr(*remote->tls_session);
 
         sock = GPOINTER_TO_INT(sock_ptr);
+    } else if (remote->tcp_socket) {
+#else
+    if (remote->tcp_socket) {
 #endif
+        sock = remote->tcp_socket;
     } else {
         crm_err("Unsupported connection type");
     }
@@ -515,21 +520,12 @@ crm_remote_recv_once(crm_remote_t * remote)
            remote->buffer_size = 2 * read_len;
         crm_trace("Expanding buffer to %u bytes", remote->buffer_size);
 
-        remote->buffer = realloc(remote->buffer, remote->buffer_size + 1);
+        remote->buffer = realloc_safe(remote->buffer, remote->buffer_size + 1);
         CRM_ASSERT(remote->buffer != NULL);
     }
 
-    if (remote->tcp_socket) {
-        errno = 0;
-        rc = read(remote->tcp_socket,
-                  remote->buffer + remote->buffer_offset,
-                  remote->buffer_size - remote->buffer_offset);
-        if(rc < 0) {
-            rc = -errno;
-        }
-
 #ifdef HAVE_GNUTLS_GNUTLS_H
-    } else if (remote->tls_session) {
+    if (remote->tls_session) {
         rc = gnutls_record_recv(*(remote->tls_session),
                                 remote->buffer + remote->buffer_offset,
                                 remote->buffer_size - remote->buffer_offset);
@@ -541,7 +537,18 @@ crm_remote_recv_once(crm_remote_t * remote)
             crm_debug("TLS receive failed: %s (%d)", gnutls_strerror(rc), rc);
             rc = -pcmk_err_generic;
         }
+    } else if (remote->tcp_socket) {
+#else
+    if (remote->tcp_socket) {
 #endif
+        errno = 0;
+        rc = read(remote->tcp_socket,
+                  remote->buffer + remote->buffer_offset,
+                  remote->buffer_size - remote->buffer_offset);
+        if(rc < 0) {
+            rc = -errno;
+        }
+
     } else {
         crm_err("Unsupported connection type");
         return -ESOCKTNOSUPPORT;
@@ -614,6 +621,10 @@ crm_remote_recv(crm_remote_t * remote, int total_timeout /*ms */ , int *disconne
             crm_err("poll timed out (%d ms) while waiting to receive msg", remaining_timeout);
             return FALSE;
 
+        } else if (rc == -EAGAIN) {
+            crm_trace("waiting for remote connection data (up to %dms)",
+                      remaining_timeout);
+
         } else if(rc < 0) {
             crm_debug("poll() failed: %s (%d)", pcmk_strerror(rc), rc);
 
@@ -737,11 +748,12 @@ check_connect_finished(gpointer userdata)
 static int
 internal_tcp_connect_async(int sock,
                            const struct sockaddr *addr, socklen_t addrlen, int timeout /* ms */ ,
-                           void *userdata, void (*callback) (void *userdata, int sock))
+                           int *timer_id, void *userdata, void (*callback) (void *userdata, int sock))
 {
     int rc = 0;
     int flag = 0;
     int interval = 500;
+    int timer;
     struct tcp_async_cb_data *cb_data = NULL;
 
     if ((flag = fcntl(sock, F_GETFL)) >= 0) {
@@ -782,7 +794,10 @@ internal_tcp_connect_async(int sock,
      * Something about the way mainloop is currently polling prevents this from working at the
      * moment though. */
     crm_trace("fd %d: scheduling to check if connect finished in %dms second", sock, interval);
-    g_timeout_add(interval, check_connect_finished, cb_data);
+    timer = g_timeout_add(interval, check_connect_finished, cb_data);
+    if (timer_id) {
+        *timer_id = timer;
+    }
 
     return 0;
 }
@@ -809,10 +824,11 @@ internal_tcp_connect(int sock, const struct sockaddr *addr, socklen_t addrlen)
  * \internal
  * \brief tcp connection to server at specified port
  * \retval negative, failed to connect.
+ * \retval positive, sock fd
  */
 int
-crm_remote_tcp_connect_async(const char *host, int port, int timeout,   /*ms */
-                             void *userdata, void (*callback) (void *userdata, int sock))
+crm_remote_tcp_connect_async(const char *host, int port, int timeout, /*ms */
+                             int *timer_id, void *userdata, void (*callback) (void *userdata, int sock))
 {
     char buffer[256];
     struct addrinfo *res = NULL;
@@ -877,8 +893,7 @@ crm_remote_tcp_connect_async(const char *host, int port, int timeout,   /*ms */
 
         if (callback) {
             if (internal_tcp_connect_async
-                (sock, rp->ai_addr, rp->ai_addrlen, timeout, userdata, callback) == 0) {
-                sock = 0;
+                (sock, rp->ai_addr, rp->ai_addrlen, timeout, timer_id, userdata, callback) == 0) {
                 goto async_cleanup; /* Success for now, we'll hear back later in the callback */
             }
 
@@ -903,5 +918,5 @@ async_cleanup:
 int
 crm_remote_tcp_connect(const char *host, int port)
 {
-    return crm_remote_tcp_connect_async(host, port, -1, NULL, NULL);
+    return crm_remote_tcp_connect_async(host, port, -1, NULL, NULL, NULL);
 }
