@@ -36,7 +36,6 @@
 #include <limits.h>
 #include <ctype.h>
 #include <pwd.h>
-#include <grp.h>
 #include <time.h>
 #include <libgen.h>
 #include <signal.h>
@@ -162,6 +161,51 @@ check_number(const char *value)
     return TRUE;
 }
 
+gboolean
+check_quorum(const char *value)
+{
+    if (safe_str_eq(value, "stop")) {
+        return TRUE;
+
+    } else if (safe_str_eq(value, "freeze")) {
+        return TRUE;
+
+    } else if (safe_str_eq(value, "ignore")) {
+        return TRUE;
+
+    } else if (safe_str_eq(value, "suicide")) {
+        return TRUE;
+    }
+    return FALSE;
+}
+
+gboolean
+check_script(const char *value)
+{
+    struct stat st;
+
+    if(safe_str_eq(value, "/dev/null")) {
+        return TRUE;
+    }
+
+    if(stat(value, &st) != 0) {
+        crm_err("Script %s does not exist", value);
+        return FALSE;
+    }
+
+    if(S_ISREG(st.st_mode) == 0) {
+        crm_err("Script %s is not a regular file", value);
+        return FALSE;
+    }
+
+    if( (st.st_mode & (S_IXUSR | S_IXGRP )) == 0) {
+        crm_err("Script %s is not executable", value);
+        return FALSE;
+    }
+
+    return TRUE;
+}
+
 gboolean
 check_utilization(const char *value)
 {
@@ -268,6 +312,9 @@ cluster_option(GHashTable * options, gboolean(*validate) (const char *),
 
         if (options == NULL) {
             return def_value;
+
+        } else if(def_value == NULL) {
+            return def_value;
         }
 
         g_hash_table_insert(options, strdup(name), strdup(def_value));
@@ -301,7 +348,6 @@ get_cluster_pref(GHashTable * options, pe_cluster_option * option_list, int len,
         }
     }
     CRM_CHECK(found, crm_err("No option named: %s", name));
-    CRM_ASSERT(value != NULL);
     return value;
 }
 
@@ -405,30 +451,6 @@ crm_itoa(int an_int)
     return buffer;
 }
 
-void
-crm_build_path(const char *path_c, mode_t mode)
-{
-    int offset = 1, len = 0;
-    char *path = strdup(path_c);
-
-    CRM_CHECK(path != NULL, return);
-    for (len = strlen(path); offset < len; offset++) {
-        if (path[offset] == '/') {
-            path[offset] = 0;
-            if (mkdir(path, mode) < 0 && errno != EEXIST) {
-                crm_perror(LOG_ERR, "Could not create directory '%s'", path);
-                break;
-            }
-            path[offset] = '/';
-        }
-    }
-    if (mkdir(path, mode) < 0 && errno != EEXIST) {
-        crm_perror(LOG_ERR, "Could not create directory '%s'", path);
-    }
-
-    free(path);
-}
-
 int
 crm_user_lookup(const char *name, uid_t * uid, gid_t * gid)
 {
@@ -1097,43 +1119,6 @@ filter_action_parameters(xmlNode * param_set, const char *version)
     free(key);
 }
 
-void
-filter_reload_parameters(xmlNode * param_set, const char *restart_string)
-{
-    int len = 0;
-    char *name = NULL;
-    char *match = NULL;
-
-    if (param_set == NULL) {
-        return;
-    }
-
-    if (param_set) {
-        xmlAttrPtr xIter = param_set->properties;
-
-        while (xIter) {
-            const char *prop_name = (const char *)xIter->name;
-
-            xIter = xIter->next;
-            name = NULL;
-            len = strlen(prop_name) + 3;
-
-            name = malloc(len);
-            if(name) {
-                sprintf(name, " %s ", prop_name);
-                name[len - 1] = 0;
-                match = strstr(restart_string, name);
-            }
-
-            if (match == NULL) {
-                crm_trace("%s not found in %s", prop_name, restart_string);
-                xml_remove_prop(param_set, prop_name);
-            }
-            free(name);
-        }
-    }
-}
-
 extern bool crm_is_daemon;
 
 /* coverity[+kill] */
@@ -1199,157 +1184,34 @@ crm_abort(const char *file, const char *function, int line,
     crm_perror(LOG_ERR, "Cannot wait on forked child %d", pid);
 }
 
-char *
-generate_series_filename(const char *directory, const char *series, int sequence, gboolean bzip)
-{
-    int len = 40;
-    char *filename = NULL;
-    const char *ext = "raw";
-
-    CRM_CHECK(directory != NULL, return NULL);
-    CRM_CHECK(series != NULL, return NULL);
-
-#if !HAVE_BZLIB_H
-    bzip = FALSE;
-#endif
-
-    len += strlen(directory);
-    len += strlen(series);
-    filename = malloc(len);
-    CRM_CHECK(filename != NULL, return NULL);
-
-    if (bzip) {
-        ext = "bz2";
-    }
-    sprintf(filename, "%s/%s-%d.%s", directory, series, sequence, ext);
-
-    return filename;
-}
-
 int
-get_last_sequence(const char *directory, const char *series)
+crm_pid_active(long pid, const char *daemon)
 {
-    FILE *file_strm = NULL;
-    int start = 0, length = 0, read_len = 0;
-    char *series_file = NULL;
-    char *buffer = NULL;
-    int seq = 0;
-    int len = 36;
+    static int have_proc_pid = 0;
 
-    CRM_CHECK(directory != NULL, return 0);
-    CRM_CHECK(series != NULL, return 0);
+    if(have_proc_pid == 0) {
+        char proc_path[PATH_MAX], exe_path[PATH_MAX];
 
-    len += strlen(directory);
-    len += strlen(series);
-    series_file = malloc(len);
-    CRM_CHECK(series_file != NULL, return 0);
-    sprintf(series_file, "%s/%s.last", directory, series);
+        /* check to make sure pid hasn't been reused by another process */
+        snprintf(proc_path, sizeof(proc_path), "/proc/%lu/exe", (long unsigned int)getpid());
 
-    file_strm = fopen(series_file, "r");
-    if (file_strm == NULL) {
-        crm_debug("Series file %s does not exist", series_file);
-        free(series_file);
-        return 0;
-    }
-
-    /* see how big the file is */
-    start = ftell(file_strm);
-    fseek(file_strm, 0L, SEEK_END);
-    length = ftell(file_strm);
-    fseek(file_strm, 0L, start);
-
-    CRM_ASSERT(length >= 0);
-    CRM_ASSERT(start == ftell(file_strm));
-
-    if (length <= 0) {
-        crm_info("%s was not valid", series_file);
-        free(buffer);
-        buffer = NULL;
-
-    } else {
-        crm_trace("Reading %d bytes from file", length);
-        buffer = calloc(1, (length + 1));
-        read_len = fread(buffer, 1, length, file_strm);
-        if (read_len != length) {
-            crm_err("Calculated and read bytes differ: %d vs. %d", length, read_len);
-            free(buffer);
-            buffer = NULL;
+        have_proc_pid = 1;
+        if(readlink(proc_path, exe_path, PATH_MAX - 1) < 0) {
+            have_proc_pid = -1;
         }
     }
 
-    seq = crm_parse_int(buffer, "0");
-    fclose(file_strm);
-
-    crm_trace("Found %d in %s", seq, series_file);
-
-    free(series_file);
-    free(buffer);
-    return seq;
-}
-
-void
-write_last_sequence(const char *directory, const char *series, int sequence, int max)
-{
-    int rc = 0;
-    int len = 36;
-    FILE *file_strm = NULL;
-    char *series_file = NULL;
-
-    CRM_CHECK(directory != NULL, return);
-    CRM_CHECK(series != NULL, return);
-
-    if (max == 0) {
-        return;
-    }
-    if (max > 0 && sequence >= max) {
-        sequence = 0;
-    }
-
-    len += strlen(directory);
-    len += strlen(series);
-    series_file = malloc(len);
-
-    if(series_file) {
-        sprintf(series_file, "%s/%s.last", directory, series);
-        file_strm = fopen(series_file, "w");
-    }
-
-    if (file_strm != NULL) {
-        rc = fprintf(file_strm, "%d", sequence);
-        if (rc < 0) {
-            crm_perror(LOG_ERR, "Cannot write to series file %s", series_file);
-        }
-
-    } else {
-        crm_err("Cannot open series file %s for writing", series_file);
-    }
-
-    if (file_strm != NULL) {
-        fflush(file_strm);
-        fclose(file_strm);
-    }
-
-    crm_trace("Wrote %d to %s", sequence, series_file);
-    free(series_file);
-}
-
-#define	LOCKSTRLEN	11
-
-int
-crm_pid_active(long pid)
-{
     if (pid <= 0) {
         return -1;
 
     } else if (kill(pid, 0) < 0 && errno == ESRCH) {
         return 0;
-    }
-#ifndef HAVE_PROC_PID
-    return 1;
-#else
-    {
+
+    } else if(daemon == NULL || have_proc_pid == -1) {
+        return 1;
+
+    } else {
         int rc = 0;
-        int running = 0;
         char proc_path[PATH_MAX], exe_path[PATH_MAX], myexe_path[PATH_MAX];
 
         /* check to make sure pid hasn't been reused by another process */
@@ -1358,29 +1220,30 @@ crm_pid_active(long pid)
         rc = readlink(proc_path, exe_path, PATH_MAX - 1);
         if (rc < 0) {
             crm_perror(LOG_ERR, "Could not read from %s", proc_path);
-            goto bail;
+            return 0;
         }
 
         exe_path[rc] = 0;
-        snprintf(proc_path, sizeof(proc_path), "/proc/%lu/exe", (long unsigned int)getpid());
-        rc = readlink(proc_path, myexe_path, PATH_MAX - 1);
-        if (rc < 0) {
-            crm_perror(LOG_ERR, "Could not read from %s", proc_path);
-            goto bail;
-        }
 
-        myexe_path[rc] = 0;
+        if(daemon[0] != '/') {
+            rc = snprintf(myexe_path, sizeof(proc_path), CRM_DAEMON_DIR"/%s", daemon);
+            myexe_path[rc] = 0;
+        } else {
+            rc = snprintf(myexe_path, sizeof(proc_path), "%s", daemon);
+            myexe_path[rc] = 0;
+        }
+        
         if (strcmp(exe_path, myexe_path) == 0) {
-            running = 1;
+            return 1;
         }
     }
 
-  bail:
-    return running;
-#endif
+    return 0;
 }
 
-static int
+#define	LOCKSTRLEN	11
+
+int
 crm_read_pidfile(const char *filename)
 {
     int fd;
@@ -1408,8 +1271,8 @@ crm_read_pidfile(const char *filename)
     return pid;
 }
 
-static int
-crm_pidfile_inuse(const char *filename, long mypid)
+int
+crm_pidfile_inuse(const char *filename, long mypid, const char *daemon)
 {
     long pid = 0;
     struct stat sbuf;
@@ -1434,7 +1297,7 @@ crm_pidfile_inuse(const char *filename, long mypid)
                     /* In use by us */
                     rc = pcmk_ok;
 
-                } else if (crm_pid_active(pid) == FALSE) {
+                } else if (crm_pid_active(pid, daemon) == FALSE) {
                     /* Contains a stale value */
                     unlink(filename);
                     rc = -ENOENT;
@@ -1451,7 +1314,7 @@ crm_pidfile_inuse(const char *filename, long mypid)
 }
 
 static int
-crm_lock_pidfile(const char *filename)
+crm_lock_pidfile(const char *filename, const char *name)
 {
     long mypid = 0;
     int fd = 0, rc = 0;
@@ -1459,7 +1322,7 @@ crm_lock_pidfile(const char *filename)
 
     mypid = (unsigned long)getpid();
 
-    rc = crm_pidfile_inuse(filename, 0);
+    rc = crm_pidfile_inuse(filename, 0, name);
     if (rc == -ENOENT) {
         /* exists but the process is not active */
 
@@ -1482,7 +1345,7 @@ crm_lock_pidfile(const char *filename)
         return -errno;
     }
 
-    return crm_pidfile_inuse(filename, mypid);
+    return crm_pidfile_inuse(filename, mypid, name);
 }
 
 void
@@ -1497,7 +1360,7 @@ crm_make_daemon(const char *name, gboolean daemonize, const char *pidfile)
     }
 
     /* Check before we even try... */
-    rc = crm_pidfile_inuse(pidfile, 1);
+    rc = crm_pidfile_inuse(pidfile, 1, name);
     if(rc < pcmk_ok && rc != -ENOENT) {
         pid = crm_read_pidfile(pidfile);
         crm_err("%s: already running [pid %ld in %s]", name, pid, pidfile);
@@ -1515,7 +1378,7 @@ crm_make_daemon(const char *name, gboolean daemonize, const char *pidfile)
         crm_exit(pcmk_ok);
     }
 
-    rc = crm_lock_pidfile(pidfile);
+    rc = crm_lock_pidfile(pidfile, name);
     if(rc < pcmk_ok) {
         crm_err("Could not lock '%s' for %s: %s (%d)", pidfile, name, pcmk_strerror(rc), rc);
         printf("Could not lock '%s' for %s: %s (%d)\n", pidfile, name, pcmk_strerror(rc), rc);
@@ -1532,78 +1395,6 @@ crm_make_daemon(const char *name, gboolean daemonize, const char *pidfile)
     (void)open(devnull, O_WRONLY);      /* Stderr: fd 2 */
 }
 
-gboolean
-crm_is_writable(const char *dir, const char *file,
-                const char *user, const char *group, gboolean need_both)
-{
-    int s_res = -1;
-    struct stat buf;
-    char *full_file = NULL;
-    const char *target = NULL;
-
-    gboolean pass = TRUE;
-    gboolean readwritable = FALSE;
-
-    CRM_ASSERT(dir != NULL);
-    if (file != NULL) {
-        full_file = crm_concat(dir, file, '/');
-        target = full_file;
-        s_res = stat(full_file, &buf);
-        if (s_res == 0 && S_ISREG(buf.st_mode) == FALSE) {
-            crm_err("%s must be a regular file", target);
-            pass = FALSE;
-            goto out;
-        }
-    }
-
-    if (s_res != 0) {
-        target = dir;
-        s_res = stat(dir, &buf);
-        if (s_res != 0) {
-            crm_err("%s must exist and be a directory", dir);
-            pass = FALSE;
-            goto out;
-
-        } else if (S_ISDIR(buf.st_mode) == FALSE) {
-            crm_err("%s must be a directory", dir);
-            pass = FALSE;
-        }
-    }
-
-    if (user) {
-        struct passwd *sys_user = NULL;
-
-        sys_user = getpwnam(user);
-        readwritable = (sys_user != NULL
-                        && buf.st_uid == sys_user->pw_uid && (buf.st_mode & (S_IRUSR | S_IWUSR)));
-        if (readwritable == FALSE) {
-            crm_err("%s must be owned and r/w by user %s", target, user);
-            if (need_both) {
-                pass = FALSE;
-            }
-        }
-    }
-
-    if (group) {
-        struct group *sys_grp = getgrnam(group);
-
-        readwritable = (sys_grp != NULL
-                        && buf.st_gid == sys_grp->gr_gid && (buf.st_mode & (S_IRGRP | S_IWGRP)));
-        if (readwritable == FALSE) {
-            if (need_both || user == NULL) {
-                pass = FALSE;
-                crm_err("%s must be owned and r/w by group %s", target, group);
-            } else {
-                crm_warn("%s should be owned and r/w by group %s", target, group);
-            }
-        }
-    }
-
-  out:
-    free(full_file);
-    return pass;
-}
-
 char *
 crm_strip_trailing_newline(char *str)
 {
@@ -1626,7 +1417,7 @@ crm_str_eq(const char *a, const char *b, gboolean use_case)
     if (use_case) {
         return g_strcmp0(a, b) == 0;
 
-        /* TODO - Figure out which calls, if any, really need to be case independant */
+        /* TODO - Figure out which calls, if any, really need to be case independent */
     } else if (a == b) {
         return TRUE;
 
@@ -1692,7 +1483,7 @@ crm_create_long_opts(struct crm_option *long_options)
      * This dummy entry allows us to differentiate between the two in crm_get_option()
      * and exit with the correct error code
      */
-    long_opts = realloc(long_opts, (index + 1) * sizeof(struct option));
+    long_opts = realloc_safe(long_opts, (index + 1) * sizeof(struct option));
     long_opts[index].name = "__dummmy__";
     long_opts[index].has_arg = 0;
     long_opts[index].flag = 0;
@@ -1704,7 +1495,7 @@ crm_create_long_opts(struct crm_option *long_options)
             continue;
         }
 
-        long_opts = realloc(long_opts, (index + 1) * sizeof(struct option));
+        long_opts = realloc_safe(long_opts, (index + 1) * sizeof(struct option));
         /*fprintf(stderr, "Creating %d %s = %c\n", index,
          * long_options[lpc].name, long_options[lpc].val);      */
         long_opts[index].name = long_options[lpc].name;
@@ -1715,7 +1506,7 @@ crm_create_long_opts(struct crm_option *long_options)
     }
 
     /* Now create the list terminator */
-    long_opts = realloc(long_opts, (index + 1) * sizeof(struct option));
+    long_opts = realloc_safe(long_opts, (index + 1) * sizeof(struct option));
     long_opts[index].name = NULL;
     long_opts[index].has_arg = 0;
     long_opts[index].flag = 0;
@@ -1739,7 +1530,7 @@ crm_set_options(const char *short_options, const char *app_usage, struct crm_opt
 
         for (lpc = 0; long_options[lpc].name != NULL; lpc++) {
             if (long_options[lpc].val && long_options[lpc].val != '-' && long_options[lpc].val < UCHAR_MAX) {
-                local_short_options = realloc(local_short_options, opt_string_len + 4);
+                local_short_options = realloc_safe(local_short_options, opt_string_len + 4);
                 local_short_options[opt_string_len++] = long_options[lpc].val;
                 /* getopt(3) says: Two colons mean an option takes an optional arg; */
                 if (long_options[lpc].has_arg == optional_argument) {
@@ -1782,6 +1573,7 @@ crm_get_option_long(int argc, char **argv, int *index, const char **longname)
         long_opts = crm_create_long_opts(crm_long_options);
     }
 
+    *index = 0;
     if (long_opts) {
         int flag = getopt_long(argc, argv, crm_short_options, long_opts, index);
 
@@ -1823,13 +1615,13 @@ crm_help(char cmd, int exit_code)
     FILE *stream = (exit_code ? stderr : stdout);
 
     if (cmd == 'v' || cmd == '$') {
-        fprintf(stream, "Pacemaker %s\n", VERSION);
+        fprintf(stream, "Pacemaker %s\n", PACEMAKER_VERSION);
         fprintf(stream, "Written by Andrew Beekhof\n");
         goto out;
     }
 
     if (cmd == '!') {
-        fprintf(stream, "Pacemaker %s (Build: %s): %s\n", VERSION, BUILD_VERSION, CRM_FEATURES);
+        fprintf(stream, "Pacemaker %s (Build: %s): %s\n", PACEMAKER_VERSION, BUILD_VERSION, CRM_FEATURES);
         goto out;
     }
 
@@ -1951,15 +1743,15 @@ stonith_ipc_server_init(qb_ipcs_service_t **ipcs, struct qb_ipcs_service_handler
 int
 attrd_update_delegate(crm_ipc_t * ipc, char command, const char *host, const char *name,
                       const char *value, const char *section, const char *set, const char *dampen,
-                      const char *user_name, gboolean is_remote)
+                      const char *user_name, int options)
 {
     int rc = -ENOTCONN;
     int max = 5;
-    enum crm_ipc_flags flags = crm_ipc_flags_none;
     xmlNode *update = create_xml_node(NULL, __FUNCTION__);
 
     static gboolean connected = TRUE;
     static crm_ipc_t *local_ipc = NULL;
+    static enum crm_ipc_flags flags = crm_ipc_flags_none;
 
     if (ipc == NULL && local_ipc == NULL) {
         local_ipc = crm_ipc_new(T_ATTRD, 0);
@@ -1987,20 +1779,25 @@ attrd_update_delegate(crm_ipc_t * ipc, char command, const char *host, const cha
     }
 
     switch (command) {
+        case 'u':
+            crm_xml_add(update, F_ATTRD_TASK, ATTRD_OP_UPDATE);
+            crm_xml_add(update, F_ATTRD_REGEX, name);
+            break;
         case 'D':
         case 'U':
         case 'v':
-            crm_xml_add(update, F_ATTRD_TASK, "update");
+            crm_xml_add(update, F_ATTRD_TASK, ATTRD_OP_UPDATE);
             crm_xml_add(update, F_ATTRD_ATTRIBUTE, name);
             break;
         case 'R':
-            crm_xml_add(update, F_ATTRD_TASK, "refresh");
+            crm_xml_add(update, F_ATTRD_TASK, ATTRD_OP_REFRESH);
             break;
-        case 'q':
-            crm_xml_add(update, F_ATTRD_TASK, "query");
+        case 'Q':
+            crm_xml_add(update, F_ATTRD_TASK, ATTRD_OP_QUERY);
+            crm_xml_add(update, F_ATTRD_ATTRIBUTE, name);
             break;
         case 'C':
-            crm_xml_add(update, F_ATTRD_TASK, "peer-remove");
+            crm_xml_add(update, F_ATTRD_TASK, ATTRD_OP_PEER_REMOVE);
             break;
     }
 
@@ -2009,7 +1806,8 @@ attrd_update_delegate(crm_ipc_t * ipc, char command, const char *host, const cha
     crm_xml_add(update, F_ATTRD_SECTION, section);
     crm_xml_add(update, F_ATTRD_HOST, host);
     crm_xml_add(update, F_ATTRD_SET, set);
-    crm_xml_add_int(update, F_ATTRD_IS_REMOTE, is_remote);
+    crm_xml_add_int(update, F_ATTRD_IS_REMOTE, is_set(options, attrd_opt_remote));
+    crm_xml_add_int(update, F_ATTRD_IS_PRIVATE, is_set(options, attrd_opt_private));
 #if ENABLE_ACL
     if (user_name) {
         crm_xml_add(update, F_ATTRD_USER, user_name);
@@ -2024,6 +1822,8 @@ attrd_update_delegate(crm_ipc_t * ipc, char command, const char *host, const cha
 
         if (connected) {
             rc = crm_ipc_send(ipc, update, flags, 0, NULL);
+        } else {
+            crm_perror(LOG_INFO, "Connection to cluster attribute manager failed");
         }
 
         if (ipc != local_ipc) {
@@ -2133,14 +1933,15 @@ did_rsc_op_fail(lrmd_event_data_t * op, int target_rc)
 }
 
 xmlNode *
-create_operation_update(xmlNode * parent, lrmd_event_data_t * op, const char *caller_version,
-                        int target_rc, const char *origin, int level)
+create_operation_update(xmlNode * parent, lrmd_event_data_t * op, const char * caller_version,
+                        int target_rc, const char * node, const char * origin, int level)
 {
     char *key = NULL;
     char *magic = NULL;
     char *op_id = NULL;
     char *op_id_additional = NULL;
     char *local_user_data = NULL;
+    const char *exit_reason = NULL;
 
     xmlNode *xml_op = NULL;
     const char *task = NULL;
@@ -2156,7 +1957,7 @@ create_operation_update(xmlNode * parent, lrmd_event_data_t * op, const char *ca
 
     task = op->op_type;
     /* remap the task name under various scenarios
-     * this makes life easier for the PE when its trying determin the current state
+     * this makes life easier for the PE when trying determine the current state
      */
     if (crm_str_eq(task, "reload", TRUE)) {
         if (op->op_status == PCMK_LRM_OP_DONE) {
@@ -2201,6 +2002,7 @@ create_operation_update(xmlNode * parent, lrmd_event_data_t * op, const char *ca
             /* Ensure 'last' gets updated too in case recording-pending="true" */
             op_id_additional = generate_op_key(op->rsc_id, "last", 0);
         }
+        exit_reason = op->exit_reason;
 
     } else if (op->interval > 0) {
         op_id = strdup(key);
@@ -2234,6 +2036,8 @@ create_operation_update(xmlNode * parent, lrmd_event_data_t * op, const char *ca
     crm_xml_add(xml_op, XML_ATTR_CRM_VERSION, caller_version);
     crm_xml_add(xml_op, XML_ATTR_TRANSITION_KEY, op->user_data);
     crm_xml_add(xml_op, XML_ATTR_TRANSITION_MAGIC, magic);
+    crm_xml_add(xml_op, XML_LRM_ATTR_EXIT_REASON, exit_reason);
+    crm_xml_add(xml_op, XML_LRM_ATTR_TARGET, node); /* For context during triage */
 
     crm_xml_add_int(xml_op, XML_LRM_ATTR_CALLID, op->call_id);
     crm_xml_add_int(xml_op, XML_LRM_ATTR_RC, op->rc);
@@ -2490,7 +2294,7 @@ add_list_element(char *list, const char *value)
     }
     len = last + 2;             /* +1 space, +1 EOS */
     len += strlen(value);
-    list = realloc(list, len);
+    list = realloc_safe(list, len);
     sprintf(list + last, " %s", value);
     return list;
 }
@@ -2526,9 +2330,10 @@ crm_md5sum(const char *buffer)
     char *digest = NULL;
     unsigned char raw_digest[MD5_DIGEST_SIZE];
 
-    if(buffer != NULL) {
-        len = strlen(buffer);
+    if (buffer == NULL) {
+        buffer = "";
     }
+    len = strlen(buffer);
 
     crm_trace("Beginning digest of %d bytes", len);
     digest = malloc(2 * MD5_DIGEST_SIZE + 1);
