@@ -61,6 +61,7 @@ struct stonith_action_s {
 
     /*! internal async track data */
     int fd_stdout;
+    int fd_stderr;
     int last_timeout_signo;
 
     /*! internal timing information */
@@ -75,6 +76,7 @@ struct stonith_action_s {
     GPid pid;
     int rc;
     char *output;
+    char *error;
 };
 
 typedef struct stonith_private_s {
@@ -121,6 +123,7 @@ struct timer_rec_s {
 typedef int (*stonith_op_t) (const char *, int, const char *, xmlNode *,
                              xmlNode *, xmlNode *, xmlNode **, xmlNode **);
 
+#if HAVE_STONITH_STONITH_H
 static const char META_TEMPLATE[] =
     "<?xml version=\"1.0\"?>\n"
     "<!DOCTYPE resource-agent SYSTEM \"ra-api-1.dtd\">\n"
@@ -140,6 +143,7 @@ static const char META_TEMPLATE[] =
     "  </actions>\n"
     "  <special tag=\"heartbeat\">\n"
     "    <version>2.0</version>\n" "  </special>\n" "</resource-agent>\n";
+#endif
 
 bool stonith_dispatch(stonith_t * st);
 int stonith_dispatch_internal(const char *buffer, ssize_t length, gpointer userdata);
@@ -152,6 +156,27 @@ int stonith_send_command(stonith_t * stonith, const char *op, xmlNode * data,
 static void stonith_connection_destroy(gpointer user_data);
 static void stonith_send_notification(gpointer data, gpointer user_data);
 static int internal_stonith_action_execute(stonith_action_t * action);
+static void log_action(stonith_action_t *action, pid_t pid);
+
+static void
+log_action(stonith_action_t *action, pid_t pid)
+{
+    if (action->output) {
+        /* Logging the whole string confuses syslog when the string is xml */
+        char *prefix = crm_strdup_printf("%s[%d] stdout:", action->agent, pid);
+
+        crm_log_output(LOG_TRACE, prefix, action->output);
+        free(prefix);
+    }
+
+    if (action->error) {
+        /* Logging the whole string confuses syslog when the string is xml */
+        char *prefix = crm_strdup_printf("%s[%d] stderr:", action->agent, pid);
+
+        crm_log_output(LOG_WARNING, prefix, action->error);
+        free(prefix);
+    }
+}
 
 static void
 stonith_connection_destroy(gpointer user_data)
@@ -192,7 +217,7 @@ create_device_registration_xml(const char *id, const char *namespace, const char
 #endif
 
     crm_xml_add(data, XML_ATTR_ID, id);
-    crm_xml_add(data, "origin", __FUNCTION__);
+    crm_xml_add(data, F_STONITH_ORIGIN, __FUNCTION__);
     crm_xml_add(data, "agent", agent);
     crm_xml_add(data, "namespace", namespace);
     if (rsc_provides) {
@@ -229,7 +254,7 @@ stonith_api_remove_device(stonith_t * st, int call_options, const char *name)
     xmlNode *data = NULL;
 
     data = create_xml_node(NULL, F_STONITH_DEVICE);
-    crm_xml_add(data, "origin", __FUNCTION__);
+    crm_xml_add(data, F_STONITH_ORIGIN, __FUNCTION__);
     crm_xml_add(data, XML_ATTR_ID, name);
     rc = stonith_send_command(st, STONITH_OP_DEVICE_DEL, data, NULL, call_options, 0);
     free_xml(data);
@@ -238,50 +263,130 @@ stonith_api_remove_device(stonith_t * st, int call_options, const char *name)
 }
 
 static int
-stonith_api_remove_level(stonith_t * st, int options, const char *node, int level)
+stonith_api_remove_level_full(stonith_t *st, int options,
+                              const char *node, const char *pattern,
+                              const char *attr, const char *value, int level)
 {
     int rc = 0;
     xmlNode *data = NULL;
 
-    data = create_xml_node(NULL, F_STONITH_LEVEL);
-    crm_xml_add(data, "origin", __FUNCTION__);
-    crm_xml_add(data, F_STONITH_TARGET, node);
-    crm_xml_add_int(data, XML_ATTR_ID, level);
+    CRM_CHECK(node || pattern || (attr && value), return -EINVAL);
+
+    data = create_xml_node(NULL, XML_TAG_FENCING_LEVEL);
+    crm_xml_add(data, F_STONITH_ORIGIN, __FUNCTION__);
+
+    if (node) {
+        crm_xml_add(data, XML_ATTR_STONITH_TARGET, node);
+
+    } else if (pattern) {
+        crm_xml_add(data, XML_ATTR_STONITH_TARGET_PATTERN, pattern);
+
+    } else {
+        crm_xml_add(data, XML_ATTR_STONITH_TARGET_ATTRIBUTE, attr);
+        crm_xml_add(data, XML_ATTR_STONITH_TARGET_VALUE, value);
+    }
+
+    crm_xml_add_int(data, XML_ATTR_STONITH_INDEX, level);
     rc = stonith_send_command(st, STONITH_OP_LEVEL_DEL, data, NULL, options, 0);
     free_xml(data);
 
     return rc;
 }
 
-xmlNode *
-create_level_registration_xml(const char *node, int level, stonith_key_value_t * device_list)
+static int
+stonith_api_remove_level(stonith_t * st, int options, const char *node, int level)
 {
-    xmlNode *data = create_xml_node(NULL, F_STONITH_LEVEL);
+    return stonith_api_remove_level_full(st, options, node,
+                                         NULL, NULL, NULL, level);
+}
 
+/*
+ * \internal
+ * \brief Create XML for stonithd topology level registration request
+ *
+ * \param[in] node        If not NULL, target level by this node name
+ * \param[in] pattern     If not NULL, target by node name using this regex
+ * \param[in] attr        If not NULL, target by this node attribute
+ * \param[in] value       If not NULL, target by this node attribute value
+ * \param[in] level       Index number of level to register
+ * \param[in] device_list List of devices in level
+ *
+ * \return Newly allocated XML tree on success, NULL otherwise
+ *
+ * \note The caller should set only one of node, pattern or attr/value.
+ */
+xmlNode *
+create_level_registration_xml(const char *node, const char *pattern,
+                              const char *attr, const char *value,
+                              int level, stonith_key_value_t *device_list)
+{
+    int len = 0;
+    char *list = NULL;
+    xmlNode *data;
+
+    CRM_CHECK(node || pattern || (attr && value), return NULL);
+
+    data = create_xml_node(NULL, XML_TAG_FENCING_LEVEL);
+    CRM_CHECK(data, return NULL);
+
+    crm_xml_add(data, F_STONITH_ORIGIN, __FUNCTION__);
     crm_xml_add_int(data, XML_ATTR_ID, level);
-    crm_xml_add(data, F_STONITH_TARGET, node);
-    crm_xml_add(data, "origin", __FUNCTION__);
+    crm_xml_add_int(data, XML_ATTR_STONITH_INDEX, level);
 
-    for (; device_list; device_list = device_list->next) {
-        xmlNode *dev = create_xml_node(data, F_STONITH_DEVICE);
+    if (node) {
+        crm_xml_add(data, XML_ATTR_STONITH_TARGET, node);
 
-        crm_xml_add(dev, XML_ATTR_ID, device_list->value);
+    } else if (pattern) {
+        crm_xml_add(data, XML_ATTR_STONITH_TARGET_PATTERN, pattern);
+
+    } else {
+        crm_xml_add(data, XML_ATTR_STONITH_TARGET_ATTRIBUTE, attr);
+        crm_xml_add(data, XML_ATTR_STONITH_TARGET_VALUE, value);
     }
 
+    for (; device_list; device_list = device_list->next) {
+
+        int adding = strlen(device_list->value);
+        if(list) {
+            adding++;                                      /* +1 space */
+        }
+
+        crm_trace("Adding %s (%dc) at offset %d", device_list->value, adding, len);
+        list = realloc_safe(list, len + adding + 1);       /* +1 EOS */
+        CRM_CHECK(list != NULL, free_xml(data); return NULL);
+        sprintf(list + len, "%s%s", len?",":"", device_list->value);
+        len += adding;
+    }
+
+    crm_xml_add(data, XML_ATTR_STONITH_DEVICES, list);
+
+    free(list);
     return data;
 }
 
+static int
+stonith_api_register_level_full(stonith_t * st, int options, const char *node,
+                                const char *pattern,
+                                const char *attr, const char *value,
+                                int level, stonith_key_value_t *device_list)
+{
+    int rc = 0;
+    xmlNode *data = create_level_registration_xml(node, pattern, attr, value,
+                                                  level, device_list);
+    CRM_CHECK(data != NULL, return -EINVAL);
+
+    rc = stonith_send_command(st, STONITH_OP_LEVEL_ADD, data, NULL, options, 0);
+    free_xml(data);
+
+    return rc;
+}
+
 static int
 stonith_api_register_level(stonith_t * st, int options, const char *node, int level,
                            stonith_key_value_t * device_list)
 {
-    int rc = 0;
-    xmlNode *data = create_level_registration_xml(node, level, device_list);
-
-    rc = stonith_send_command(st, STONITH_OP_LEVEL_ADD, data, NULL, options, 0);
-    free_xml(data);
-
-    return rc;
+    return stonith_api_register_level_full(st, options, node, NULL, NULL, NULL,
+                                           level, device_list);
 }
 
 static void
@@ -308,7 +413,7 @@ append_arg(gpointer key, gpointer value, gpointer user_data)
         last = strlen(*args);
     }
 
-    *args = realloc(*args, last + len);
+    *args = realloc_safe(*args, last + len);
     crm_trace("Appending: %s=%s", (char *)key, (char *)value);
     sprintf((*args) + last, "%s=%s\n", (char *)key, (char *)value);
 }
@@ -401,7 +506,7 @@ append_host_specific_args(const char *victim, const char *map, GHashTable * para
 }
 
 static char *
-make_args(const char *action, const char *victim, uint32_t victim_nodeid, GHashTable * device_args,
+make_args(const char *agent, const char *action, const char *victim, uint32_t victim_nodeid, GHashTable * device_args,
           GHashTable * port_map)
 {
     char buffer[512];
@@ -456,7 +561,10 @@ make_args(const char *action, const char *victim, uint32_t victim_nodeid, GHashT
         }
 
         /* Check if we need to supply the victim in any other form */
-        if (param == NULL) {
+        if(safe_str_eq(agent, "fence_legacy")) {
+            value = agent;
+
+        } else if (param == NULL) {
             const char *map = g_hash_table_lookup(device_args, STONITH_ATTR_ARGMAP);
 
             if (map == NULL) {
@@ -550,8 +658,14 @@ stonith_action_clear_tracking_data(stonith_action_t * action)
         close(action->fd_stdout);
         action->fd_stdout = 0;
     }
+    if (action->fd_stderr) {
+        close(action->fd_stderr);
+        action->fd_stderr = 0;
+    }
     free(action->output);
     action->output = NULL;
+    free(action->error);
+    action->error = NULL;
     action->rc = 0;
     action->pid = 0;
     action->last_timeout_signo = 0;
@@ -580,7 +694,7 @@ stonith_action_create(const char *agent,
 
     action = calloc(1, sizeof(stonith_action_t));
     crm_debug("Initiating action %s for agent %s (target=%s)", _action, agent, victim);
-    action->args = make_args(_action, victim, victim_nodeid, device_args, port_map);
+    action->args = make_args(agent, _action, victim, victim_nodeid, device_args, port_map);
     action->agent = strdup(agent);
     action->action = strdup(_action);
     if (victim) {
@@ -626,8 +740,7 @@ read_output(int fd)
             buffer[more] = 0; /* Make sure its nul-terminated for logging
                               * 'more' is always less than our buffer size
                               */
-            crm_trace("Got %d more bytes: %.200s...", more, buffer);
-            output = realloc(output, len + more + 1);
+            output = realloc_safe(output, len + more + 1);
             snprintf(output + len, more + 1, "%s", buffer);
             len += more;
         }
@@ -663,11 +776,16 @@ stonith_action_async_done(mainloop_child_t * p, pid_t pid, int core, int signo,
 
     if (action->timer_sigterm > 0) {
         g_source_remove(action->timer_sigterm);
+        action->timer_sigterm = 0;
     }
     if (action->timer_sigkill > 0) {
         g_source_remove(action->timer_sigkill);
+        action->timer_sigkill = 0;
     }
 
+    action->output = read_output(action->fd_stdout);
+    action->error = read_output(action->fd_stderr);
+
     if (action->last_timeout_signo) {
         action->rc = -ETIME;
         crm_notice("Child process %d performing action '%s' timed out with signal %d",
@@ -679,12 +797,30 @@ stonith_action_async_done(mainloop_child_t * p, pid_t pid, int core, int signo,
                    pid, action->action, signo);
 
     } else {
-        action->rc = exitcode;
         crm_debug("Child process %d performing action '%s' exited with rc %d",
                   pid, action->action, exitcode);
+        if (exitcode > 0) {
+            /* Try to provide a useful error code based on the fence agent's
+             * error output.
+             */
+            if (action->error == NULL) {
+                exitcode = -ENODATA;
+
+            } else if (strstr(action->error, "imed out")) {
+                /* Some agents have their own internal timeouts */
+                exitcode = -ETIMEDOUT;
+
+            } else if (strstr(action->error, "Unrecognised action")) {
+                exitcode = -EOPNOTSUPP;
+
+            } else {
+                exitcode = -pcmk_err_generic;
+            }
+        }
+        action->rc = exitcode;
     }
 
-    action->output = read_output(action->fd_stdout);
+    log_action(action, pid);
 
     if (action->rc != pcmk_ok && update_remaining_timeout(action)) {
         int rc = internal_stonith_action_execute(action);
@@ -708,8 +844,10 @@ internal_stonith_action_execute(stonith_action_t * action)
     int total = 0;
     int p_read_fd, p_write_fd;  /* parent read/write file descriptors */
     int c_read_fd, c_write_fd;  /* child read/write file descriptors */
+    int c_stderr_fd, p_stderr_fd; /* parent/child side file descriptors for stderr */
     int fd1[2];
     int fd2[2];
+    int fd3[2];
     int is_retry = 0;
 
     /* clear any previous tracking data */
@@ -726,7 +864,7 @@ internal_stonith_action_execute(stonith_action_t * action)
         is_retry = 1;
     }
 
-    c_read_fd = c_write_fd = p_read_fd = p_write_fd = -1;
+    c_read_fd = c_write_fd = p_read_fd = p_write_fd = c_stderr_fd = p_stderr_fd = -1;
 
     if (action->args == NULL || action->agent == NULL)
         goto fail;
@@ -742,6 +880,11 @@ internal_stonith_action_execute(stonith_action_t * action)
     c_read_fd = fd2[0];
     p_write_fd = fd2[1];
 
+    if (pipe(fd3))
+        goto fail;
+    p_stderr_fd = fd3[0];
+    c_stderr_fd = fd3[1];
+
     crm_debug("forking");
     pid = fork();
     if (pid < 0) {
@@ -759,17 +902,19 @@ internal_stonith_action_execute(stonith_action_t * action)
             goto fail;
         close(2);
         /* coverity[leaked_handle] False positive */
-        if (dup(c_write_fd) < 0)
+        if (dup(c_stderr_fd) < 0)
             goto fail;
         close(0);
         /* coverity[leaked_handle] False positive */
         if (dup(c_read_fd) < 0)
             goto fail;
 
-        /* keep c_write_fd open so parent can report all errors. */
+        /* keep c_stderr_fd open so parent can report all errors. */
+        /* keep c_write_fd open so hostlist can be sent to parent. */
         close(c_read_fd);
         close(p_read_fd);
         close(p_write_fd);
+        close(p_stderr_fd);
 
         /* keep retries from executing out of control */
         if (is_retry) {
@@ -786,6 +931,11 @@ internal_stonith_action_execute(stonith_action_t * action)
         crm_perror(LOG_NOTICE, "Could not change the output of %s to be non-blocking",
                    action->agent);
     }
+    ret = fcntl(p_stderr_fd, F_SETFL, fcntl(p_stderr_fd, F_GETFL, 0) | O_NONBLOCK);
+    if (ret < 0) {
+        crm_perror(LOG_NOTICE, "Could not change the stderr of %s to be non-blocking",
+                   action->agent);
+    }
 
     do {
         crm_debug("sending args");
@@ -809,6 +959,7 @@ internal_stonith_action_execute(stonith_action_t * action)
     /* async */
     if (action->async) {
         action->fd_stdout = p_read_fd;
+        action->fd_stderr = p_stderr_fd;
         mainloop_child_add(pid, 0/* Move the timeout here? */, action->action, action, stonith_action_async_done);
         crm_trace("Op: %s on %s, pid: %d, timeout: %ds", action->action, action->agent, pid,
                   action->remaining_timeout);
@@ -825,6 +976,7 @@ internal_stonith_action_execute(stonith_action_t * action)
 
         close(c_write_fd);
         close(c_read_fd);
+        close(c_stderr_fd);
         return 0;
 
     } else {
@@ -864,8 +1016,12 @@ internal_stonith_action_execute(stonith_action_t * action)
         }
 
         action->output = read_output(p_read_fd);
+        action->error = read_output(p_stderr_fd);
 
         action->rc = -ECONNABORTED;
+
+        log_action(action, pid);
+
         rc = action->rc;
         if (timeout == 0) {
             action->rc = -ETIME;
@@ -892,6 +1048,9 @@ internal_stonith_action_execute(stonith_action_t * action)
     if (p_write_fd >= 0) {
         close(p_write_fd);
     }
+    if (p_stderr_fd >= 0) {
+        close(p_stderr_fd);
+    }
 
     if (c_read_fd >= 0) {
         close(c_read_fd);
@@ -899,6 +1058,9 @@ internal_stonith_action_execute(stonith_action_t * action)
     if (c_write_fd >= 0) {
         close(c_write_fd);
     }
+    if (c_stderr_fd >= 0) {
+        close(c_stderr_fd);
+    }
 
     return rc;
 }
@@ -1054,21 +1216,25 @@ strdup_null(const char *val)
 }
 
 static void
-stonith_plugin(int priority, const char *fmt, ...)
-G_GNUC_PRINTF(2, 3);
+stonith_plugin(int priority, const char *fmt, ...) __attribute__((__format__ (__printf__, 2, 3)));
 
 static void
-stonith_plugin(int priority, const char *fmt, ...)
+stonith_plugin(int priority, const char *format, ...)
 {
-    va_list args;
-    char *str;
     int err = errno;
 
-    va_start(args, fmt);
-    str = g_strdup_vprintf(fmt, args);
-    va_end(args);
-    do_crm_log_alias(priority, __FILE__, __func__, __LINE__, "%s", str);
-    g_free(str);
+    va_list ap;
+    int len = 0;
+    char *string = NULL;
+
+    va_start(ap, format);
+
+    len = vasprintf (&string, format, ap);
+    CRM_ASSERT(len > 0);
+
+    do_crm_log_alias(priority, __FILE__, __func__, __LINE__, "%s", string);
+
+    free(string);
     errno = err;
 }
 #endif
@@ -1091,57 +1257,62 @@ stonith_api_device_metadata(stonith_t * stonith, int call_options, const char *a
     if (safe_str_eq(provider, "redhat")) {
         stonith_action_t *action = stonith_action_create(agent, "metadata", NULL, 0, 5, NULL, NULL);
         int exec_rc = stonith_action_execute(action, &rc, &buffer);
+        xmlNode *xml = NULL;
+        xmlNode *actions = NULL;
+        xmlXPathObject *xpathObj = NULL;
 
         if (exec_rc < 0 || rc != 0 || buffer == NULL) {
+            crm_warn("Could not obtain metadata for %s", agent);
             crm_debug("Query failed: %d %d: %s", exec_rc, rc, crm_str(buffer));
             free(buffer);       /* Just in case */
             return -EINVAL;
+        }
 
-        } else {
-
-            xmlNode *xml = string2xml(buffer);
-            xmlNode *actions = NULL;
-            xmlXPathObject *xpathObj = NULL;
-
-            xpathObj = xpath_search(xml, "//actions");
-            if (numXpathResults(xpathObj) > 0) {
-                actions = getXpathResult(xpathObj, 0);
-            }
-
-            freeXpathObject(xpathObj);
-
-            /* Now fudge the metadata so that the start/stop actions appear */
-            xpathObj = xpath_search(xml, "//action[@name='stop']");
-            if (numXpathResults(xpathObj) <= 0) {
-                xmlNode *tmp = NULL;
-
-                tmp = create_xml_node(actions, "action");
-                crm_xml_add(tmp, "name", "stop");
-                crm_xml_add(tmp, "timeout", "20s");
-
-                tmp = create_xml_node(actions, "action");
-                crm_xml_add(tmp, "name", "start");
-                crm_xml_add(tmp, "timeout", "20s");
-            }
-
-            freeXpathObject(xpathObj);
-
-            /* Now fudge the metadata so that the port isn't required in the configuration */
-            xpathObj = xpath_search(xml, "//parameter[@name='port']");
-            if (numXpathResults(xpathObj) > 0) {
-                /* We'll fill this in */
-                xmlNode *tmp = getXpathResult(xpathObj, 0);
-
-                crm_xml_add(tmp, "required", "0");
-            }
-
-            freeXpathObject(xpathObj);
+        xml = string2xml(buffer);
+        if(xml == NULL) {
+            crm_warn("Metadata for %s is invalid", agent);
             free(buffer);
-            buffer = dump_xml_formatted(xml);
-            free_xml(xml);
-            if (!buffer) {
-                return -EINVAL;
-            }
+            return -EINVAL;
+        }
+
+        xpathObj = xpath_search(xml, "//actions");
+        if (numXpathResults(xpathObj) > 0) {
+            actions = getXpathResult(xpathObj, 0);
+        }
+
+        freeXpathObject(xpathObj);
+
+        /* Now fudge the metadata so that the start/stop actions appear */
+        xpathObj = xpath_search(xml, "//action[@name='stop']");
+        if (numXpathResults(xpathObj) <= 0) {
+            xmlNode *tmp = NULL;
+
+            tmp = create_xml_node(actions, "action");
+            crm_xml_add(tmp, "name", "stop");
+            crm_xml_add(tmp, "timeout", "20s");
+
+            tmp = create_xml_node(actions, "action");
+            crm_xml_add(tmp, "name", "start");
+            crm_xml_add(tmp, "timeout", "20s");
+        }
+
+        freeXpathObject(xpathObj);
+
+        /* Now fudge the metadata so that the port isn't required in the configuration */
+        xpathObj = xpath_search(xml, "//parameter[@name='port']");
+        if (numXpathResults(xpathObj) > 0) {
+            /* We'll fill this in */
+            xmlNode *tmp = getXpathResult(xpathObj, 0);
+
+            crm_xml_add(tmp, "required", "0");
+        }
+
+        freeXpathObject(xpathObj);
+        free(buffer);
+        buffer = dump_xml_formatted_with_text(xml);
+        free_xml(xml);
+        if (!buffer) {
+            return -EINVAL;
         }
 
     } else {
@@ -1253,7 +1424,7 @@ stonith_api_query(stonith_t * stonith, int call_options, const char *target,
     CRM_CHECK(devices != NULL, return -EINVAL);
 
     data = create_xml_node(NULL, F_STONITH_DEVICE);
-    crm_xml_add(data, "origin", __FUNCTION__);
+    crm_xml_add(data, F_STONITH_ORIGIN, __FUNCTION__);
     crm_xml_add(data, F_STONITH_TARGET, target);
     crm_xml_add(data, F_STONITH_ACTION, "off");
     rc = stonith_send_command(stonith, STONITH_OP_QUERY, data, &output, call_options, timeout);
@@ -1271,7 +1442,10 @@ stonith_api_query(stonith_t * stonith, int call_options, const char *target,
 
             CRM_LOG_ASSERT(match != NULL);
             if(match != NULL) {
-                crm_info("%s[%d] = %s", "//@agent", lpc, xmlGetNodePath(match));
+                xmlChar *match_path = xmlGetNodePath(match);
+
+                crm_info("%s[%d] = %s", "//@agent", lpc, match_path);
+                free(match_path);
                 *devices = stonith_key_value_add(*devices, NULL, crm_element_value(match, XML_ATTR_ID));
             }
         }
@@ -1294,7 +1468,7 @@ stonith_api_call(stonith_t * stonith,
     xmlNode *data = NULL;
 
     data = create_xml_node(NULL, F_STONITH_DEVICE);
-    crm_xml_add(data, "origin", __FUNCTION__);
+    crm_xml_add(data, F_STONITH_ORIGIN, __FUNCTION__);
     crm_xml_add(data, F_STONITH_DEVICE, id);
     crm_xml_add(data, F_STONITH_ACTION, action);
     crm_xml_add(data, F_STONITH_TARGET, victim);
@@ -1465,8 +1639,13 @@ get_stonith_provider(const char *agent, const char *provider)
 #endif
     }
 
-    crm_err("No such device: %s", agent);
-    return NULL;
+    if (safe_str_eq(provider, "internal")) {
+        return provider;
+
+    } else {
+        crm_err("No such device: %s", agent);
+        return NULL;
+    }
 }
 
 static gint
@@ -1580,8 +1759,8 @@ stonith_api_signon(stonith_t * stonith, const char *name, int *stonith_fd)
 
         if (native->ipc && crm_ipc_connect(native->ipc)) {
             *stonith_fd = crm_ipc_get_fd(native->ipc);
-
         } else if (native->ipc) {
+            crm_perror(LOG_ERR, "Connection to STONITH manager failed");
             rc = -ENOTCONN;
         }
 
@@ -1975,7 +2154,7 @@ xml_to_event(xmlNode * msg)
 {
     stonith_event_t *event = calloc(1, sizeof(stonith_event_t));
     const char *ntype = crm_element_value(msg, F_SUBTYPE);
-    char *data_addr = g_strdup_printf("//%s", ntype);
+    char *data_addr = crm_strdup_printf("//%s", ntype);
     xmlNode *data = get_xpath_object(data_addr, msg, LOG_DEBUG);
 
     crm_log_xml_trace(msg, "stonith_notify");
@@ -2000,7 +2179,7 @@ xml_to_event(xmlNode * msg)
         }
     }
 
-    g_free(data_addr);
+    free(data_addr);
     return event;
 }
 
@@ -2317,8 +2496,10 @@ stonith_api_new(void)
     new_stonith->cmds->remove_device   = stonith_api_remove_device;
     new_stonith->cmds->register_device = stonith_api_register_device;
 
-    new_stonith->cmds->remove_level    = stonith_api_remove_level;
-    new_stonith->cmds->register_level  = stonith_api_register_level;
+    new_stonith->cmds->remove_level          = stonith_api_remove_level;
+    new_stonith->cmds->remove_level_full     = stonith_api_remove_level_full;
+    new_stonith->cmds->register_level        = stonith_api_register_level;
+    new_stonith->cmds->register_level_full   = stonith_api_register_level_full;
 
     new_stonith->cmds->remove_callback       = stonith_api_del_callback;
     new_stonith->cmds->register_callback     = stonith_api_add_callback;
