@@ -47,6 +47,9 @@
 #  include "crm/common/cib_secrets.h"
 #endif
 
+/* ops currently active (in-flight) */
+extern GList *inflight_ops;
+
 static inline void
 set_fd_opts(int fd, int opts)
 {
@@ -92,9 +95,9 @@ svc_read_output(int fd, svc_action_t * op, bool is_stderr)
     do {
         rc = read(fd, buf, buf_read_len);
         if (rc > 0) {
-            crm_trace("Got %d characters starting with %.20s", rc, buf);
+            crm_trace("Got %d chars: %.80s", rc, buf);
             buf[rc] = 0;
-            data = realloc(data, len + rc + 1);
+            data = realloc_safe(data, len + rc + 1);
             len += sprintf(data + len, "%s", buf);
 
         } else if (errno != EINTR) {
@@ -199,6 +202,7 @@ add_OCF_env_vars(svc_action_t * op)
     set_ocf_env("OCF_RA_VERSION_MAJOR", "1", NULL);
     set_ocf_env("OCF_RA_VERSION_MINOR", "0", NULL);
     set_ocf_env("OCF_ROOT", OCF_ROOT_DIR, NULL);
+    set_ocf_env("OCF_EXIT_REASON_PREFIX", PCMK_OCF_REASON_PREFIX, NULL);
 
     if (op->rsc) {
         set_ocf_env("OCF_RESOURCE_INSTANCE", op->rsc, NULL);
@@ -219,13 +223,14 @@ recurring_action_timer(gpointer data)
 {
     svc_action_t *op = data;
 
-    crm_debug("Scheduling another invokation of %s", op->id);
+    crm_debug("Scheduling another invocation of %s", op->id);
 
     /* Clean out the old result */
     free(op->stdout_data);
     op->stdout_data = NULL;
     free(op->stderr_data);
     op->stderr_data = NULL;
+    op->opaque->repeat_timer = 0;
 
     services_action_async(op, NULL);
     return FALSE;
@@ -254,7 +259,11 @@ operation_finalize(svc_action_t * op)
 
     op->pid = 0;
 
-    if (!recurring) {
+    inflight_ops = g_list_remove(inflight_ops, op);
+
+    handle_blocked_ops();
+
+    if (!recurring && op->synchronous == FALSE) {
         /*
          * If this is a recurring action, do not free explicitly.
          * It will get freed whenever the action gets cancelled.
@@ -262,6 +271,8 @@ operation_finalize(svc_action_t * op)
         services_action_free(op);
         return TRUE;
     }
+
+    services_action_cleanup(op);
     return FALSE;
 }
 
@@ -269,7 +280,7 @@ static void
 operation_finished(mainloop_child_t * p, pid_t pid, int core, int signo, int exitcode)
 {
     svc_action_t *op = mainloop_child_userdata(p);
-    char *prefix = g_strdup_printf("%s:%d", op->id, op->pid);
+    char *prefix = crm_strdup_printf("%s:%d", op->id, op->pid);
 
     mainloop_clear_child_userdata(p);
     op->status = PCMK_LRM_OP_DONE;
@@ -316,70 +327,273 @@ operation_finished(mainloop_child_t * p, pid_t pid, int core, int signo, int exi
         crm_debug("%s - exited with rc=%d", prefix, exitcode);
     }
 
-    g_free(prefix);
-    prefix = g_strdup_printf("%s:%d:stderr", op->id, op->pid);
+    free(prefix);
+    prefix = crm_strdup_printf("%s:%d:stderr", op->id, op->pid);
     crm_log_output(LOG_NOTICE, prefix, op->stderr_data);
 
-    g_free(prefix);
-    prefix = g_strdup_printf("%s:%d:stdout", op->id, op->pid);
+    free(prefix);
+    prefix = crm_strdup_printf("%s:%d:stdout", op->id, op->pid);
     crm_log_output(LOG_DEBUG, prefix, op->stdout_data);
 
-    g_free(prefix);
+    free(prefix);
     operation_finalize(op);
 }
 
+/*!
+ * \internal
+ * \brief Set operation rc and status per errno from stat(), fork() or execvp()
+ *
+ * \param[in,out] op     Operation to set rc and status for
+ * \param[in]     error  Value of errno after system call
+ *
+ * \return void
+ */
 static void
 services_handle_exec_error(svc_action_t * op, int error)
 {
-    op->rc = PCMK_OCF_EXEC_ERROR;
-    op->status = PCMK_LRM_OP_ERROR;
+    int rc_not_installed, rc_insufficient_priv, rc_exec_error;
 
-    /* Need to mimic the return codes for each standard as thats what we'll convert back from in get_uniform_rc() */
+    /* Mimic the return codes for each standard as that's what we'll convert back from in get_uniform_rc() */
     if (safe_str_eq(op->standard, "lsb") && safe_str_eq(op->action, "status")) {
-        switch (error) {    /* see execve(2) */
-            case ENOENT:   /* No such file or directory */
-            case EISDIR:   /* Is a directory */
-                op->rc = PCMK_LSB_STATUS_NOT_INSTALLED;
-                op->status = PCMK_LRM_OP_NOT_INSTALLED;
-                break;
-            case EACCES:   /* permission denied (various errors) */
-                /* LSB status ops don't support 'not installed' */
-                break;
-        }
+        rc_not_installed = PCMK_LSB_STATUS_NOT_INSTALLED;
+        rc_insufficient_priv = PCMK_LSB_STATUS_INSUFFICIENT_PRIV;
+        rc_exec_error = PCMK_LSB_STATUS_UNKNOWN;
 
 #if SUPPORT_NAGIOS
     } else if (safe_str_eq(op->standard, "nagios")) {
-        switch (error) {
-            case ENOENT:   /* No such file or directory */
-            case EISDIR:   /* Is a directory */
-                op->rc = NAGIOS_NOT_INSTALLED;
-                op->status = PCMK_LRM_OP_NOT_INSTALLED;
-                break;
-            case EACCES:   /* permission denied (various errors) */
-                op->rc = NAGIOS_INSUFFICIENT_PRIV;
-                break;
-        }
+        rc_not_installed = NAGIOS_NOT_INSTALLED;
+        rc_insufficient_priv = NAGIOS_INSUFFICIENT_PRIV;
+        rc_exec_error = PCMK_OCF_EXEC_ERROR;
 #endif
 
     } else {
-        switch (error) {
-            case ENOENT:   /* No such file or directory */
-            case EISDIR:   /* Is a directory */
-                op->rc = PCMK_OCF_NOT_INSTALLED; /* Valid for LSB */
-                op->status = PCMK_LRM_OP_NOT_INSTALLED;
-                break;
-            case EACCES:   /* permission denied (various errors) */
-                op->rc = PCMK_OCF_INSUFFICIENT_PRIV; /* Valid for LSB */
-                break;
-        }
+        rc_not_installed = PCMK_OCF_NOT_INSTALLED;
+        rc_insufficient_priv = PCMK_OCF_INSUFFICIENT_PRIV;
+        rc_exec_error = PCMK_OCF_EXEC_ERROR;
+    }
+
+    switch (error) {   /* see execve(2), stat(2) and fork(2) */
+        case ENOENT:   /* No such file or directory */
+        case EISDIR:   /* Is a directory */
+        case ENOTDIR:  /* Path component is not a directory */
+        case EINVAL:   /* Invalid executable format */
+        case ENOEXEC:  /* Invalid executable format */
+            op->rc = rc_not_installed;
+            op->status = PCMK_LRM_OP_NOT_INSTALLED;
+            break;
+        case EACCES:   /* permission denied (various errors) */
+        case EPERM:    /* permission denied (various errors) */
+            op->rc = rc_insufficient_priv;
+            op->status = PCMK_LRM_OP_ERROR;
+            break;
+        default:
+            op->rc = rc_exec_error;
+            op->status = PCMK_LRM_OP_ERROR;
     }
 }
 
-/* Returns FALSE if 'op' should be free'd by the caller */
+static void
+action_launch_child(svc_action_t *op)
+{
+    int lpc;
+
+    /* SIGPIPE is ignored (which is different from signal blocking) by the gnutls library.
+     * Depending on the libqb version in use, libqb may set SIGPIPE to be ignored as well. 
+     * We do not want this to be inherited by the child process. By resetting this the signal
+     * to the default behavior, we avoid some potential odd problems that occur during OCF
+     * scripts when SIGPIPE is ignored by the environment. */
+    signal(SIGPIPE, SIG_DFL);
+
+#if defined(HAVE_SCHED_SETSCHEDULER)
+    if (sched_getscheduler(0) != SCHED_OTHER) {
+        struct sched_param sp;
+
+        memset(&sp, 0, sizeof(sp));
+        sp.sched_priority = 0;
+
+        if (sched_setscheduler(0, SCHED_OTHER, &sp) == -1) {
+            crm_perror(LOG_ERR, "Could not reset scheduling policy to SCHED_OTHER for %s", op->id);
+        }
+    }
+#endif
+    if (setpriority(PRIO_PROCESS, 0, 0) == -1) {
+        crm_perror(LOG_ERR, "Could not reset process priority to 0 for %s", op->id);
+    }
+
+    /* Man: The call setpgrp() is equivalent to setpgid(0,0)
+     * _and_ compiles on BSD variants too
+     * need to investigate if it works the same too.
+     */
+    setpgid(0, 0);
+
+    /* close all descriptors except stdin/out/err and channels to logd */
+    for (lpc = getdtablesize() - 1; lpc > STDERR_FILENO; lpc--) {
+        close(lpc);
+    }
+
+#if SUPPORT_CIBSECRETS
+    if (replace_secret_params(op->rsc, op->params) < 0) {
+        /* replacing secrets failed! */
+        if (safe_str_eq(op->action,"stop")) {
+            /* don't fail on stop! */
+            crm_info("proceeding with the stop operation for %s", op->rsc);
+
+        } else {
+            crm_err("failed to get secrets for %s, "
+                    "considering resource not configured", op->rsc);
+            _exit(PCMK_OCF_NOT_CONFIGURED);
+        }
+    }
+#endif
+    /* Setup environment correctly */
+    add_OCF_env_vars(op);
+
+    /* execute the RA */
+    execvp(op->opaque->exec, op->opaque->args);
+
+    /* Most cases should have been already handled by stat() */
+    services_handle_exec_error(op, errno);
+
+    _exit(op->rc);
+}
+
+static void
+action_synced_wait(svc_action_t * op, sigset_t mask)
+{
+
+#ifndef HAVE_SYS_SIGNALFD_H
+    CRM_ASSERT(FALSE);
+#else
+    int status = 0;
+    int timeout = op->timeout;
+    int sfd = -1;
+    time_t start = -1;
+    struct pollfd fds[3];
+    int wait_rc = 0;
+
+    sfd = signalfd(-1, &mask, SFD_NONBLOCK);
+    if (sfd < 0) {
+        crm_perror(LOG_ERR, "signalfd() failed");
+    }
+
+    fds[0].fd = op->opaque->stdout_fd;
+    fds[0].events = POLLIN;
+    fds[0].revents = 0;
+
+    fds[1].fd = op->opaque->stderr_fd;
+    fds[1].events = POLLIN;
+    fds[1].revents = 0;
+
+    fds[2].fd = sfd;
+    fds[2].events = POLLIN;
+    fds[2].revents = 0;
+
+    crm_trace("Waiting for %d", op->pid);
+    start = time(NULL);
+    do {
+        int poll_rc = poll(fds, 3, timeout);
+
+        if (poll_rc > 0) {
+            if (fds[0].revents & POLLIN) {
+                svc_read_output(op->opaque->stdout_fd, op, FALSE);
+            }
+
+            if (fds[1].revents & POLLIN) {
+                svc_read_output(op->opaque->stderr_fd, op, TRUE);
+            }
+
+            if (fds[2].revents & POLLIN) {
+                struct signalfd_siginfo fdsi;
+                ssize_t s;
+
+                s = read(sfd, &fdsi, sizeof(struct signalfd_siginfo));
+                if (s != sizeof(struct signalfd_siginfo)) {
+                    crm_perror(LOG_ERR, "Read from signal fd %d failed", sfd);
+
+                } else if (fdsi.ssi_signo == SIGCHLD) {
+                    wait_rc = waitpid(op->pid, &status, WNOHANG);
+
+                    if (wait_rc < 0){
+                        crm_perror(LOG_ERR, "waitpid() for %d failed", op->pid);
+
+                    } else if (wait_rc > 0) {
+                        break;
+                    }
+                }
+            }
+
+        } else if (poll_rc == 0) {
+            timeout = 0;
+            break;
+
+        } else if (poll_rc < 0) {
+            if (errno != EINTR) {
+                crm_perror(LOG_ERR, "poll() failed");
+                break;
+            }
+        }
+
+        timeout = op->timeout - (time(NULL) - start) * 1000;
+
+    } while ((op->timeout < 0 || timeout > 0));
+
+    crm_trace("Child done: %d", op->pid);
+    if (wait_rc <= 0) {
+        int killrc = kill(op->pid, SIGKILL);
+
+        op->rc = PCMK_OCF_UNKNOWN_ERROR;
+        if (op->timeout > 0 && timeout <= 0) {
+            op->status = PCMK_LRM_OP_TIMEOUT;
+            crm_warn("%s:%d - timed out after %dms", op->id, op->pid, op->timeout);
+
+        } else {
+            op->status = PCMK_LRM_OP_ERROR;
+        }
+
+        if (killrc && errno != ESRCH) {
+            crm_err("kill(%d, KILL) failed: %d", op->pid, errno);
+        }
+        /*
+         * From sigprocmask(2):
+         * It is not possible to block SIGKILL or SIGSTOP.  Attempts to do so are silently ignored.
+         *
+         * This makes it safe to skip WNOHANG here
+         */
+        waitpid(op->pid, &status, 0);
+
+    } else if (WIFEXITED(status)) {
+        op->status = PCMK_LRM_OP_DONE;
+        op->rc = WEXITSTATUS(status);
+        crm_info("Managed %s process %d exited with rc=%d", op->id, op->pid, op->rc);
+
+    } else if (WIFSIGNALED(status)) {
+        int signo = WTERMSIG(status);
+
+        op->status = PCMK_LRM_OP_ERROR;
+        crm_err("Managed %s process %d exited with signal=%d", op->id, op->pid, signo);
+    }
+#ifdef WCOREDUMP
+    if (WCOREDUMP(status)) {
+        crm_err("Managed %s process %d dumped core", op->id, op->pid);
+    }
+#endif
+
+    svc_read_output(op->opaque->stdout_fd, op, FALSE);
+    svc_read_output(op->opaque->stderr_fd, op, TRUE);
+
+    close(op->opaque->stdout_fd);
+    close(op->opaque->stderr_fd);
+    close(sfd);
+
+#endif
+
+}
+
+/* For an asynchronous 'op', returns FALSE if 'op' should be free'd by the caller */
+/* For a synchronous 'op', returns FALSE if 'op' fails */
 gboolean
 services_os_action_execute(svc_action_t * op, gboolean synchronous)
 {
-    int lpc;
     int stdout_fd[2];
     int stderr_fd[2];
     sigset_t mask;
@@ -434,27 +648,6 @@ services_os_action_execute(svc_action_t * op, gboolean synchronous)
                 return FALSE;
             }
         case 0:                /* Child */
-#if defined(HAVE_SCHED_SETSCHEDULER)
-            if (sched_getscheduler(0) != SCHED_OTHER) {
-                struct sched_param sp;
-
-                memset(&sp, 0, sizeof(sp));
-                sp.sched_priority = 0;
-
-                if (sched_setscheduler(0, SCHED_OTHER, &sp) == -1) {
-                    crm_perror(LOG_ERR, "Could not reset scheduling policy to SCHED_OTHER for %s", op->id);
-                }
-            }
-#endif
-            if (setpriority(PRIO_PROCESS, 0, 0) == -1) {
-                crm_perror(LOG_ERR, "Could not reset process priority to 0 for %s", op->id);
-            }
-
-            /* Man: The call setpgrp() is equivalent to setpgid(0,0)
-             * _and_ compiles on BSD variants too
-             * need to investigate if it works the same too.
-             */
-            setpgid(0, 0);
             close(stdout_fd[0]);
             close(stderr_fd[0]);
             if (STDOUT_FILENO != stdout_fd[1]) {
@@ -470,34 +663,7 @@ services_os_action_execute(svc_action_t * op, gboolean synchronous)
                 close(stderr_fd[1]);
             }
 
-            /* close all descriptors except stdin/out/err and channels to logd */
-            for (lpc = getdtablesize() - 1; lpc > STDERR_FILENO; lpc--) {
-                close(lpc);
-            }
-
-#if SUPPORT_CIBSECRETS
-            if (replace_secret_params(op->rsc, op->params) < 0) {
-                /* replacing secrets failed! */
-                if (safe_str_eq(op->action,"stop")) {
-                    /* don't fail on stop! */
-                    crm_info("proceeding with the stop operation for %s", op->rsc);
-
-                } else {
-                    crm_err("failed to get secrets for %s, "
-                            "considering resource not configured", op->rsc);
-                    _exit(PCMK_OCF_NOT_CONFIGURED);
-                }
-            }
-#endif
-            /* Setup environment correctly */
-            add_OCF_env_vars(op);
-
-            /* execute the RA */
-            execvp(op->opaque->exec, op->opaque->args);
-
-            /* Most cases should have been already handled by stat() */
-            services_handle_exec_error(op, errno);
-            _exit(op->rc);
+            action_launch_child(op);
     }
 
     /* Only the parent reaches here */
@@ -511,140 +677,24 @@ services_os_action_execute(svc_action_t * op, gboolean synchronous)
     set_fd_opts(op->opaque->stderr_fd, O_NONBLOCK);
 
     if (synchronous) {
-#ifndef HAVE_SYS_SIGNALFD_H
-        CRM_ASSERT(FALSE);
-#else
-        int status = 0;
-        int timeout = op->timeout;
-        int sfd = -1;
-        time_t start = -1;
-        struct pollfd fds[3];
-        int wait_rc = 0;
-
-        sfd = signalfd(-1, &mask, SFD_NONBLOCK);
-        if (sfd < 0) {
-            crm_perror(LOG_ERR, "signalfd() failed");
-        }
-
-        fds[0].fd = op->opaque->stdout_fd;
-        fds[0].events = POLLIN;
-        fds[0].revents = 0;
-
-        fds[1].fd = op->opaque->stderr_fd;
-        fds[1].events = POLLIN;
-        fds[1].revents = 0;
-
-        fds[2].fd = sfd;
-        fds[2].events = POLLIN;
-        fds[2].revents = 0;
-
-        crm_trace("Waiting for %d", op->pid);
-        start = time(NULL);
-        do {
-            int poll_rc = poll(fds, 3, timeout);
-
-            if (poll_rc > 0) {
-                if (fds[0].revents & POLLIN) {
-                    svc_read_output(op->opaque->stdout_fd, op, FALSE);
-                }
-
-                if (fds[1].revents & POLLIN) {
-                    svc_read_output(op->opaque->stderr_fd, op, TRUE);
-                }
-
-                if (fds[2].revents & POLLIN) {
-                    struct signalfd_siginfo fdsi;
-                    ssize_t s;
-
-                    s = read(sfd, &fdsi, sizeof(struct signalfd_siginfo));
-                    if (s != sizeof(struct signalfd_siginfo)) {
-                        crm_perror(LOG_ERR, "Read from signal fd %d failed", sfd);
-
-                    } else if (fdsi.ssi_signo == SIGCHLD) {
-                        wait_rc = waitpid(op->pid, &status, WNOHANG);
-
-                        if (wait_rc < 0){
-                            crm_perror(LOG_ERR, "waitpid() for %d failed", op->pid);
-
-                        } else if (wait_rc > 0) {
-                            break;
-                        }
-                    }
-                }
-
-            } else if (poll_rc == 0) {
-                timeout = 0;
-                break;
-
-            } else if (poll_rc < 0) {
-                if (errno != EINTR) {
-                    crm_perror(LOG_ERR, "poll() failed");
-                    break;
-                }
-            }
-
-            timeout = op->timeout - (time(NULL) - start) * 1000;
-
-        } while ((op->timeout < 0 || timeout > 0));
-
-        crm_trace("Child done: %d", op->pid);
-        if (wait_rc <= 0) {
-            int killrc = kill(op->pid, SIGKILL);
-
-            op->rc = PCMK_OCF_UNKNOWN_ERROR;
-            if (op->timeout > 0 && timeout <= 0) {
-                op->status = PCMK_LRM_OP_TIMEOUT;
-                crm_warn("%s:%d - timed out after %dms", op->id, op->pid, op->timeout);
-
-            } else {
-                op->status = PCMK_LRM_OP_ERROR;
-            }
-
-            if (killrc && errno != ESRCH) {
-                crm_err("kill(%d, KILL) failed: %d", op->pid, errno);
-            }
-            /*
-             * From sigprocmask(2):
-             * It is not possible to block SIGKILL or SIGSTOP.  Attempts to do so are silently ignored.
-             *
-             * This makes it safe to skip WNOHANG here
-             */
-            waitpid(op->pid, &status, 0);
-
-        } else if (WIFEXITED(status)) {
-            op->status = PCMK_LRM_OP_DONE;
-            op->rc = WEXITSTATUS(status);
-            crm_info("Managed %s process %d exited with rc=%d", op->id, op->pid, op->rc);
-
-        } else if (WIFSIGNALED(status)) {
-            int signo = WTERMSIG(status);
-
-            op->status = PCMK_LRM_OP_ERROR;
-            crm_err("Managed %s process %d exited with signal=%d", op->id, op->pid, signo);
-        }
-#ifdef WCOREDUMP
-        if (WCOREDUMP(status)) {
-            crm_err("Managed %s process %d dumped core", op->id, op->pid);
-        }
-#endif
-
-        svc_read_output(op->opaque->stdout_fd, op, FALSE);
-        svc_read_output(op->opaque->stderr_fd, op, TRUE);
-
-        close(op->opaque->stdout_fd);
-        close(op->opaque->stderr_fd);
-        close(sfd);
+        action_synced_wait(op, mask);
 
         if (sigismember(&old_mask, SIGCHLD) == 0) {
             if (sigprocmask(SIG_UNBLOCK, &mask, NULL) < 0) {
                 crm_perror(LOG_ERR, "sigprocmask() to unblocked failed");
             }
         }
-#endif
 
     } else {
+
         crm_trace("Async waiting for %d - %s", op->pid, op->opaque->exec);
-        mainloop_child_add(op->pid, op->timeout, op->id, op, operation_finished);
+        mainloop_child_add_with_flags(op->pid,
+                                      op->timeout,
+                                      op->id,
+                                      op,
+                                      (op->flags & SVC_ACTION_LEAVE_GROUP) ? mainloop_leave_pid_group : 0,
+                                      operation_finished);
+
 
         op->opaque->stdout_gsource = mainloop_add_fd(op->id,
                                                      G_PRIORITY_LOW,
@@ -653,6 +703,8 @@ services_os_action_execute(svc_action_t * op, gboolean synchronous)
         op->opaque->stderr_gsource = mainloop_add_fd(op->id,
                                                      G_PRIORITY_LOW,
                                                      op->opaque->stderr_fd, op, &stderr_callbacks);
+
+        services_add_inflight_op(op);
     }
 
     return TRUE;
@@ -765,14 +817,14 @@ resources_os_list_nagios_agents(void)
     /* Make sure both the plugin and its metadata exist */
     for (gIter = plugin_list; gIter != NULL; gIter = gIter->next) {
         const char *plugin = gIter->data;
-        char *metadata = g_strdup_printf(NAGIOS_METADATA_DIR "/%s.xml", plugin);
+        char *metadata = crm_strdup_printf(NAGIOS_METADATA_DIR "/%s.xml", plugin);
         struct stat st;
 
         if (stat(metadata, &st) == 0) {
             result = g_list_append(result, strdup(plugin));
         }
 
-        g_free(metadata);
+        free(metadata);
     }
     g_list_free_full(plugin_list, free);
     return result;
