@@ -19,9 +19,14 @@
 #ifndef __MH_SERVICES_PRIVATE_H__
 #  define __MH_SERVICES_PRIVATE_H__
 
+#if SUPPORT_DBUS
+#  include <dbus/dbus.h>
+#endif
+
+#define MAX_ARGC        255
 struct svc_action_private_s {
     char *exec;
-    char *args[255];
+    char *args[MAX_ARGC];
 
     guint repeat_timer;
     void (*callback) (svc_action_t * op);
@@ -31,6 +36,10 @@ struct svc_action_private_s {
 
     int stdout_fd;
     mainloop_io_t *stdout_gsource;
+#if SUPPORT_DBUS
+    DBusPendingCall* pending;
+    unsigned timerid;
+#endif
 };
 
 GList *services_os_get_directory_list(const char *root, gboolean files, gboolean executable);
@@ -50,4 +59,14 @@ gboolean cancel_recurring_action(svc_action_t * op);
 gboolean recurring_action_timer(gpointer data);
 gboolean operation_finalize(svc_action_t * op);
 
+void services_add_inflight_op(svc_action_t *op);
+
+void handle_blocked_ops(void);
+
+gboolean is_op_blocked(const char *rsc);
+
+#if SUPPORT_DBUS
+void services_set_op_pending(svc_action_t *op, DBusPendingCall *pending);
+#endif
+
 #endif                          /* __MH_SERVICES_PRIVATE_H__ */
