@@ -77,7 +77,7 @@ upstart_cleanup(void)
 }
 
 static gboolean
-upstart_job_by_name(const gchar * arg_name, gchar ** out_unit)
+upstart_job_by_name(const gchar * arg_name, gchar ** out_unit, int timeout)
 {
 /*
   com.ubuntu.Upstart0_6.GetJobByName (in String name, out ObjectPath job)
@@ -97,7 +97,7 @@ upstart_job_by_name(const gchar * arg_name, gchar ** out_unit)
 
     dbus_error_init(&error);
     CRM_LOG_ASSERT(dbus_message_append_args(msg, DBUS_TYPE_STRING, &arg_name, DBUS_TYPE_INVALID));
-    reply = pcmk_dbus_send_recv(msg, upstart_proxy, &error);
+    reply = pcmk_dbus_send_recv(msg, upstart_proxy, &error, timeout);
     dbus_message_unref(msg);
 
     if(error.name) {
@@ -190,7 +190,7 @@ upstart_job_listall(void)
                                        method); // method name
     CRM_ASSERT(msg != NULL);
 
-    reply = pcmk_dbus_send_recv(msg, upstart_proxy, &error);
+    reply = pcmk_dbus_send_recv(msg, upstart_proxy, &error, DBUS_TIMEOUT_USE_DEFAULT);
     dbus_message_unref(msg);
 
     if(error.name) {
@@ -246,11 +246,11 @@ upstart_job_listall(void)
 gboolean
 upstart_job_exists(const char *name)
 {
-    return upstart_job_by_name(name, NULL);
+    return upstart_job_by_name(name, NULL, DBUS_TIMEOUT_USE_DEFAULT);
 }
 
 static char *
-get_first_instance(const gchar * job)
+get_first_instance(const gchar * job, int timeout)
 {
     char *instance = NULL;
     const char *method = "GetAllInstances";
@@ -268,13 +268,17 @@ get_first_instance(const gchar * job)
     CRM_ASSERT(msg != NULL);
 
     dbus_message_append_args(msg, DBUS_TYPE_INVALID);
-    reply = pcmk_dbus_send_recv(msg, upstart_proxy, &error);
+    reply = pcmk_dbus_send_recv(msg, upstart_proxy, &error, timeout);
     dbus_message_unref(msg);
 
     if(error.name) {
         crm_err("Call to %s failed: %s", method, error.name);
         goto done;
 
+    } else if(reply == NULL) {
+        crm_err("Call to %s failed: no reply", method);
+        goto done;
+
     } else if (!dbus_message_iter_init(reply, &args)) {
         crm_err("Call to %s failed: Message has no arguments", method);
         goto done;
@@ -304,37 +308,29 @@ get_first_instance(const gchar * job)
     return instance;
 }
 
-gboolean
-upstart_job_running(const gchar * name)
+static void
+upstart_job_check(const char *name, const char *state, void *userdata)
 {
-    bool running = FALSE;
-    char *job = NULL;
+    svc_action_t * op = userdata;
 
-    if(upstart_job_by_name(name, &job)) {
-        char *path = get_first_instance(job);
-
-        if (path) {
-            char *state = pcmk_dbus_get_property(
-                upstart_proxy, BUS_NAME, path, UPSTART_06_API ".Instance", "state");
-
-            crm_info("State of %s: %s", name, state);
-            if (state) {
-                running = !g_strcmp0(state, "running");
-            }
-            free(state);
-        }
-        free(path);
+    if (state && g_strcmp0(state, "running") == 0) {
+        op->rc = PCMK_OCF_OK;
+    /* } else if (g_strcmp0(state, "activating") == 0) { */
+    /*     op->rc = PCMK_OCF_PENDING; */
+    } else {
+        op->rc = PCMK_OCF_NOT_RUNNING;
     }
 
-    free(job);
-    crm_info("%s is%s running", name, running ? "" : " not");
-    return running;
+    if (op->synchronous == FALSE) {
+        services_set_op_pending(op, NULL);
+        operation_finalize(op);
+    }
 }
 
 static char *
 upstart_job_metadata(const char *name)
 {
-    return g_strdup_printf("<?xml version=\"1.0\"?>\n"
+    return crm_strdup_printf("<?xml version=\"1.0\"?>\n"
                            "<!DOCTYPE resource-agent SYSTEM \"ra-api-1.dtd\">\n"
                            "<resource-agent name=\"%s\" version=\"0.1\">\n"
                            "  <version>1.0</version>\n"
@@ -393,6 +389,7 @@ upstart_async_dispatch(DBusPendingCall *pending, void *user_data)
     if(pending) {
         reply = dbus_pending_call_steal_reply(pending);
     }
+
     if(pcmk_dbus_find_error(op->action, pending, reply, &error)) {
 
         /* ignore "already started" or "not running" errors */
@@ -420,16 +417,17 @@ upstart_async_dispatch(DBusPendingCall *pending, void *user_data)
         }
     }
 
+    CRM_LOG_ASSERT(pending == op->opaque->pending);
+    services_set_op_pending(op, NULL);
     operation_finalize(op);
 
-    if(pending) {
-        dbus_pending_call_unref(pending);
-    }
     if(reply) {
         dbus_message_unref(reply);
     }
 }
 
+/* For an asynchronous 'op', returns FALSE if 'op' should be free'd by the caller */
+/* For a synchronous 'op', returns FALSE if 'op' fails */
 gboolean
 upstart_job_exec(svc_action_t * op, gboolean synchronous)
 {
@@ -452,7 +450,7 @@ upstart_job_exec(svc_action_t * op, gboolean synchronous)
         goto cleanup;
     }
 
-    if(!upstart_job_by_name(op->agent, &job)) {
+    if(!upstart_job_by_name(op->agent, &job, op->timeout)) {
         crm_debug("Could not obtain job named '%s' to %s", op->agent, action);
         if (!g_strcmp0(action, "stop")) {
             op->rc = PCMK_OCF_OK;
@@ -465,10 +463,30 @@ upstart_job_exec(svc_action_t * op, gboolean synchronous)
     }
 
     if (safe_str_eq(op->action, "monitor") || safe_str_eq(action, "status")) {
-        if (upstart_job_running(op->agent)) {
-            op->rc = PCMK_OCF_OK;
-        } else {
-            op->rc = PCMK_OCF_NOT_RUNNING;
+
+        char *path = get_first_instance(job, op->timeout);
+
+        op->rc = PCMK_OCF_NOT_RUNNING;
+        if(path) {
+            DBusPendingCall *pending = NULL;
+            char *state = pcmk_dbus_get_property(
+                upstart_proxy, BUS_NAME, path, UPSTART_06_API ".Instance", "state",
+                op->synchronous?NULL:upstart_job_check, op,
+                op->synchronous?NULL:&pending, op->timeout);
+
+            free(job);
+            free(path);
+
+            if(op->synchronous) {
+                upstart_job_check("state", state, op);
+                free(state);
+                return op->rc == PCMK_OCF_OK;
+            } else if (pending) {
+                services_set_op_pending(op, pending);
+                services_add_inflight_op(op);
+                return TRUE;
+            }
+            return FALSE;
         }
         goto cleanup;
 
@@ -503,13 +521,20 @@ upstart_job_exec(svc_action_t * op, gboolean synchronous)
 
     CRM_LOG_ASSERT(dbus_message_append_args(msg, DBUS_TYPE_BOOLEAN, &arg_wait, DBUS_TYPE_INVALID));
 
-    if (synchronous == FALSE) {
+    if (op->synchronous == FALSE) {
+        DBusPendingCall* pending = pcmk_dbus_send(msg, upstart_proxy, upstart_async_dispatch, op, op->timeout);
         free(job);
-        return pcmk_dbus_send(msg, upstart_proxy, upstart_async_dispatch, op);
+
+        if(pending) {
+            services_set_op_pending(op, pending);
+            services_add_inflight_op(op);
+            return TRUE;
+        }
+        return FALSE;
     }
 
     dbus_error_init(&error);
-    reply = pcmk_dbus_send_recv(msg, upstart_proxy, &error);
+    reply = pcmk_dbus_send_recv(msg, upstart_proxy, &error, op->timeout);
 
     if(error.name) {
         if(!upstart_mask_error(op, error.name)) {
@@ -545,9 +570,8 @@ upstart_job_exec(svc_action_t * op, gboolean synchronous)
         dbus_message_unref(reply);
     }
 
-    if (synchronous == FALSE) {
-        operation_finalize(op);
-        return TRUE;
+    if (op->synchronous == FALSE) {
+        return operation_finalize(op);
     }
     return op->rc == PCMK_OCF_OK;
 }
