@@ -35,6 +35,8 @@
 
 #include <dirent.h>
 #include <ctype.h>
+
+gboolean pcmk_quorate = FALSE;
 gboolean fatal_error = FALSE;
 GMainLoop *mainloop = NULL;
 
@@ -105,11 +107,7 @@ static uint32_t
 get_process_list(void)
 {
     int lpc = 0;
-    uint32_t procs = 0;
-
-    if(is_classic_ais_cluster()) {
-        procs |= crm_proc_plugin;
-    }
+    uint32_t procs = crm_get_cluster_proc();
 
     for (lpc = 0; lpc < SIZEOF(pcmk_children); lpc++) {
         if (pcmk_children[lpc].pid != 0) {
@@ -145,19 +143,12 @@ pcmk_process_exit(pcmk_child_t * child)
         mainloop_set_trigger(shutdown_trigger);
         update_node_processes(local_nodeid, NULL, get_process_list());
 
+    } else if (child->respawn && crm_is_true(getenv("PCMK_fail_fast"))) {
+        crm_err("Rebooting system because of %s", child->name);
+        pcmk_panic(__FUNCTION__);
+
     } else if (child->respawn) {
-        gboolean fail_fast = crm_is_true(getenv("PCMK_fail_fast"));
-
         crm_notice("Respawning failed child process: %s", child->name);
-
-#ifdef RB_HALT_SYSTEM
-        if (fail_fast) {
-            crm_err("Rebooting system", child->name);
-            sync();
-            reboot(RB_HALT_SYSTEM);
-            crm_exit(DAEMON_RESPAWN_STOP);
-        }
-#endif
         start_child(child);
     }
 }
@@ -168,21 +159,37 @@ pcmk_child_exit(mainloop_child_t * p, pid_t pid, int core, int signo, int exitco
     pcmk_child_t *child = mainloop_child_userdata(p);
     const char *name = mainloop_child_name(p);
 
-    if (signo) {
-        crm_notice("Child process %s terminated with signal %d (pid=%d, core=%d)",
-                   name, signo, pid, core);
+    if (signo && signo == SIGKILL) {
+        crm_warn("The %s process (%d) terminated with signal %d (core=%d)", name, pid, signo, core);
+
+    } else if (signo) {
+        crm_err("The %s process (%d) terminated with signal %d (core=%d)", name, pid, signo, core);
 
     } else {
-        do_crm_log(exitcode == 0 ? LOG_INFO : LOG_ERR,
-                   "Child process %s (%d) exited: %s (%d)", name, pid, pcmk_strerror(exitcode), exitcode);
-    }
+        switch(exitcode) {
+            case pcmk_ok:
+                crm_info("The %s process (%d) exited: %s (%d)", name, pid, pcmk_strerror(exitcode), exitcode);
+                break;
 
-    if (exitcode == 100) {
-        crm_warn("Pacemaker child process %s no longer wishes to be respawned. "
-                 "Shutting ourselves down.", name);
-        child->respawn = FALSE;
-        fatal_error = TRUE;
-        pcmk_shutdown(15);
+            case DAEMON_RESPAWN_STOP:
+                crm_warn("The %s process (%d) can no longer be respawned, shutting the cluster down.", name, pid);
+                child->respawn = FALSE;
+                fatal_error = TRUE;
+                pcmk_shutdown(SIGTERM);
+                break;
+
+            case pcmk_err_panic:
+                do_crm_log_always(LOG_EMERG, "The %s process (%d) instructed the machine to reset", name, pid);
+                child->respawn = FALSE;
+                fatal_error = TRUE;
+                pcmk_panic(__FUNCTION__);
+                pcmk_shutdown(SIGTERM);
+                break;
+
+            default:
+                crm_err("The %s process (%d) exited: %s (%d)", name, pid, pcmk_strerror(exitcode), exitcode);
+                break;
+        }
     }
 
     pcmk_process_exit(child);
@@ -315,7 +322,7 @@ start_child(pcmk_child_t * child)
 
                 /* Keep the root group (so we can access corosync), but add the haclient group (so we can access ipc) */
             } else if (initgroups(child->uid, gid) < 0) {
-                crm_err("Cannot initalize groups for %s: %s (%d)", child->uid, pcmk_strerror(errno), errno);
+                crm_err("Cannot initialize groups for %s: %s (%d)", child->uid, pcmk_strerror(errno), errno);
             }
         }
 
@@ -368,7 +375,7 @@ pcmk_shutdown_worker(gpointer user_data)
     int lpc = 0;
 
     if (phase == 0) {
-        crm_notice("Shuting down Pacemaker");
+        crm_notice("Shutting down Pacemaker");
         phase = max;
 
         /* Add a second, more frequent, check to speed up shutdown */
@@ -412,6 +419,15 @@ pcmk_shutdown_worker(gpointer user_data)
 
     /* send_cluster_id(); */
     crm_notice("Shutdown complete");
+
+    {
+        const char *delay = daemon_option("shutdown_delay");
+        if(delay) {
+            sync();
+            sleep(crm_get_msec(delay) / 1000);
+        }
+    }
+
     g_main_loop_quit(mainloop);
 
     if (fatal_error) {
@@ -422,6 +438,18 @@ pcmk_shutdown_worker(gpointer user_data)
     return TRUE;
 }
 
+static void
+pcmk_ignore(int nsig)
+{
+    crm_info("Ignoring signal %s (%d)", strsignal(nsig), nsig);
+}
+
+static void
+pcmk_sigquit(int nsig)
+{
+    pcmk_panic(__FUNCTION__);
+}
+
 void
 pcmk_shutdown(int nsig)
 {
@@ -465,7 +493,7 @@ pcmk_ipc_dispatch(qb_ipcs_connection_t * qbc, void *data, size_t size)
     task = crm_element_value(msg, F_CRM_TASK);
     if (crm_str_eq(task, CRM_OP_QUIT, TRUE)) {
         /* Time to quit */
-        crm_notice("Shutting down in responce to ticket %s (%s)",
+        crm_notice("Shutting down in response to ticket %s (%s)",
                    crm_element_value(msg, F_CRM_REFERENCE), crm_element_value(msg, F_CRM_ORIGIN));
         pcmk_shutdown(15);
 
@@ -521,6 +549,12 @@ struct qb_ipcs_service_handlers mcp_ipc_callbacks = {
     .connection_destroyed = pcmk_ipc_destroy
 };
 
+/*!
+ * \internal
+ * \brief Send an XML message with process list of all known peers to client(s)
+ *
+ * \param[in] client  Send message to this client, or all clients if NULL
+ */
 void
 update_process_clients(crm_client_t *client)
 {
@@ -528,7 +562,9 @@ update_process_clients(crm_client_t *client)
     crm_node_t *node = NULL;
     xmlNode *update = create_xml_node(NULL, "nodes");
 
-    crm_trace("Sending process list to %d children", crm_hash_table_size(client_connections));
+    if (is_corosync_cluster()) {
+        crm_xml_add_int(update, "quorate", pcmk_quorate);
+    }
 
     g_hash_table_iter_init(&iter, crm_peer_cache);
     while (g_hash_table_iter_next(&iter, NULL, (gpointer *) & node)) {
@@ -541,9 +577,11 @@ update_process_clients(crm_client_t *client)
     }
 
     if(client) {
+        crm_trace("Sending process list to client %s", client->id);
         crm_ipcs_send(client, 0, update, crm_ipc_server_event);
 
     } else {
+        crm_trace("Sending process list to %d clients", crm_hash_table_size(client_connections));
         g_hash_table_iter_init(&iter, client_connections);
         while (g_hash_table_iter_next(&iter, NULL, (gpointer *) & client)) {
             crm_ipcs_send(client, 0, update, crm_ipc_server_event);
@@ -553,6 +591,10 @@ update_process_clients(crm_client_t *client)
     free_xml(update);
 }
 
+/*!
+ * \internal
+ * \brief Send a CPG message with local node's process list to all peers
+ */
 void
 update_process_peers(void)
 {
@@ -578,6 +620,16 @@ update_process_peers(void)
     send_cpg_iov(iov);
 }
 
+/*!
+ * \internal
+ * \brief Update a node's process list, notifying clients and peers if needed
+ *
+ * \param[in] id     Node ID of affected node
+ * \param[in] uname  Uname of affected node
+ * \param[in] procs  Affected node's process list mask
+ *
+ * \return TRUE if the process list changed, FALSE otherwise
+ */
 gboolean
 update_node_processes(uint32_t id, const char *uname, uint32_t procs)
 {
@@ -591,15 +643,16 @@ update_node_processes(uint32_t id, const char *uname, uint32_t procs)
             node->processes = procs;
             changed = TRUE;
 
+            /* If local node's processes have changed, notify clients/peers */
+            if (id == local_nodeid) {
+                update_process_clients(NULL);
+                update_process_peers();
+            }
+
         } else {
             crm_trace("Node %s still has process list: %.32x", node->uname, procs);
         }
     }
-
-    if (changed && id == local_nodeid) {
-        update_process_clients(NULL);
-        update_process_peers();
-    }
     return changed;
 }
 
@@ -646,7 +699,7 @@ check_active_before_startup_processes(gpointer user_data)
                 continue;
             } else if (start_seq != pcmk_children[lpc].start_seq) {
                 continue;
-            } else if (crm_pid_active(pcmk_children[lpc].pid) != 1) {
+            } else if (crm_pid_active(pcmk_children[lpc].pid, pcmk_children[lpc].name) != 1) {
                 crm_notice("Process %s terminated (pid=%d)",
                            pcmk_children[lpc].name, pcmk_children[lpc].pid);
                 pcmk_process_exit(&(pcmk_children[lpc]));
@@ -661,59 +714,29 @@ check_active_before_startup_processes(gpointer user_data)
     return keep_tracking;
 }
 
-static void
+static bool
 find_and_track_existing_processes(void)
 {
     DIR *dp;
     struct dirent *entry;
-    struct stat statbuf;
     int start_tracker = 0;
+    char entry_name[64];
 
     dp = opendir("/proc");
     if (!dp) {
         /* no proc directory to search through */
         crm_notice("Can not read /proc directory to track existing components");
-        return;
+        return FALSE;
     }
 
     while ((entry = readdir(dp)) != NULL) {
-        char procpath[128];
-        char value[64];
-        char key[16];
-        FILE *file;
         int pid;
         int max = SIZEOF(pcmk_children);
         int i;
 
-        strcpy(procpath, "/proc/");
-        /* strlen("/proc/") + strlen("/status") + 1 = 14
-         * 128 - 14 = 114 */
-        strncat(procpath, entry->d_name, 114);
-
-        if (lstat(procpath, &statbuf)) {
+        if (crm_procfs_process_info(entry, entry_name, &pid) < 0) {
             continue;
         }
-        if (!S_ISDIR(statbuf.st_mode) || !isdigit(entry->d_name[0])) {
-            continue;
-        }
-
-        strcat(procpath, "/status");
-
-        file = fopen(procpath, "r");
-        if (!file) {
-            continue;
-        }
-        if (fscanf(file, "%15s%63s", key, value) != 2) {
-            fclose(file);
-            continue;
-        }
-        fclose(file);
-
-        pid = atoi(entry->d_name);
-        if (pid <= 0) {
-            continue;
-        }
-
         for (i = 0; i < max; i++) {
             const char *name = pcmk_children[i].name;
 
@@ -723,14 +746,12 @@ find_and_track_existing_processes(void)
             if (pcmk_children[i].flag == crm_proc_stonith_ng) {
                 name = "stonithd";
             }
-            if (safe_str_eq(name, value)) {
-                if (crm_pid_active(pid) != 1) {
-                    continue;
-                }
-                crm_notice("Tracking existing %s process (pid=%d)", value, pid);
+            if (safe_str_eq(entry_name, name) && (crm_pid_active(pid, NULL) == 1)) {
+                crm_notice("Tracking existing %s process (pid=%d)", name, pid);
                 pcmk_children[i].pid = pid;
                 pcmk_children[i].active_before_startup = TRUE;
                 start_tracker = 1;
+                break;
             }
         }
     }
@@ -740,6 +761,8 @@ find_and_track_existing_processes(void)
                               NULL);
     }
     closedir(dp);
+
+    return start_tracker;
 }
 
 static void
@@ -778,6 +801,17 @@ mcp_cpg_destroy(gpointer user_data)
     crm_exit(ENOTCONN);
 }
 
+/*!
+ * \internal
+ * \brief Process a CPG message (process list or manual peer cache removal)
+ *
+ * \param[in] handle     CPG connection (ignored)
+ * \param[in] groupName  CPG group name (ignored)
+ * \param[in] nodeid     ID of affected node
+ * \param[in] pid        Process ID (ignored)
+ * \param[in] msg        CPG XML message
+ * \param[in] msg_len    Length of msg in bytes (ignored)
+ */
 static void
 mcp_cpg_deliver(cpg_handle_t handle,
                  const struct cpg_name *groupName,
@@ -786,15 +820,20 @@ mcp_cpg_deliver(cpg_handle_t handle,
     xmlNode *xml = string2xml(msg);
     const char *task = crm_element_value(xml, F_CRM_TASK);
 
-    crm_trace("Received %s %.200s", task, msg);
-    if (task == NULL && nodeid != local_nodeid) {
-        uint32_t procs = 0;
-        const char *uname = crm_element_value(xml, "uname");
+    crm_trace("Received CPG message (%s): %.200s",
+              (task? task : "process list"), msg);
 
-        crm_element_value_int(xml, "proclist", (int *)&procs);
-        /* crm_debug("Got proclist %.32x from %s", procs, uname); */
-        if (update_node_processes(nodeid, uname, procs)) {
-            update_process_clients(NULL);
+    if (task == NULL) {
+        if (nodeid == local_nodeid) {
+            crm_info("Ignoring process list sent by peer for local node");
+        } else {
+            uint32_t procs = 0;
+            const char *uname = crm_element_value(xml, "uname");
+
+            crm_element_value_int(xml, "proclist", (int *)&procs);
+            if (update_node_processes(nodeid, uname, procs)) {
+                update_process_clients(NULL);
+            }
         }
 
     } else if (crm_str_eq(task, CRM_OP_RM_NODE_CACHE, TRUE)) {
@@ -805,6 +844,10 @@ mcp_cpg_deliver(cpg_handle_t handle,
         name = crm_element_value(xml, XML_ATTR_UNAME);
         reap_crm_member(id, name);
     }
+
+    if (xml != NULL) {
+        free_xml(xml);
+    }
 }
 
 static void
@@ -814,22 +857,42 @@ mcp_cpg_membership(cpg_handle_t handle,
                     const struct cpg_address *left_list, size_t left_list_entries,
                     const struct cpg_address *joined_list, size_t joined_list_entries)
 {
-    /* Don't care about CPG membership, but we do want to broadcast our own presence */
+    /* Update peer cache if needed */
+    pcmk_cpg_membership(handle, groupName, member_list, member_list_entries,
+                        left_list, left_list_entries,
+                        joined_list, joined_list_entries);
+
+    /* Always broadcast our own presence after any membership change */
     update_process_peers();
 }
 
 static gboolean
 mcp_quorum_callback(unsigned long long seq, gboolean quorate)
 {
-    /* Nothing to do */
+    pcmk_quorate = quorate;
     return TRUE;
 }
 
 static void
 mcp_quorum_destroy(gpointer user_data)
+{
+    crm_info("connection lost");
+}
+
+#if SUPPORT_CMAN
+static gboolean
+mcp_cman_dispatch(unsigned long long seq, gboolean quorate)
+{
+    pcmk_quorate = quorate;
+    return TRUE;
+}
+
+static void
+mcp_cman_destroy(gpointer user_data)
 {
     crm_info("connection closed");
 }
+#endif
 
 int
 main(int argc, char **argv)
@@ -851,6 +914,8 @@ main(int argc, char **argv)
 
     crm_log_preinit(NULL, argc, argv);
     crm_set_options(NULL, "mode [options]", long_options, "Start/Stop Pacemaker\n");
+    mainloop_add_signal(SIGHUP, pcmk_ignore);
+    mainloop_add_signal(SIGQUIT, pcmk_sigquit);
 
     while (1) {
         flag = crm_get_option(argc, argv, &option_index);
@@ -875,7 +940,7 @@ main(int argc, char **argv)
                 shutdown = TRUE;
                 break;
             case 'F':
-                printf("Pacemaker %s (Build: %s)\n Supporting v%s: %s\n", VERSION, BUILD_VERSION,
+                printf("Pacemaker %s (Build: %s)\n Supporting v%s: %s\n", PACEMAKER_VERSION, BUILD_VERSION,
                        CRM_FEATURE_SET, CRM_FEATURES);
                 crm_exit(pcmk_ok);
             default:
@@ -942,8 +1007,9 @@ main(int argc, char **argv)
         crm_exit(ENODATA);
     }
 
-    crm_notice("Starting Pacemaker %s (Build: %s): %s", VERSION, BUILD_VERSION, CRM_FEATURES);
+    crm_notice("Starting Pacemaker %s (Build: %s): %s", PACEMAKER_VERSION, BUILD_VERSION, CRM_FEATURES);
     mainloop = g_main_new(FALSE);
+    sysrq_init();
 
     rc = getrlimit(RLIMIT_CORE, &cores);
     if (rc < 0) {
@@ -972,6 +1038,7 @@ main(int argc, char **argv)
         }
 #endif
     }
+    rc = pcmk_ok;
 
     if (crm_user_lookup(CRM_DAEMON_USER, &pcmk_uid, &pcmk_gid) < 0) {
         crm_err("Cluster user %s does not exist, aborting Pacemaker startup", CRM_DAEMON_USER);
@@ -1011,23 +1078,38 @@ main(int argc, char **argv)
         crm_exit(ENOPROTOOPT);
     }
 
+    if(pcmk_locate_sbd() > 0) {
+        setenv("PCMK_watchdog", "true", 1);
+    } else {
+        setenv("PCMK_watchdog", "false", 1);
+    }
+
+    find_and_track_existing_processes();
+
     cluster.destroy = mcp_cpg_destroy;
     cluster.cpg.cpg_deliver_fn = mcp_cpg_deliver;
     cluster.cpg.cpg_confchg_fn = mcp_cpg_membership;
 
+    crm_set_autoreap(FALSE);
+
     if(cluster_connect_cpg(&cluster) == FALSE) {
         crm_err("Couldn't connect to Corosync's CPG service");
-        crm_exit(ENOPROTOOPT);
+        rc = -ENOPROTOOPT;
     }
 
-    rc = pcmk_ok;
-    if (is_corosync_cluster()) {
+    if (rc == pcmk_ok && is_corosync_cluster()) {
         /* Keep the membership list up-to-date for crm_node to query */
         if(cluster_connect_quorum(mcp_quorum_callback, mcp_quorum_destroy) == FALSE) {
             rc = -ENOTCONN;
         }
     }
 
+#if SUPPORT_CMAN
+    if (rc == pcmk_ok && is_cman_cluster()) {
+        init_cman_connection(mcp_cman_dispatch, mcp_cman_destroy);
+    }
+#endif
+
     if(rc == pcmk_ok) {
         local_name = get_local_node_name();
         update_node_processes(local_nodeid, local_name, get_process_list());
@@ -1035,7 +1117,6 @@ main(int argc, char **argv)
         mainloop_add_signal(SIGTERM, pcmk_shutdown);
         mainloop_add_signal(SIGINT, pcmk_shutdown);
 
-        find_and_track_existing_processes();
         init_children_processes();
 
         crm_info("Starting mainloop");
