@@ -96,6 +96,34 @@ resource_alloc_functions_t resource_class_alloc_functions[] = {
      }
 };
 
+gboolean
+update_action_flags(action_t * action, enum pe_action_flags flags)
+{
+    static unsigned long calls = 0;
+    gboolean changed = FALSE;
+    gboolean clear = is_set(flags, pe_action_clear);
+    enum pe_action_flags last = action->flags;
+
+    if (clear) {
+        pe_clear_action_bit(action, flags);
+    } else {
+        pe_set_action_bit(action, flags);
+    }
+
+    if (last != action->flags) {
+        calls++;
+        changed = TRUE;
+        /* Useful for tracking down _who_ changed a specific flag */
+        /* CRM_ASSERT(calls != 534); */
+        clear_bit(flags, pe_action_clear);
+        crm_trace("%s on %s: %sset flags 0x%.6x (was 0x%.6x, now 0x%.6x, %lu)",
+                  action->uuid, action->node ? action->node->details->uname : "[none]",
+                  clear ? "un-" : "", flags, last, action->flags, calls);
+    }
+
+    return changed;
+}
+
 static gboolean
 check_rsc_parameters(resource_t * rsc, node_t * node, xmlNode * rsc_entry,
                      gboolean active_here, pe_working_set_t * data_set)
@@ -199,6 +227,7 @@ check_action_definition(resource_t * rsc, node_t * active_node, xmlNode * xml_op
 
     const char *task = crm_element_value(xml_op, XML_LRM_ATTR_TASK);
     const char *op_version;
+    const char *digest_secure = NULL;
 
     CRM_CHECK(active_node != NULL, return FALSE);
     if (safe_str_eq(task, RSC_STOP)) {
@@ -239,13 +268,28 @@ check_action_definition(resource_t * rsc, node_t * active_node, xmlNode * xml_op
     } else if (interval == 0 && safe_str_eq(task, RSC_MIGRATED)) {
         /* Reload based on the start action not a migrate */
         task = RSC_START;
+    } else if (interval == 0 && safe_str_eq(task, RSC_PROMOTE)) {
+        /* Reload based on the start action not a promote */
+        task = RSC_START;
     }
 
-    digest_data = rsc_action_digest_cmp(rsc, xml_op, active_node, data_set);
     op_version = crm_element_value(xml_op, XML_ATTR_CRM_VERSION);
+    digest_data = rsc_action_digest_cmp(rsc, xml_op, active_node, data_set);
 
-    /* Changes that force a restart */
-    if (digest_data->rc == RSC_DIGEST_RESTART) {
+    if(is_set(data_set->flags, pe_flag_sanitized)) {
+        digest_secure = crm_element_value(xml_op, XML_LRM_ATTR_SECURE_DIGEST);
+    }
+
+    if(digest_data->rc != RSC_DIGEST_MATCH
+       && digest_secure
+       && digest_data->digest_secure_calc
+       && strcmp(digest_data->digest_secure_calc, digest_secure) == 0) {
+        fprintf(stdout, "Only 'private' parameters to %s_%s_%d on %s changed: %s\n",
+                rsc->id, task, interval, active_node->details->uname,
+                crm_element_value(xml_op, XML_ATTR_TRANSITION_MAGIC));
+
+    } else if (digest_data->rc == RSC_DIGEST_RESTART) {
+        /* Changes that force a restart */
         const char *digest_restart = crm_element_value(xml_op, XML_LRM_ATTR_RESTART_DIGEST);
 
         did_change = TRUE;
@@ -282,12 +326,11 @@ check_action_definition(resource_t * rsc, node_t * active_node, xmlNode * xml_op
             update_action_flags(op, pe_action_allow_reload_conversion);
 #else
             /* Re-sending the recurring op is sufficient - the old one will be cancelled automatically */
-            op = custom_action(rsc, key, task, NULL, FALSE, TRUE, data_set);
-            custom_action_order(rsc, start_key(rsc), NULL,
-                                NULL, NULL, op, pe_order_runnable_left, data_set);
+            op = custom_action(rsc, key, task, active_node, TRUE, TRUE, data_set);
+            set_bit(op->flags, pe_action_reschedule);
 #endif
 
-        } else if (digest_restart) {
+        } else if (digest_restart && rsc->isolation_wrapper == NULL && (uber_parent(rsc))->isolation_wrapper == NULL) {
             pe_rsc_trace(rsc, "Reloading '%s' action for resource %s", task, rsc->id);
 
             /* Allow this resource to reload - unless something else causes a full restart */
@@ -300,7 +343,7 @@ check_action_definition(resource_t * rsc, node_t * active_node, xmlNode * xml_op
             pe_rsc_trace(rsc, "Resource %s doesn't know how to reload", rsc->id);
 
             /* Re-send the start/demote/promote op
-             * Recurring ops will be detected independantly
+             * Recurring ops will be detected independently
              */
             custom_action(rsc, key, task, NULL, FALSE, TRUE, data_set);
         }
@@ -359,7 +402,7 @@ check_actions_for(xmlNode * rsc_entry, resource_t * rsc, node_t * node, pe_worki
         DeleteRsc(rsc, node, FALSE, data_set);
     }
 
-    for (rsc_op = __xml_first_child(rsc_entry); rsc_op != NULL; rsc_op = __xml_next(rsc_op)) {
+    for (rsc_op = __xml_first_child(rsc_entry); rsc_op != NULL; rsc_op = __xml_next_element(rsc_op)) {
         if (crm_str_eq((const char *)rsc_op->name, XML_LRM_TAG_RSC_OP, TRUE)) {
             op_list = g_list_prepend(op_list, rsc_op);
         }
@@ -396,7 +439,7 @@ check_actions_for(xmlNode * rsc_entry, resource_t * rsc, node_t * node, pe_worki
             (is_set(rsc->flags, pe_rsc_maintenance) || node->details->maintenance)) {
             CancelXmlOp(rsc, rsc_op, node, "maintenance mode", data_set);
 
-        } else if (is_probe || safe_str_eq(task, RSC_START) || interval > 0
+        } else if (is_probe || safe_str_eq(task, RSC_START) || safe_str_eq(task, RSC_PROMOTE) || interval > 0
                    || safe_str_eq(task, RSC_MIGRATED)) {
             did_change = check_action_definition(rsc, node, rsc_op, data_set);
         }
@@ -484,7 +527,7 @@ check_actions(pe_working_set_t * data_set)
     xmlNode *node_state = NULL;
 
     for (node_state = __xml_first_child(status); node_state != NULL;
-         node_state = __xml_next(node_state)) {
+         node_state = __xml_next_element(node_state)) {
         if (crm_str_eq((const char *)node_state->name, XML_CIB_TAG_STATE, TRUE)) {
             id = crm_element_value(node_state, XML_ATTR_ID);
             lrm_rscs = find_xml_node(node_state, XML_CIB_TAG_LRM, FALSE);
@@ -497,7 +540,8 @@ check_actions(pe_working_set_t * data_set)
 
             /* Still need to check actions for a maintenance node to cancel existing monitor operations */
             } else if (can_run_resources(node) == FALSE && node->details->maintenance == FALSE) {
-                crm_trace("Skipping param check for %s: cant run resources", node->details->uname);
+                crm_trace("Skipping param check for %s: can't run resources",
+                          node->details->uname);
                 continue;
             }
 
@@ -506,7 +550,7 @@ check_actions(pe_working_set_t * data_set)
                 xmlNode *rsc_entry = NULL;
 
                 for (rsc_entry = __xml_first_child(lrm_rscs); rsc_entry != NULL;
-                     rsc_entry = __xml_next(rsc_entry)) {
+                     rsc_entry = __xml_next_element(rsc_entry)) {
                     if (crm_str_eq((const char *)rsc_entry->name, XML_LRM_TAG_RESOURCE, TRUE)) {
 
                         if (xml_has_children(rsc_entry)) {
@@ -755,7 +799,7 @@ apply_system_health(pe_working_set_t * data_set)
             for (; gIter2 != NULL; gIter2 = gIter2->next) {
                 resource_t *rsc = (resource_t *) gIter2->data;
 
-                rsc2node_new(health_strategy, rsc, system_health, node, data_set);
+                rsc2node_new(health_strategy, rsc, system_health, NULL, node, data_set);
             }
         }
 
@@ -785,62 +829,18 @@ stage0(pe_working_set_t * data_set)
     return TRUE;
 }
 
-static void
-wait_for_probe(resource_t * rsc, const char *action, action_t * probe_complete,
-               pe_working_set_t * data_set)
-{
-    if (probe_complete == NULL) {
-        return;
-    }
-
-    if (rsc->children) {
-        GListPtr gIter = rsc->children;
-
-        for (; gIter != NULL; gIter = gIter->next) {
-            resource_t *child = (resource_t *) gIter->data;
-
-            wait_for_probe(child, action, probe_complete, data_set);
-        }
-
-    } else {
-        char *key = NULL;
-
-        if (safe_str_eq(action, RSC_STOP) && g_list_length(rsc->running_on) == 1) {
-            node_t *node = (node_t *) rsc->running_on->data;
-
-            /* Stop actions on nodes that are shutting down do not need to wait for probes to complete
-             * Doing so prevents node shutdown in the presence of nodes that are coming up
-             * The purpose of waiting is to not stop resources until we know for sure the
-             *  intended destination is able to take them
-             */
-            if (node && node->details->shutdown) {
-                crm_debug("Skipping %s before %s_%s_0 due to %s shutdown",
-                          probe_complete->uuid, rsc->id, action, node->details->uname);
-                return;
-            }
-        }
-
-        key = generate_op_key(rsc->id, action, 0);
-        custom_action_order(NULL, NULL, probe_complete, rsc, key, NULL,
-                            pe_order_optional, data_set);
-    }
-}
-
 /*
  * Check nodes for resources started outside of the LRM
  */
 gboolean
 probe_resources(pe_working_set_t * data_set)
 {
-    action_t *probe_complete = NULL;
     action_t *probe_node_complete = NULL;
-    action_t *probe_cluster_nodes_complete = NULL;
 
     GListPtr gIter = NULL;
     GListPtr gIter2 = NULL;
 
-    gIter = data_set->nodes;
-    for (; gIter != NULL; gIter = gIter->next) {
+    for (gIter = data_set->nodes; gIter != NULL; gIter = gIter->next) {
         node_t *node = (node_t *) gIter->data;
         const char *probed = g_hash_table_lookup(node->details->attrs, CRM_OP_PROBED);
 
@@ -861,89 +861,50 @@ probe_resources(pe_working_set_t * data_set)
             /* TODO enable container node probes once ordered probing is implemented. */
             continue;
 
-        } else if (probe_complete == NULL) {
-            probe_complete = get_pseudo_op(CRM_OP_PROBED, data_set);
-            if (is_set(data_set->flags, pe_flag_have_remote_nodes)) {
-                probe_cluster_nodes_complete = get_pseudo_op(CRM_OP_NODES_PROBED, data_set);
-            }
+        } else if (node->details->rsc_discovery_enabled == FALSE) {
+            /* resource discovery is disabled for this node */
+            continue;
         }
 
         if (probed != NULL && crm_is_true(probed) == FALSE) {
-            action_t *probe_op = custom_action(NULL, g_strdup_printf("%s-%s", CRM_OP_REPROBE, node->details->uname),
+            action_t *probe_op = custom_action(NULL, crm_strdup_printf("%s-%s", CRM_OP_REPROBE, node->details->uname),
                                                CRM_OP_REPROBE, node, FALSE, TRUE, data_set);
 
             add_hash_param(probe_op->meta, XML_ATTR_TE_NOWAIT, XML_BOOLEAN_TRUE);
             continue;
         }
 
-        probe_node_complete = custom_action(NULL, g_strdup_printf("%s-%s", CRM_OP_PROBED, node->details->uname),
-                                            CRM_OP_PROBED, node, FALSE, TRUE, data_set);
-        if (crm_is_true(probed)) {
-            crm_trace("unset");
-            update_action_flags(probe_node_complete, pe_action_optional);
-        } else {
-            crm_trace("set");
-            update_action_flags(probe_node_complete, pe_action_optional | pe_action_clear);
-        }
-        crm_trace("%s - %d", node->details->uname, probe_node_complete->flags & pe_action_optional);
-        probe_node_complete->priority = INFINITY;
-        add_hash_param(probe_node_complete->meta, XML_ATTR_TE_NOWAIT, XML_BOOLEAN_TRUE);
-
-        if (node->details->pending) {
-            update_action_flags(probe_node_complete, pe_action_runnable | pe_action_clear);
-            crm_info("Action %s on %s is unrunnable (pending)",
-                     probe_node_complete->uuid, probe_node_complete->node->details->uname);
-        }
-
-        if (is_remote_node(node)) {
-            order_actions(probe_node_complete, probe_complete,
-                      pe_order_runnable_left /*|pe_order_implies_then */ );
-        } else if (probe_cluster_nodes_complete == NULL) {
-            order_actions(probe_node_complete, probe_complete,
-                      pe_order_runnable_left /*|pe_order_implies_then */ );
-        } else {
-            order_actions(probe_node_complete, probe_cluster_nodes_complete,
-                      pe_order_runnable_left /*|pe_order_implies_then */ );
-        }
-
-        gIter2 = data_set->resources;
-        for (; gIter2 != NULL; gIter2 = gIter2->next) {
+        for (gIter2 = data_set->resources; gIter2 != NULL; gIter2 = gIter2->next) {
             resource_t *rsc = (resource_t *) gIter2->data;
 
-            if (rsc->cmds->create_probe(rsc, node, probe_node_complete, FALSE, data_set)) {
-                update_action_flags(probe_complete, pe_action_optional | pe_action_clear);
-                update_action_flags(probe_node_complete, pe_action_optional | pe_action_clear);
-
-                if (probe_cluster_nodes_complete
-                    && (rsc->is_remote_node || rsc_contains_remote_node(data_set, rsc))) {
-                    update_action_flags(probe_cluster_nodes_complete, pe_action_optional | pe_action_clear);
-                    /* allow remote connection resources and resources
-                     * containing remote connection resources to run after all
-                     * cluster nodes are probed */
-                    wait_for_probe(rsc, RSC_START, probe_cluster_nodes_complete, data_set);
-                } else {
-                    wait_for_probe(rsc, RSC_START, probe_complete, data_set);
-                }
-            }
+            rsc->cmds->create_probe(rsc, node, probe_node_complete, FALSE, data_set);
         }
     }
-
-    gIter = data_set->resources;
-    for (; gIter != NULL; gIter = gIter->next) {
-        resource_t *rsc = (resource_t *) gIter->data;
-
-        if (rsc->is_remote_node || rsc_contains_remote_node(data_set, rsc)) {
-            /* allow remote connection resources and any resources containing
-             * remote connection resources to run after cluster nodes are probed.*/
-            wait_for_probe(rsc, RSC_STOP, probe_cluster_nodes_complete, data_set);
-        } else {
-            wait_for_probe(rsc, RSC_STOP, probe_complete, data_set);
-        }
-    }
-
     return TRUE;
 }
 
+static void
+rsc_discover_filter(resource_t *rsc, node_t *node)
+{
+    GListPtr gIter = rsc->children;
+    resource_t *top = uber_parent(rsc);
+    node_t *match;
+
+    if (rsc->exclusive_discover == FALSE && top->exclusive_discover == FALSE) {
+        return;
+    }
+
+    for (; gIter != NULL; gIter = gIter->next) {
+        resource_t *child_rsc = (resource_t *) gIter->data;
+        rsc_discover_filter(child_rsc, node);
+    }
+
+    match = g_hash_table_lookup(rsc->allowed_nodes, node->details->id);
+    if (match && match->rsc_discover_mode != discover_exclusive) {
+        match->weight = -INFINITY;
+    }
+}
+
 /*
  * Count how many valid nodes we have (so we know the maximum number of
  *  colors we can resolve).
@@ -982,6 +943,7 @@ stage2(pe_working_set_t * data_set)
             resource_t *rsc = (resource_t *) gIter2->data;
 
             common_apply_stickiness(rsc, node, data_set);
+            rsc_discover_filter(rsc, node);
         }
     }
 
@@ -1029,7 +991,8 @@ sort_rsc_process_order(gconstpointer a, gconstpointer b, gpointer data)
     resource_t *resource1 = (resource_t *) convert_const_pointer(a);
     resource_t *resource2 = (resource_t *) convert_const_pointer(b);
 
-    node_t *node = NULL;
+    node_t *r1_node = NULL;
+    node_t *r2_node = NULL;
     GListPtr gIter = NULL;
     GHashTable *r1_nodes = NULL;
     GHashTable *r2_nodes = NULL;
@@ -1078,17 +1041,17 @@ sort_rsc_process_order(gconstpointer a, gconstpointer b, gpointer data)
     r2_weight = -INFINITY;
 
     if (resource1->running_on) {
-        node = g_list_nth_data(resource1->running_on, 0);
-        node = g_hash_table_lookup(r1_nodes, node->details->id);
-        if (node != NULL) {
-            r1_weight = node->weight;
+        r1_node = g_list_nth_data(resource1->running_on, 0);
+        r1_node = g_hash_table_lookup(r1_nodes, r1_node->details->id);
+        if (r1_node != NULL) {
+            r1_weight = r1_node->weight;
         }
     }
     if (resource2->running_on) {
-        node = g_list_nth_data(resource2->running_on, 0);
-        node = g_hash_table_lookup(r2_nodes, node->details->id);
-        if (node != NULL) {
-            r2_weight = node->weight;
+        r2_node = g_list_nth_data(resource2->running_on, 0);
+        r2_node = g_hash_table_lookup(r2_nodes, r2_node->details->id);
+        if (r2_node != NULL) {
+            r2_weight = r2_node->weight;
         }
     }
 
@@ -1104,10 +1067,10 @@ sort_rsc_process_order(gconstpointer a, gconstpointer b, gpointer data)
 
     reason = "score";
     for (gIter = nodes; gIter != NULL; gIter = gIter->next) {
-        node_t *r1_node = NULL;
-        node_t *r2_node = NULL;
+        node_t *node = (node_t *) gIter->data;
 
-        node = (node_t *) gIter->data;
+        r1_node = NULL;
+        r2_node = NULL;
 
         r1_weight = -INFINITY;
         if (r1_nodes) {
@@ -1137,6 +1100,11 @@ sort_rsc_process_order(gconstpointer a, gconstpointer b, gpointer data)
     }
 
   done:
+    crm_trace("%s (%d) on %s %c %s (%d) on %s: %s",
+              resource1->id, r1_weight, r1_node ? r1_node->details->id : "n/a",
+              rc < 0 ? '>' : rc > 0 ? '<' : '=',
+              resource2->id, r2_weight, r2_node ? r2_node->details->id : "n/a", reason);
+
     if (r1_nodes) {
         g_hash_table_destroy(r1_nodes);
     }
@@ -1144,9 +1112,6 @@ sort_rsc_process_order(gconstpointer a, gconstpointer b, gpointer data)
         g_hash_table_destroy(r2_nodes);
     }
 
-    crm_trace("%s (%d) %c %s (%d) on %s: %s",
-              resource1->id, r1_weight, rc < 0 ? '>' : rc > 0 ? '<' : '=',
-              resource2->id, r2_weight, node ? node->details->id : "n/a", reason);
     return rc;
 }
 
@@ -1164,7 +1129,10 @@ allocate_resources(pe_working_set_t * data_set)
                 continue;
             }
             pe_rsc_trace(rsc, "Allocating: %s", rsc->id);
-            rsc->cmds->allocate(rsc, NULL, data_set);
+            /* for remote node connection resources, always prefer the partial migration
+             * target during resource allocation if the rsc is in the middle of a
+             * migration */ 
+            rsc->cmds->allocate(rsc, rsc->partial_migration_target, data_set);
         }
     }
 
@@ -1179,6 +1147,41 @@ allocate_resources(pe_working_set_t * data_set)
     }
 }
 
+static void
+cleanup_orphans(resource_t * rsc, pe_working_set_t * data_set)
+{
+    GListPtr gIter = NULL;
+
+    if (is_set(data_set->flags, pe_flag_stop_rsc_orphans) == FALSE) {
+        return;
+    }
+
+    /* Don't recurse into ->children, those are just unallocated clone instances */
+    if(is_not_set(rsc->flags, pe_rsc_orphan)) {
+        return;
+    }
+
+    for (gIter = data_set->nodes; gIter != NULL; gIter = gIter->next) {
+        node_t *node = (node_t *) gIter->data;
+
+        if (node->details->online && get_failcount(node, rsc, NULL, data_set)) {
+            action_t *clear_op = NULL;
+
+            clear_op = custom_action(rsc, crm_concat(rsc->id, CRM_OP_CLEAR_FAILCOUNT, '_'),
+                                     CRM_OP_CLEAR_FAILCOUNT, node, FALSE, TRUE, data_set);
+
+            add_hash_param(clear_op->meta, XML_ATTR_TE_NOWAIT, XML_BOOLEAN_TRUE);
+            pe_rsc_info(rsc, "Clearing failcount (%d) for orphaned resource %s on %s (%s)",
+                        get_failcount(node, rsc, NULL, data_set), rsc->id, node->details->uname,
+                        clear_op->uuid);
+
+            custom_action_order(rsc, NULL, clear_op,
+                            rsc, generate_op_key(rsc->id, RSC_STOP, 0), NULL,
+                            pe_order_optional, data_set);
+        }
+    }
+}
+
 gboolean
 stage5(pe_working_set_t * data_set)
 {
@@ -1248,15 +1251,21 @@ stage5(pe_working_set_t * data_set)
          ptest[14637]: 2010/09/27_17:56:25 notice: TRACE: do_calculations: pengine.c:306 Processing fencing and shutdown cases
          ptest[14637]: 2010/09/27_17:56:25 notice: TRACE: do_calculations: pengine.c:313 Applying ordering constraints
          ptest[14637]: 2010/09/27_17:56:25 notice: TRACE: do_calculations: pengine.c:320 Create transition graph
-         */
+        */
 
         probe_resources(data_set);
     }
 
+    crm_trace("Handle orphans");
+
+    for (gIter = data_set->resources; gIter != NULL; gIter = gIter->next) {
+        resource_t *rsc = (resource_t *) gIter->data;
+        cleanup_orphans(rsc, data_set);
+    }
+
     crm_trace("Creating actions");
 
-    gIter = data_set->resources;
-    for (; gIter != NULL; gIter = gIter->next) {
+    for (gIter = data_set->resources; gIter != NULL; gIter = gIter->next) {
         resource_t *rsc = (resource_t *) gIter->data;
 
         rsc->cmds->create_actions(rsc, data_set);
@@ -1303,7 +1312,7 @@ any_managed_resources(pe_working_set_t * data_set)
 }
 
 /*
- * Create dependancies for stonith and shutdown operations
+ * Create dependencies for stonith and shutdown operations
  */
 gboolean
 stage6(pe_working_set_t * data_set)
@@ -1330,6 +1339,23 @@ stage6(pe_working_set_t * data_set)
 
         /* remote-nodes associated with a container resource (such as a vm) are not fenced */
         if (is_container_remote_node(node)) {
+            /* Guest */
+            if (need_stonith
+                && node->details->remote_requires_reset
+                && pe_can_fence(data_set, node)) {
+                resource_t *container = node->details->remote_rsc->container;
+                char *key = stop_key(container);
+                GListPtr stop_list = find_actions(container->actions, key, NULL);
+
+                crm_info("Implying node %s is down when container %s is stopped (%p)",
+                         node->details->uname, container->id, stop_list);
+                if(stop_list) {
+                    stonith_constraints(node, stop_list->data, data_set);
+                }
+
+                g_list_free(stop_list);
+                free(key);
+            }
             continue;
         }
 
@@ -1362,7 +1388,7 @@ stage6(pe_working_set_t * data_set)
 
             crm_notice("Scheduling Node %s for shutdown", node->details->uname);
 
-            down_op = custom_action(NULL, g_strdup_printf("%s-%s", CRM_OP_SHUTDOWN, node->details->uname),
+            down_op = custom_action(NULL, crm_strdup_printf("%s-%s", CRM_OP_SHUTDOWN, node->details->uname),
                                     CRM_OP_SHUTDOWN, node, FALSE, TRUE, data_set);
 
             shutdown_constraints(node, down_op, data_set);
@@ -1419,11 +1445,12 @@ stage6(pe_working_set_t * data_set)
         }
     }
 
-    if (last_stonith) {
-        order_actions(last_stonith, done, pe_order_implies_then);
 
-    } else if (dc_fence) {
+    if (dc_fence) {
         order_actions(dc_down, done, pe_order_implies_then);
+
+    } else if (last_stonith) {
+        order_actions(last_stonith, done, pe_order_implies_then);
     }
 
     order_actions(done, all_stopped, pe_order_implies_then);
@@ -1431,7 +1458,7 @@ stage6(pe_working_set_t * data_set)
 }
 
 /*
- * Determin the sets of independant actions and the correct order for the
+ * Determine the sets of independent actions and the correct order for the
  *  actions in each set.
  *
  * Mark dependencies of un-runnable actions un-runnable
@@ -1590,6 +1617,7 @@ rsc_order_first(resource_t * lh_rsc, order_constraint_t * order, pe_working_set_
 }
 
 extern gboolean update_action(action_t * action);
+extern void update_colo_start_chain(action_t * action);
 
 static void
 apply_remote_node_ordering(pe_working_set_t *data_set)
@@ -1604,9 +1632,38 @@ apply_remote_node_ordering(pe_working_set_t *data_set)
         resource_t *remote_rsc = NULL;
         resource_t *container = NULL;
 
+        if (action->rsc == NULL) {
+            continue;
+        }
+
+        /* Special case. */
+        if (action->rsc &&
+            action->rsc->is_remote_node &&
+            safe_str_eq(action->task, CRM_OP_CLEAR_FAILCOUNT)) {
+
+            /* if we are clearing the failcount of an actual remote node connect
+             * resource, then make sure this happens before allowing the connection
+             * to start if we are planning on starting the connection during this
+             * transition */ 
+            custom_action_order(action->rsc,
+                NULL,
+                action,
+                action->rsc,
+                generate_op_key(action->rsc->id, RSC_START, 0),
+                NULL,
+                pe_order_optional,
+                data_set);
+
+                continue;
+        }
+
+        /* detect if the action occurs on a remote node. if so create
+         * ordering constraints that guarantee the action occurs while
+         * the remote node is active (after start, before stop...) things
+         * like that */ 
         if (action->node == NULL ||
             is_remote_node(action->node) == FALSE ||
-            action->rsc == NULL ||
+            action->node->details->remote_rsc == NULL ||
             is_set(action->flags, pe_action_pseudo)) {
             continue;
         }
@@ -1617,6 +1674,7 @@ apply_remote_node_ordering(pe_working_set_t *data_set)
         if (safe_str_eq(action->task, "monitor") ||
             safe_str_eq(action->task, "start") ||
             safe_str_eq(action->task, "promote") ||
+            safe_str_eq(action->task, "notify") ||
             safe_str_eq(action->task, CRM_OP_LRM_REFRESH) ||
             safe_str_eq(action->task, CRM_OP_CLEAR_FAILCOUNT) ||
             safe_str_eq(action->task, "delete")) {
@@ -1680,20 +1738,176 @@ apply_remote_node_ordering(pe_working_set_t *data_set)
                 action,
                 pe_order_preserve | pe_order_implies_then | pe_order_runnable_left,
                 data_set);
-
         } else if (safe_str_eq(action->task, "stop")) {
-            custom_action_order(action->rsc,
-                NULL,
-                action,
-                remote_rsc,
-                generate_op_key(remote_rsc->id, RSC_STOP, 0),
-                NULL,
-                pe_order_preserve | pe_order_implies_first,
-                data_set);
+            gboolean after_start = FALSE;
+
+            /* handle special case with baremetal remote where stop actions need to be
+             * ordered after the connection resource starts somewhere else. */
+            if (is_baremetal_remote_node(action->node)) {
+                node_t *cluster_node = remote_rsc->running_on ? remote_rsc->running_on->data : NULL;
+
+                /* if the current cluster node a baremetal connection resource
+                 * is residing on is unclean or went offline we can't process any
+                 * operations on that remote node until after it starts somewhere else. */
+                if (cluster_node == NULL ||
+                    cluster_node->details->unclean == TRUE ||
+                    cluster_node->details->online == FALSE) {
+                    after_start = TRUE;
+                } else if (g_list_length(remote_rsc->running_on) > 1 &&
+                           remote_rsc->partial_migration_source &&
+                            remote_rsc->partial_migration_target) {
+                    /* if we're caught in the middle of migrating a connection resource,
+                     * then we have to wait until after the resource migrates before performing
+                     * any actions. */
+                    after_start = TRUE;
+                }
+            }
+
+            if (after_start) {
+                custom_action_order(remote_rsc,
+                    generate_op_key(remote_rsc->id, RSC_START, 0),
+                    NULL,
+                    action->rsc,
+                    NULL,
+                    action,
+                    pe_order_preserve | pe_order_implies_then | pe_order_runnable_left,
+                    data_set);
+            } else {
+                custom_action_order(action->rsc,
+                    NULL,
+                    action,
+                    remote_rsc,
+                    generate_op_key(remote_rsc->id, RSC_STOP, 0),
+                    NULL,
+                    pe_order_preserve | pe_order_implies_first,
+                    data_set);
+            }
         }
     }
 }
 
+static void
+order_probes(pe_working_set_t * data_set) 
+{
+#if 0
+    GListPtr gIter = NULL;
+
+    for (gIter = data_set->resources; gIter != NULL; gIter = gIter->next) {
+        resource_t *rsc = (resource_t *) gIter->data;
+
+        /* Given "A then B", we would prefer to wait for A to be
+         * started before probing B.
+         *
+         * If A was a filesystem on which the binaries and data for B
+         * lived, it would have been useful if the author of B's agent
+         * could assume that A is running before B.monitor will be
+         * called.
+         *
+         * However we can't _only_ probe once A is running, otherwise
+         * we'd not detect the state of B if A could not be started
+         * for some reason.
+         *
+         * In practice however, we cannot even do an opportunistic
+         * version of this because B may be moving:
+         *
+         *   B.probe -> B.start
+         *   B.probe -> B.stop
+         *   B.stop -> B.start
+         *   A.stop -> A.start
+         *   A.start -> B.probe
+         *
+         * So far so good, but if we add the result of this code:
+         *
+         *   B.stop -> A.stop
+         *
+         * Then we get a loop:
+         *
+         *   B.probe -> B.stop -> A.stop -> A.start -> B.probe
+         *
+         * We could kill the 'B.probe -> B.stop' dependency, but that
+         * could mean stopping B "too" soon, because B.start must wait
+         * for the probes to complete.
+         *
+         * Another option is to allow it only if A is a non-unique
+         * clone with clone-max == node-max (since we'll never be
+         * moving it).  However, we could still be stopping one
+         * instance at the same time as starting another.
+
+         * The complexity of checking for allowed conditions combined
+         * with the ever narrowing usecase suggests that this code
+         * should remain disabled until someone gets smarter.
+         */
+        action_t *start = NULL;
+        GListPtr actions = NULL;
+        GListPtr probes = NULL;
+        char *key = NULL;
+
+        key = start_key(rsc);
+        actions = find_actions(rsc->actions, key, NULL);
+        free(key);
+
+        if (actions) {
+            start = actions->data;
+            g_list_free(actions);
+        }
+
+        if(start == NULL) {
+            crm_err("No start action for %s", rsc->id);
+            continue;
+        }
+
+        key = generate_op_key(rsc->id, CRMD_ACTION_STATUS, 0);
+        probes = find_actions(rsc->actions, key, NULL);
+        free(key);
+
+        for (actions = start->actions_before; actions != NULL; actions = actions->next) {
+            action_wrapper_t *before = (action_wrapper_t *) actions->data;
+
+            GListPtr pIter = NULL;
+            action_t *first = before->action;
+            resource_t *first_rsc = first->rsc;
+
+            if(first->required_runnable_before) {
+                GListPtr clone_actions = NULL;
+                for (clone_actions = first->actions_before; clone_actions != NULL; clone_actions = clone_actions->next) {
+                    before = (action_wrapper_t *) clone_actions->data;
+
+                    crm_trace("Testing %s -> %s (%p) for %s", first->uuid, before->action->uuid, before->action->rsc, start->uuid);
+
+                    CRM_ASSERT(before->action->rsc);
+                    first_rsc = before->action->rsc;
+                    break;
+                }
+
+            } else if(safe_str_neq(first->task, RSC_START)) {
+                crm_trace("Not a start op %s for %s", first->uuid, start->uuid);
+            }
+
+            if(first_rsc == NULL) {
+                continue;
+
+            } else if(uber_parent(first_rsc) == uber_parent(start->rsc)) {
+                crm_trace("Same parent %s for %s", first_rsc->id, start->uuid);
+                continue;
+
+            } else if(FALSE && uber_parent(first_rsc)->variant < pe_clone) {
+                crm_trace("Not a clone %s for %s", first_rsc->id, start->uuid);
+                continue;
+            }
+
+            crm_err("Appplying %s before %s %d", first->uuid, start->uuid, uber_parent(first_rsc)->variant);
+
+            for (pIter = probes; pIter != NULL; pIter = pIter->next) {
+                action_t *probe = (action_t *) pIter->data;
+
+                crm_err("Ordering %s before %s", first->uuid, probe->uuid);
+                order_actions(first, probe, pe_order_optional);
+            }
+        }
+    }
+#endif
+}
+
 gboolean
 stage7(pe_working_set_t * data_set)
 {
@@ -1710,8 +1924,7 @@ stage7(pe_working_set_t * data_set)
      */
     data_set->ordering_constraints = g_list_reverse(data_set->ordering_constraints);
 
-    gIter = data_set->ordering_constraints;
-    for (; gIter != NULL; gIter = gIter->next) {
+    for (gIter = data_set->ordering_constraints; gIter != NULL; gIter = gIter->next) {
         order_constraint_t *order = (order_constraint_t *) gIter->data;
         resource_t *rsc = order->lh_rsc;
 
@@ -1734,10 +1947,18 @@ stage7(pe_working_set_t * data_set)
         }
     }
 
+    for (gIter = data_set->actions; gIter != NULL; gIter = gIter->next) {
+        action_t *action = (action_t *) gIter->data;
+
+        update_colo_start_chain(action);
+    }
+
+    crm_trace("Ordering probes");
+    order_probes(data_set);
+
     crm_trace("Updating %d actions", g_list_length(data_set->actions));
 
-    gIter = data_set->actions;
-    for (; gIter != NULL; gIter = gIter->next) {
+    for (gIter = data_set->actions; gIter != NULL; gIter = gIter->next) {
         action_t *action = (action_t *) gIter->data;
 
         update_action(action);
@@ -1745,8 +1966,7 @@ stage7(pe_working_set_t * data_set)
 
     crm_trace("Processing reloads");
 
-    gIter = data_set->resources;
-    for (; gIter != NULL; gIter = gIter->next) {
+    for (gIter = data_set->resources; gIter != NULL; gIter = gIter->next) {
         resource_t *rsc = (resource_t *) gIter->data;
 
         rsc_reload(rsc, data_set);
@@ -1800,6 +2020,35 @@ sort_notify_entries(gconstpointer a, gconstpointer b)
     return strcmp(entry_a->node->details->id, entry_b->node->details->id);
 }
 
+static char *
+expand_node_list(GListPtr list)
+{
+    GListPtr gIter = NULL;
+    char *node_list = NULL;
+
+    if (list == NULL) {
+        return strdup(" ");
+    }
+
+    for (gIter = list; gIter != NULL; gIter = gIter->next) {
+        node_t *node = (node_t *) gIter->data;
+
+        if (node->details->uname) {
+            int existing_len = 0;
+            int len = 2 + strlen(node->details->uname);
+
+            if(node_list) {
+                existing_len = strlen(node_list);
+            }
+            crm_trace("Adding %s (%dc) at offset %d", node->details->uname, len - 2, existing_len);
+            node_list = realloc_safe(node_list, len + existing_len);
+            sprintf(node_list + existing_len, "%s%s", existing_len == 0 ? "":" ", node->details->uname);
+        }
+    }
+
+    return node_list;
+}
+
 static void
 expand_list(GListPtr list, char **rsc_list, char **node_list)
 {
@@ -1861,8 +2110,8 @@ expand_list(GListPtr list, char **rsc_list, char **node_list)
             }
 
             crm_trace("Adding %s (%dc) at offset %d", rsc_id, len - 2, existing_len);
-            *rsc_list = realloc(*rsc_list, len + existing_len);
-            sprintf(*rsc_list + existing_len, "%s ", rsc_id);
+            *rsc_list = realloc_safe(*rsc_list, len + existing_len);
+            sprintf(*rsc_list + existing_len, "%s%s", existing_len == 0 ? "":" ", rsc_id);
         }
 
         if (entry->node != NULL) {
@@ -1878,8 +2127,8 @@ expand_list(GListPtr list, char **rsc_list, char **node_list)
             }
 
             crm_trace("Adding %s (%dc) at offset %d", uname, len - 2, existing_len);
-            *node_list = realloc(*node_list, len + existing_len);
-            sprintf(*node_list + existing_len, "%s ", uname);
+            *node_list = realloc_safe(*node_list, len + existing_len);
+            sprintf(*node_list + existing_len, "%s%s", existing_len == 0 ? "":" ", uname);
         }
     }
 
@@ -2096,6 +2345,10 @@ collect_notification_data(resource_t * rsc, gboolean state, gboolean activity,
                           notify_data_t * n_data)
 {
 
+    if(n_data->allowed_nodes == NULL) {
+        n_data->allowed_nodes = rsc->allowed_nodes;
+    }
+
     if (rsc->children) {
         GListPtr gIter = rsc->children;
 
@@ -2178,7 +2431,7 @@ collect_notification_data(resource_t * rsc, gboolean state, gboolean activity,
 }
 
 gboolean
-expand_notification_data(notify_data_t * n_data)
+expand_notification_data(notify_data_t * n_data, pe_working_set_t * data_set)
 {
     /* Expand the notification entries into a key=value hashtable
      * This hashtable is later used in action2xml()
@@ -2186,6 +2439,7 @@ expand_notification_data(notify_data_t * n_data)
     gboolean required = FALSE;
     char *rsc_list = NULL;
     char *node_list = NULL;
+    GListPtr nodes = NULL;
 
     if (n_data->stop) {
         n_data->stop = g_list_sort(n_data->stop, sort_notify_entries);
@@ -2257,6 +2511,14 @@ expand_notification_data(notify_data_t * n_data)
     expand_list(n_data->inactive, &rsc_list, NULL);
     g_hash_table_insert(n_data->keys, strdup("notify_inactive_resource"), rsc_list);
 
+    nodes = g_hash_table_get_values(n_data->allowed_nodes);
+    node_list = expand_node_list(nodes);
+    g_hash_table_insert(n_data->keys, strdup("notify_available_uname"), node_list);
+    g_list_free(nodes);
+
+    node_list = expand_node_list(data_set->nodes);
+    g_hash_table_insert(n_data->keys, strdup("notify_all_uname"), node_list);
+
     if (required && n_data->pre) {
         update_action_flags(n_data->pre, pe_action_optional | pe_action_clear);
         update_action_flags(n_data->pre_done, pe_action_optional | pe_action_clear);
@@ -2322,6 +2584,16 @@ create_notifications(resource_t * rsc, notify_data_t * n_data, pe_working_set_t
             for (; gIter != NULL; gIter = gIter->next) {
                 node_t *current_node = (node_t *) gIter->data;
 
+                /* if this stop action is a pseudo action as a result of the current
+                 * node being fenced, this stop action is implied by the fencing 
+                 * action. There's no reason to send the fenced node a stop notification */ 
+                if (stop &&
+                    is_set(stop->flags, pe_action_pseudo) &&
+                    current_node->details->unclean) {
+
+                    continue;
+                }
+
                 pe_notify(rsc, current_node, n_data->pre, n_data->pre_done, n_data, data_set);
                 if (task == action_demote || stop == NULL
                     || is_set(stop->flags, pe_action_optional)) {
