@@ -31,6 +31,7 @@ info Performing the following tests from $io_dir
 create_mode="false"
 
 echo ""
+
 do_test simple1 "Offline     "
 do_test simple2 "Start       "
 do_test simple3 "Start 2     "
@@ -66,7 +67,7 @@ do_test bug-lf-2619 "Move group on clone failure"
 do_test group-fail "Ensure stop order is preserved for partially active groups"
 do_test group-unmanaged "No need to restart r115 because r114 is unmanaged"
 do_test group-unmanaged-stopped "Make sure r115 is stopped when r114 fails"
-do_test group-dependants "Account for the location preferences of things colocated with a group"
+do_test group-dependents "Account for the location preferences of things colocated with a group"
 
 echo ""
 do_test rsc_dep1 "Must not     "
@@ -94,6 +95,7 @@ do_test novell-251689 "Resource definition change + target_role=stopped"
 do_test bug-lf-2106 "Restart all anonymous clone instances after config change"
 do_test params-6 "Params: Detect reload in previously migrated resource"
 do_test nvpair-id-ref "Support id-ref in nvpair with optional name"
+do_test not-reschedule-unneeded-monitor "Do not reschedule unneeded monitors while resource definitions have changed"
 
 echo ""
 do_test target-0 "Target Role : baseline"
@@ -112,7 +114,7 @@ do_test probe-0 "Probe (anon clone)"
 do_test probe-1 "Pending Probe"
 do_test probe-2 "Correctly re-probe cloned groups"
 do_test probe-3 "Probe (pending node)"
-do_test probe-4 "Probe (pending node + stopped resource)" --rc 4
+do_test probe-4 "Probe (pending node + stopped resource)"
 do_test standby "Standby"
 do_test comments "Comments"
 
@@ -126,6 +128,17 @@ do_test one-or-more-5 "Start A and F even though C and D are stopped"
 do_test one-or-more-6 "Leave A running even though B is stopped"
 do_test one-or-more-7 "Leave A running even though C is stopped"
 do_test bug-5140-require-all-false "Allow basegrp:0 to stop"
+do_test clone-require-all-1 "clone B starts node 3 and 4"
+do_test clone-require-all-2 "clone B remains stopped everywhere"
+do_test clone-require-all-3 "clone B stops everywhere because A stops everywhere"
+do_test clone-require-all-4 "clone B remains on node 3 and 4 with only one instance of A remaining."
+do_test clone-require-all-5 "clone B starts on node 1 3 and 4"
+do_test clone-require-all-6 "clone B remains active after shutting down instances of A"
+do_test clone-require-all-7 "clone A and B both start at the same time. all instances of A start before B."
+do_test clone-require-all-no-interleave-1 "C starts everywhere after A and B"
+do_test clone-require-all-no-interleave-2 "C starts on nodes 1, 2, and 4 with only one active instance of B"
+do_test clone-require-all-no-interleave-3 "C remains active when instance of B is stopped on one node and started on another."
+do_test one-or-more-unrunnnable-instances "Avoid dependencies on instances that wont ever be started"
 
 echo ""
 do_test order1 "Order start 1     "
@@ -138,7 +151,7 @@ do_test order7 "Order (manditory)  "
 do_test order-optional "Order (score=0)  "
 do_test order-required "Order (score=INFINITY)  "
 do_test bug-lf-2171 "Prevent group start when clone is stopped"
-do_test order-clone "Clone ordering should be able to prevent startup of dependant clones"
+do_test order-clone "Clone ordering should be able to prevent startup of dependent clones"
 do_test order-sets "Ordering for resource sets"
 do_test order-serialize "Serialize resources without inhibiting migration"
 do_test order-serialize-set "Serialize a set of resources without inhibiting migration"
@@ -148,6 +161,7 @@ do_test order-optional-keyword "Order (optional keyword)"
 do_test order-mandatory "Order (mandatory keyword)"
 do_test bug-lf-2493 "Don't imply colocation requirements when applying ordering constraints with clones"
 do_test ordered-set-basic-startup "Constraint set with default order settings."
+do_test ordered-set-natural "Allow natural set ordering"
 do_test order-wrong-kind "Order (error)"
 
 echo ""
@@ -160,11 +174,13 @@ do_test coloc-attr "Colocation based on node attributes"
 do_test coloc-negative-group "Negative colocation with a group"
 do_test coloc-intra-set "Intra-set colocation"
 do_test bug-lf-2435 "Colocation sets with a negative score"
-do_test coloc-clone-stays-active "Ensure clones don't get stopped/demoted because a dependant must stop"
+do_test coloc-clone-stays-active "Ensure clones don't get stopped/demoted because a dependent must stop"
 do_test coloc_fp_logic "Verify floating point calculations in colocation are working"
 do_test colo_master_w_native "cl#5070 - Verify promotion order is affected when colocating master to native rsc."
 do_test colo_slave_w_native  "cl#5070 - Verify promotion order is affected when colocating slave to native rsc."
 do_test anti-colocation-order "cl#5187 - Prevent resources in an anti-colocation from even temporarily running on a same node"
+do_test enforce-colo1 "Always enforce B with A INFINITY."
+do_test complex_enforce_colo "Always enforce B with A INFINITY. (make sure heat-engine stops)"
 
 echo ""
 do_test rsc-sets-seq-true "Resource Sets - sequential=false"
@@ -210,6 +226,7 @@ do_test monitor-recovery "on-fail=block + resource recovery detected by recurrin
 do_test stop-failure-no-quorum "Stop failure without quorum"
 do_test stop-failure-no-fencing "Stop failure without fencing available"
 do_test stop-failure-with-fencing "Stop failure with fencing available"
+do_test multiple-active-block-group "Support of multiple-active=block for resource groups"
 
 echo ""
 do_test quorum-1 "No quorum - ignore"
@@ -292,6 +309,9 @@ do_test 9-am-then-bm-b-migrating-a-stopping "Advanced migrate logic, A then B, B
 do_test 10-a-then-bm-b-move-a-clone "Advanced migrate logic, A clone then B, migrate B while stopping A"
 do_test 11-a-then-bm-b-move-a-clone-starting "Advanced migrate logic, A clone then B, B moving while A is start/stopping"
 
+do_test a-promote-then-b-migrate "A promote then B start. migrate B"
+do_test a-demote-then-b-migrate "A demote then B stop. migrate B"
+
 #echo ""
 #do_test complex1 "Complex	"
 
@@ -316,10 +336,10 @@ do_test inc11 "Primitive colocation with clones"
 do_test inc12 "Clone shutdown"
 do_test cloned-group "Make sure only the correct number of cloned groups are started"
 do_test cloned-group-stop "Ensure stopping qpidd also stops glance and cinder"
-do_test clone-no-shuffle "Dont prioritize allocation of instances that must be moved"
+do_test clone-no-shuffle "Don't prioritize allocation of instances that must be moved"
 do_test clone-max-zero "Orphan processing with clone-max=0"
 do_test clone-anon-dup "Bug LF#2087 - Correctly parse the state of anonymous clones that are active more than once per node"
-do_test bug-lf-2160 "Dont shuffle clones due to colocation"
+do_test bug-lf-2160 "Don't shuffle clones due to colocation"
 do_test bug-lf-2213 "clone-node-max enforcement for cloned groups"
 do_test bug-lf-2153 "Clone ordering constraints"
 do_test bug-lf-2361 "Ensure clones observe mandatory ordering constraints if the LHS is unrunnable"
@@ -340,6 +360,21 @@ do_test clone-interleave-1 "Clone-3 cannot start on pcmk-1 due to interleaved or
 do_test clone-interleave-2 "Clone-3 must stop on pcmk-1 due to interleaved ordering (no colocation)"
 do_test clone-interleave-3 "Clone-3 must be recovered on pcmk-1 due to interleaved ordering (no colocation)"
 
+echo ""
+do_test cloned_start_one  "order first clone then clone... first clone_min=2"
+do_test cloned_start_two  "order first clone then clone... first clone_min=2"
+do_test cloned_stop_one   "order first clone then clone... first clone_min=2"
+do_test cloned_stop_two   "order first clone then clone... first clone_min=2"
+do_test clone_min_interleave_start_one "order first clone then clone... first clone_min=2 and then has interleave=true"
+do_test clone_min_interleave_start_two "order first clone then clone... first clone_min=2 and then has interleave=true"
+do_test clone_min_interleave_stop_one  "order first clone then clone... first clone_min=2 and then has interleave=true"
+do_test clone_min_interleave_stop_two  "order first clone then clone... first clone_min=2 and then has interleave=true"
+do_test clone_min_start_one "order first clone then primitive... first clone_min=2"
+do_test clone_min_start_two "order first clone then primitive... first clone_min=2"
+do_test clone_min_stop_all  "order first clone then primitive... first clone_min=2"
+do_test clone_min_stop_one  "order first clone then primitive... first clone_min=2"
+do_test clone_min_stop_two  "order first clone then primitive... first clone_min=2"
+
 echo ""
 do_test unfence-startup "Clean unfencing"
 do_test unfence-definition "Unfencing when the agent changes"
@@ -367,8 +402,8 @@ do_test master-ordering "Prevent resources from starting that need a master"
 do_test bug-1765 "Master-Master Colocation (dont stop the slaves)"
 do_test master-group "Promotion of cloned groups"
 do_test bug-lf-1852 "Don't shuffle master/slave instances unnecessarily"
-do_test master-failed-demote "Dont retry failed demote actions"
-do_test master-failed-demote-2 "Dont retry failed demote actions (notify=false)"
+do_test master-failed-demote "Don't retry failed demote actions"
+do_test master-failed-demote-2 "Don't retry failed demote actions (notify=false)"
 do_test master-depend "Ensure resources that depend on the master don't get allocated until the master does"
 do_test master-reattach "Re-attach to a running master"
 do_test master-allow-start "Don't include master score if it would prevent allocation"
@@ -390,11 +425,13 @@ do_test master_monitor_restart "cl#5072 - Ensure master monitor operation will s
 do_test bug-rh-880249 "Handle replacement of an m/s resource with a primitive"
 do_test bug-5143-ms-shuffle "Prevent master shuffling due to promotion score"
 do_test master-demote-block "Block promotion if demote fails with on-fail=block"
-do_test master-dependant-ban "Don't stop instances from being active because a dependant is banned from that host"
+do_test master-dependent-ban "Don't stop instances from being active because a dependent is banned from that host"
 do_test master-stop "Stop instances due to location constraint with role=Started"
 do_test master-partially-demoted-group "Allow partially demoted group to finish demoting"
 do_test bug-cl-5213 "Ensure role colocation with -INFINITY is enforced"
 do_test bug-cl-5219 "Allow unrelated resources with a common colocation target to remain promoted"
+do_test master-asymmetrical-order "Fix the behaviors of multi-state resources with asymmetrical ordering"
+do_test master-notify "Master promotion with notifies"
 
 echo ""
 do_test history-1 "Correctly parse stateful-1 resource state"
@@ -419,7 +456,7 @@ do_test interleave-2 "coloc - interleaved   "
 do_test interleave-3 "coloc - interleaved (2)"
 do_test interleave-pseudo-stop "Interleaved clone during stonith"
 do_test interleave-stop "Interleaved clone during stop"
-do_test interleave-restart "Interleaved clone during dependancy restart"
+do_test interleave-restart "Interleaved clone during dependency restart"
 
 echo ""
 do_test notify-0 "Notify reference"
@@ -439,7 +476,7 @@ do_test 764 "OSDL #764 - Missing monitor op for DoFencing:child_DoFencing:1"
 do_test 797 "OSDL #797 - Assert triggered: task_id_i > max_call_id"
 do_test 829 "OSDL #829"
 do_test 994 "OSDL #994 - Stopping the last resource in a resource group causes the entire group to be restarted"
-do_test 994-2 "OSDL #994 - with a dependant resource"
+do_test 994-2 "OSDL #994 - with a dependent resource"
 do_test 1360 "OSDL #1360 - Clone stickiness"
 do_test 1484 "OSDL #1484 - on_fail=stop"
 do_test 1494 "OSDL #1494 - Clone stability"
@@ -448,11 +485,11 @@ do_test stonith-0 "Stonith loop - 1"
 do_test stonith-1 "Stonith loop - 2"
 do_test stonith-2 "Stonith loop - 3"
 do_test stonith-3 "Stonith startup"
-do_test stonith-4 "Stonith node state" --rc 4
+do_test stonith-4 "Stonith node state"
 do_test bug-1572-1 "Recovery of groups depending on master/slave"
 do_test bug-1572-2 "Recovery of groups depending on master/slave when the master is never re-promoted"
 do_test bug-1685 "Depends-on-master ordering"
-do_test bug-1822 "Dont promote partially active groups"
+do_test bug-1822 "Don't promote partially active groups"
 do_test bug-pm-11 "New resource added to a m/s group"
 do_test bug-pm-12 "Recover only the failed portion of a cloned group"
 do_test bug-n-387749 "Don't shuffle clone instances"
@@ -538,6 +575,15 @@ do_test utilization-order3 "Utilization Order - Migrate"
 do_test utilization-order4 "Utilization Order - Live Mirgration (bnc#695440)"
 do_test utilization-shuffle "Don't displace prmExPostgreSQLDB2 on act2, Start prmExPostgreSQLDB1 on act3"
 do_test load-stopped-loop "Avoid transition loop due to load_stopped (cl#5044)"
+do_test load-stopped-loop-2 "cl#5235 - Prevent graph loops that can be introduced by load_stopped -> migrate_to ordering"
+
+echo ""
+do_test colocated-utilization-primitive-1 "Colocated Utilization - Primitive"
+do_test colocated-utilization-primitive-2 "Colocated Utilization - Choose the most capable node"
+do_test colocated-utilization-group "Colocated Utilization - Group"
+do_test colocated-utilization-clone "Colocated Utilization - Clone"
+
+do_test utilization-check-allowed-nodes "Only check the capacities of the nodes that can run the resource"
 
 echo ""
 do_test reprobe-target_rc "Ensure correct target_rc for reprobe of inactive resources"
@@ -742,6 +788,7 @@ do_test container-group-3 "Container in group - stop failed"
 do_test container-group-4 "Container in group - reached migration-threshold"
 do_test container-is-remote-node "Place resource within container when container is remote-node"
 do_test bug-rh-1097457 "Kill user defined container/contents ordering"
+do_test bug-cl-5247 "Graph loop when recovering m/s resource in a container"
 
 echo ""
 do_test whitebox-fail1 "Fail whitebox container rsc."
@@ -756,14 +803,35 @@ do_test whitebox-orphaned    "Properly shutdown orphaned whitebox container"
 do_test whitebox-orphan-ms   "Properly tear down orphan ms resources on remote-nodes"
 do_test whitebox-unexpectedly-running "Recover container nodes the cluster did not start."
 do_test whitebox-migrate1 "Migrate both container and connection resource"
+do_test whitebox-imply-stop-on-fence "imply stop action on container node rsc when host node is fenced"
+do_test whitebox-nested-group "Verify guest remote-node works nested in a group"
 
 echo ""
 do_test remote-startup-probes  "Baremetal remote-node startup probes"
 do_test remote-startup         "Startup a newly discovered remote-nodes with no status."
 do_test remote-fence-unclean   "Fence unclean baremetal remote-node"
+do_test remote-fence-unclean2  "Fence baremetal remote-node after cluster node fails and connection can not be recovered"
 do_test remote-move            "Move remote-node connection resource"
 do_test remote-disable         "Disable a baremetal remote-node"
 do_test remote-orphaned        "Properly shutdown orphaned connection resource"
+do_test remote-orphaned2       "verify we can handle orphaned remote connections with active resources on the remote"
+do_test remote-recover         "Recover connection resource after cluster-node fails."
 do_test remote-stale-node-entry "Make sure we properly handle leftover remote-node entries in the node section"
+do_test remote-partial-migrate  "Make sure partial migrations are handled before ops on the remote node."
+do_test remote-partial-migrate2 "Make sure partial migration target is prefered for remote connection."
+do_test remote-recover-fail     "Make sure start failure causes fencing if rsc are active on remote."
+do_test remote-start-fail       "Make sure a start failure does not result in fencing if no active resources are on remote."
+do_test remote-unclean2         "Make monitor failure always results in fencing, even if no rsc are active on remote."
+do_test remote-fence-before-reconnect "Fence before clearing recurring monitor failure"
+
+echo ""
+do_test resource-discovery      "Exercises resource-discovery location constraint option."
+do_test rsc-discovery-per-node  "Disable resource discovery per node"
+
+echo ""
+do_test isolation-start-all   "Start docker isolated resources."
+do_test isolation-restart-all "Restart docker isolated resources."
+do_test isolation-clone       "Cloned isolated primitive."
+
 echo ""
 test_results
